#!/bin/sh
# Copyright (C) 2007, 2008  Condel project, TU-Berlin.
# Written by Sven Grottke, Jan Sablatnig, Jiehua Chen.
# 
# This file is part of Adam, the DVE testbed.
# 
# Adam is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3, or (at your option) any later
# version.
# 
# Adam is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
# 
# You should have received a copy of the GNU General Public License
# along with Adam; see the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.
# 
# Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.

# usage: bootstrap.sh [-c]
# options: -c: set up for commandline interface (default is Tkenv)
#          -f: force complete rebuild of Akaroa and OMNeT++
failure() {
    echo "error: $@"
    exit 1
}

clear_akaroa() {
    # remove the lock file if it exists
    [ -f "${AKAROA_LOCK}" ] && rm -f "${AKAROA_LOCK}" || return 0

    # make clean & remove extra files
    [ -d "${AKAROA_BASE}" ] || {
        echo "Akaroa does not exist, skipping setup..."
        return 1
    }
    pushd "${AKAROA_BASE}" 2>&1 1>/dev/null
    make clean || failure "make clean failed in ${AKAROA_BASE}"
    rm -Rf Makefile.config_tools Makefile.config_install src/config src/Makefile.dependencies || failure "could not remove extra Akaroa files"
    popd 2>&1 1>/dev/null

    return 0
}

setup_akaroa() {
    echo "rebuilding akaroa"
    [ -d "${AKAROA_BASE}" ] || {
        echo "Akaroa does not exist, skipping Akaroa setup"
        return 1
    }
    pushd "${AKAROA_BASE}" 2>&1 1>/dev/null
    ./configure --prefix="${AKAROA_INSTALL}" && echo "Akaroa configured" || failure "configure failed in ${AKAROA_BASE}"
    make || failure "make failed in ${AKAROA_BASE}"
    make install || failure "make install failed in ${AKAROA_BASE}"
    # clean the .svn stuff from the install destination
    [ -e "${AKAROA_INSTALL}/include/akaroa/.svn" ] && rm -Rf "${AKAROA_INSTALL}/include/akaroa/.svn"
    popd 2>&1 1>/dev/null

    # create lock file
    touch "${AKAROA_LOCK}"
    
    return 0    
}

clear_omnetpp() {
    # remove lock file if it exists
    [ -f "${OMNETPP_LOCK}" ] && rm -f "${OMNETPP_LOCK}"

    # make clean if Makefile exists
    pushd "${OMNETPP_BASE}" 2>&1 1>/dev/null
    [ -f "${OMNETPP_BASE}/Makefile" ] && make clean
    popd 2>&1 1>/dev/null

    # remove Makefile if it exists
    [ -f "${OMNETPP_BASE}/Makefile" ] && rm "${OMNETPP_BASE}/Makefile"

    # remove configure.user, config.log and config.status
    [ -f "${OMNETPP_CONFIG}" ] && rm "${OMNETPP_CONFIG}"
    [ -f "${OMNETPP_BASE}/config.status" ] && rm "${OMNETPP_BASE}/config.status"
    [ -f "${OMNETPP_BASE}/config.log" ] && rm "${OMNETPP_BASE}/config.log"
}

setup_omnetpp() {
    echo "rebuilding OMNeT++"
    
    # if configure.user does not exist, create it from the stub
    [ -f "${OMNETPP_CONFIG}" ] || {
        cat "${OMNETPP_BASE}/configure.user.stub" > "${OMNETPP_CONFIG}"
        echo "AKAROA_CFLAGS=\"-I${AKAROA_INSTALL}/include\"" >> "${OMNETPP_CONFIG}"
        echo "AKAROA_LIBS=\"-L${AKAROA_INSTALL}/lib -lakaroa -lfl\"" >> "${OMNETPP_CONFIG}"
    }

    # if Makefile does not exists, run configure
    [ -f "${OMNETPP_BASE}/Makefile" ] || {
        pushd "${OMNETPP_BASE}" 2>&1 1>/dev/null
        ./configure || failure "OMNeT++ configure failed"
        popd 2>&1 1>/dev/null
    }

    # call make
    pushd "${OMNETPP_BASE}" 2>&1 1>/dev/null
    make || failure "make failed in ${OMNETPP_BASE}"
    popd 2>&1 1>/dev/null
    
    # create lock file
    touch "${OMNETPP_LOCK}"

    return 0
}

# set base directory
pushd `dirname $0` 2>&1 1>/dev/null 
CONDEL_ROOT=`pwd`
popd 2>&1 1>/dev/null

# set default arguments
CMDENV="Tkenv"

# set Akaroa base path
AKAROA_BASE="${CONDEL_ROOT}/akaroa"
# DO NOT EVER CHANGE THIS!!! clear_akaroa() involves a rm -Rf on this,
# so setting it to something like ${HOME} might give you a very bad surprise
AKAROA_INSTALL="${AKAROA_BASE}/install"
AKAROA_LOCK="${AKAROA_BASE}/has_been_configured"

# set OMNeT++ base path
OMNETPP_BASE="${CONDEL_ROOT}/omnetpp"
OMNETPP_LOCK="${OMNETPP_BASE}/has_been_configured"
OMNETPP_CONFIG="${OMNETPP_BASE}/configure.user"

# modify PATH
PATH="${PATH}:${OMNETPP_BASE}/bin"
export PATH

# parse arguments
while [ $# -gt 0 ]; do
    case $1 in
        "-c") 
            CMDENV="Cmdenv"
            ;;
        "-f")
            echo "forcing rebuild of Akaroa and OMNeT++"
            clear_akaroa
            clear_omnetpp
            ;;
    esac
    shift
done

# setup Akaroa if it has not been done before
[ -f "${AKAROA_LOCK}" ] && echo "Akaroa already setup" || setup_akaroa

# setup ONeT++ if it has not been done before
[ -f "${OMNETPP_LOCK}" ] && echo "OMNeT++ already setup" || setup_omnetpp


# set makemake utility
MAKEMAKE="${OMNETPP_BASE}/bin/opp_makemake"

# create config file
CONFIG_FILE="${CONDEL_ROOT}/omnetppconfig"
${MAKEMAKE} -N -f --genconfig "${CONFIG_FILE}" || {
    failure "could not create ${CONFIG_FILE}"
}

# subdirectories that have to be compiled
SUBDIRS="general scenarios scenarios/pong scenarios/swarm consistency consistency/loose consistency/opt replication replication/total experiments experiments/pong_loose experiments/pong_opt experiments/pong_loose/src experiments/pong_opt/src experiments/swarm_loose experiments/swarm_opt experiments/swarm_loose/src experiments/swarm_opt/src experiments/timeout experiments/timeout/src"

# create include directives
SUBDIR_INC="-I${CONDEL_ROOT}"
for d in ${SUBDIRS}; do
    SUBDIR_INC="${SUBDIR_INC} -I${CONDEL_ROOT}/${d}"
done

# directories where no Makefiles will be generated
EXCLUDE_DIRS="docs util run ourbitmaps akaroa omnetpp"
EXCLUDES=""
for d in ${EXCLUDE_DIRS}; do
    EXCLUDES="${EXCLUDES} -X ${CONDEL_ROOT}/${d}"
done

# experiment-specific dependencies
PONG_LOOSE_BIN="experiments/pong_loose/bin"
PONG_LOOSE_DIRS="general scenarios scenarios/pong consistency consistency/loose replication replication/total experiments experiments/pong_loose experiments/pong_loose/src"
PONG_LOOSE_INC="-I${CONDEL_ROOT}"
for d in ${PONG_LOOSE_DIRS}; do
    PONG_LOOSE_INC="${PONG_LOOSE_INC} -I${CONDEL_ROOT}/${d}"
done

PONG_IDEAL_BIN="experiments/pong_ideal/bin"

PONG_OPT_BIN="experiments/pong_opt/bin"
PONG_OPT_DIRS="general scenarios scenarios/pong consistency consistency/opt replication replication/total experiments experiments/pong_opt experiments/pong_opt/src"
PONG_OPT_INC="-I${CONDEL_ROOT}"
for d in ${PONG_OPT_DIRS}; do
    PONG_OPT_INC="${PONG_OPT_INC} -I${CONDEL_ROOT}/${d}"
done

# experiment-specific dependencies
SWARM_LOOSE_BIN="experiments/swarm_loose/bin"
SWARM_LOOSE_DIRS="general scenarios scenarios/swarm consistency consistency/loose replication replication/total experiments experiments/swarm_loose experiments/swarm_loose/src"
SWARM_LOOSE_INC="-I${CONDEL_ROOT}"
for d in ${SWARM_LOOSE_DIRS}; do
    SWARM_LOOSE_INC="${SWARM_LOOSE_INC} -I${CONDEL_ROOT}/${d}"
done

SWARM_IDEAL_BIN="experiments/swarm_ideal/bin"

SWARM_OPT_BIN="experiments/swarm_opt/bin"
SWARM_OPT_DIRS="general scenarios scenarios/swarm consistency consistency/opt replication replication/total experiments experiments/swarm_opt experiments/swarm_opt/src"
SWARM_OPT_INC="-I${CONDEL_ROOT}"
for d in ${SWARM_OPT_DIRS}; do
    SWARM_OPT_INC="${SWARM_OPT_INC} -I${CONDEL_ROOT}/${d}"
done

TIMEOUT_BIN="experiments/timeout/bin"
TIMEOUT_DIRS="general experiments experiments/timeout experiments/timeout/src"
TIMEOUT_INC="-I${CONDEL_ROOT}"
for d in ${TIMEOUT_DIRS}; do
    TIMEOUT_INC="${TIMEOUT_INC} -I${CONDEL_ROOT}/${d}"
done

# set common options for opp_makemake
OPP_FLAGS="-N -f -c ${CONFIG_FILE} -x -b ${CONDEL_ROOT} -u ${CMDENV}"

# create Makefiles in all subdirectories
for d in ${SUBDIRS}; do
    echo "making Makefiles in ${d}"
    cd "${d}" && ${MAKEMAKE} ${OPP_FLAGS} -n -r ${SUBDIR_INC}
    cd ${CONDEL_ROOT}
done

# make the top level directory
cd ${CONDEL_ROOT}/experiments
${MAKEMAKE} ${OPP_FLAGS} -r -n -X common_files -X pong_ideal -X swarm_ideal
cd ${CONDEL_ROOT}

# make the actual executables for each experiment
cd ${PONG_LOOSE_BIN}
${MAKEMAKE} ${OPP_FLAGS} -r ${PONG_LOOSE_INC} -w -o simulation || failure "opp_makemake failed"
cd ..
${MAKEMAKE} ${OPP_FLAGS} -n -r ${SUBDIR_INC} -X results -X scripts && cd ${CONDEL_ROOT}
cd ${CONDEL_ROOT}

cd ${PONG_OPT_BIN}
${MAKEMAKE} ${OPP_FLAGS} -r ${PONG_OPT_INC} -w -o simulation || failure "opp_makemake failed"
cd ..
${MAKEMAKE} ${OPP_FLAGS} -n -r ${SUBDIR_INC} -X results -X scripts && cd ${CONDEL_ROOT}
cd ${CONDEL_ROOT}

# make the actual executables for each experiment
cd ${SWARM_LOOSE_BIN}
${MAKEMAKE} ${OPP_FLAGS} -r ${SWARM_LOOSE_INC} -w -o simulation || failure "opp_makemake failed"
cd ..
${MAKEMAKE} ${OPP_FLAGS} -n -r ${SUBDIR_INC} -X results -X scripts && cd ${CONDEL_ROOT}
cd ${CONDEL_ROOT}

cd ${SWARM_OPT_BIN}
${MAKEMAKE} ${OPP_FLAGS} -r ${SWARM_OPT_INC} -w -o simulation || failure "opp_makemake failed"
cd ..
${MAKEMAKE} ${OPP_FLAGS} -n -r ${SUBDIR_INC} -X results -X scripts && cd ${CONDEL_ROOT}
cd ${CONDEL_ROOT}


cd ${TIMEOUT_BIN}
${MAKEMAKE} ${OPP_FLAGS} -r ${TIMEOUT_INC} -w -o simulation || failure "opp_makemake failed"
cd ..
${MAKEMAKE} ${OPP_FLAGS} -n -r ${SUBDIR_INC} -X results -X scripts && cd ${CONDEL_ROOT}
cd ${CONDEL_ROOT}

# the main directory
${MAKEMAKE} ${OPP_FLAGS} -n -r ${EXCLUDES} -i makefrag ${SUBDIR_INC} || {
    failure "opp_makemake failed"
}

# create wrapper scripts
BIN_DIRS="${PONG_LOOSE_BIN} ${PONG_IDEAL_BIN} ${PONG_OPT_BIN} ${TIMEOUT_BIN} ${SWARM_LOOSE_BIN} ${SWARM_IDEAL_BIN} ${SWARM_OPT_BIN}"
for d in ${BIN_DIRS}; do
    f="${d}/simulation.sh"
    echo "creating ${f}"
    cat >"${f}" <<EOF
#!/bin/sh

LD_LIBRARY_PATH="\${LD_LIBRARY_PATH}\${LD_LIBRARY_PATH:+":"}${OMNETPP_BASE}/lib"
export LD_LIBRARY_PATH

exec "${CONDEL_ROOT}/${d}/simulation" "\$@"
EOF
    chmod a+rx "${f}"
done

# link swarm_ideal binaries
echo "creating symlinks for ideal consistency"
[ -h "${PONG_IDEAL_BIN}/simulation" ] || ln -s "${CONDEL_ROOT}/${PONG_LOOSE_BIN}/simulation" "${PONG_IDEAL_BIN}/simulation"
[ -h "${SWARM_IDEAL_BIN}/simulation" ] || ln -s "${CONDEL_ROOT}/${SWARM_LOOSE_BIN}/simulation" "${SWARM_IDEAL_BIN}/simulation"
