/*
Copyright (C) 2007, 2008  Condel project, TU-Berlin.
Written by Jan Sablatnig, Sven Grottke, Jiehua Chen.

This file is part of Adam, the DVE testbed.

Adam is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3, or (at your option) any later
version.

Adam is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Adam; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.
*/
#include "consistency_loose.hh"
#include <omnetpp.h>
#include "instance.hh"
#include "instance_factory.hh"
#include "scenario.hh"
#include "message_router.hh"
#include "host.hh"

cLooseConsistency::cLooseConsistency(cHost& Host)
: cConsistency(Host), m_ghost(Host, "ghost"),
  m_regularUpdates(0.0), m_updateTimer(0.0),
  m_isOwner(), m_maxDistance(), m_timeout(0.0)
{ }

void cLooseConsistency::initialize(void)
{
  cConsistency::initialize();

  m_ghost.clear();

  m_regularUpdates= host().par("regular_updates");

  if (host().hasPar("timeout"))
    m_timeout= host().par("timeout");
  else
    m_timeout= 0.0;

  m_updateTimer= 0.0;

  m_isOwner.clear();

  m_maxDistance.clear();
}

void cLooseConsistency::addInstance(const cInstance& instance)
{
  // Create an insertion-message
  scoped_ptr<cMsg> msg(host().getMsgFactory().createMsg("AddInstance",
                                                          m_timeout));

  msg->setObjectID(instance.getObjectID());
  msg->setObjectName(instance.getName().c_str());

  msg->setAttributes(instance.getAttributes());

  // Now, send that message to myself and to the routing layer
  host().getRouter().send(msg.get());
  setOwner(instance.getObjectID());

  host().getLoad().addCPU(11);          // wild guess at instantiation cost (10)
}

void cLooseConsistency::addInstance(const cInstance& instance, double maxdist)
{
    addInstance(instance);
    setMaxdist(instance.getObjectID(), maxdist);
}

void cLooseConsistency::handleTick(void)
{
  const double delta = host().now() - m_lastTickTime;

  host().getLoad().addMem(host().getWorld().getTotalLength() + m_ghost.getTotalLength());

#if VERBOSITY >=2
  ev << "tick on host " << host().name() << ", " << delta << " seconds "
    "since last tick\n";
#endif
  
  // simulate; for the original, non-deterministic and deterministic events
  // will be processed; for the ghost, only deterministic events will be
  // processed
  host().getWorld().simulate(host().now(), host().getLoad());
  m_ghost.simulate(host().now(), host().getLoad());

  // This must be the last.  That isn't entirely logical, but necessary because
  // random effects are caused here and there needs to be a chance to send
  // them off right away before ownership is reevevaluated next tick.
  host().getScenario().ai(host().now());

  // Check for and possibly do a regular update
  bool do_regular= false;
  m_updateTimer += delta;
  if (m_regularUpdates > 0.0)
    while (m_updateTimer >= m_regularUpdates)
      {
      m_updateTimer -= m_regularUpdates;
      do_regular= true;
      }

  if (do_regular)
    {
    // Do a regular update, sending an update for each and every instance
    cWorld::const_iterator iter;
    for (iter = host().getWorld().begin(); iter != host().getWorld().end(); iter++)
      updateInstance((*iter).second->getObjectID());
    }
  else
    {
    // For each instance, check if distance between player and ghost is
    // too large; if it is, send an update message
    cWorld::const_iterator iter;
    for (iter = host().getWorld().begin(); iter != host().getWorld().end(); iter++)
      {
//      cout << "DIST[" << (*iter).second->getObjectID() << "]: " <<
//          ((*iter).second->
//          getDistance(m_ghost.getInstance((*iter).second->getObjectID())))
//          << " / " << m_maxDistance[(*iter).second->getObjectID()] << " "
//          << m_ghost.isReplicated((*iter).second->getObjectID()) << endl;
         
      if ((*iter).second->
          getDistance(m_ghost.getInstance((*iter).second->getObjectID()))
          > m_maxDistance[(*iter).second->getObjectID()])
        updateInstance((*iter).second->getObjectID());
      host().getLoad().addCPU(3);
      }
    }

  m_lastTickTime= host().now();
}

void cLooseConsistency::changeAttribute(OBJECT_ID obj, ATTRIBUTE_ID attr,
        double val, bool)
{
    cInstance &ei= host().getWorld().getInstance(obj);
    cAttributeSet eas(ei.getAttributes());
    if (!eas.hasAttribute(attr))
      return;

    eas.get(attr).setValue(val);
    ei.setAttributes(eas);

    host().getLoad().addCPU(5);
}

void cLooseConsistency::updateInstance(const OBJECT_ID &obj)
{
    // send messages only if owner
    if (!m_isOwner[obj])
	return;

    cInstance &ni= host().getWorld().getInstance(obj);

    scoped_ptr<cMsg> msg(host().getMsgFactory().createMsg("UpdateInstance"));

    msg->setObjectID(ni.getObjectID());
    msg->setObjectName(ni.getName().c_str());

    msg->setAttributes(ni.getAttributes());

    // send message
    host().getRouter().send(msg.get());

    // update ghost
    cInstance &gi(m_ghost.getInstance(obj));
    gi.setAttributes(ni.getAttributes());

    host().getLoad().addCPU(15);
}


void cLooseConsistency::handleMessage(cMsg *msg)
{
  // I can't use the real selfMessage method cause of a hack with batching
  bool isSelfMessage = msg->senderModuleId() == host().id();

  switch (msg->getType()) {
  case cMsg::DATA:
    {
    // remember if this object was seen for the first time
    const bool isNew = !isReplicated(msg->getObjectID());

    if (isNew)
      {
      // still need to replicate this one
      scoped_ptr<cInstance>
        ni(cInstanceFactory::createInstance(msg->getObjectID(),
                                            msg->getObjectName()));
      host().getWorld().addInstance(ni.get());
      scoped_ptr<cInstance>
	ng(cInstanceFactory::createInstance(msg->getObjectID(),
                                            msg->getObjectName()));
      m_ghost.addInstance(ng.get());
      
      if (m_isOwner.find(msg->getObjectID()) == m_isOwner.end())
        setOwner(msg->getObjectID(), false);
      if (m_maxDistance.find(msg->getObjectID()) == m_maxDistance.end())
        setMaxdist(msg->getObjectID(), 0);

      host().getLoad().addCPU(10);
      }

    // refuse updates for objects where I am the owner, UNLESS the object
    // wasn't replicated before or the message comes from myself
    if (isOwner(msg->getObjectID()) && !isSelfMessage && !isNew)
      break;

    cInstance &u(host().getWorld().getInstance(msg->getObjectID()));
    double old_simtime= u.simTime();
    scoped_ptr<cInstance> old_inst(u.clone());

    // ignore out-of-order messages
    if (u.updateTime() > msg->timestamp())
      {
#if VERBOSITY >=1
      ev << "Dropped out of order packet for " << u.getName() << " sent "
         << msg->timestamp() << "."<< endl;
#endif
      break;
      }

    u.setAttributes(msg->getAttributes());
    u.setSimTime(msg->timestamp());
    if (!isSelfMessage || isNew)
      u.setUpdateTime(msg->timestamp());

    host().getLoad().addCPU(6);         // copy cost is 5

    if (msg->timestamp() < old_simtime)
	{
#if VERBOSITY >= 3
    ev << "Updating instance " << u.getName() << " at time " << host().now() <<" from "
       << msg->timestamp() << "to" << old_simtime << "." <<endl;
    { cAttributeSet as= u.getAttributes(); cAttributeSet::const_iterator ci;
      for (ci=as.begin(); ci != as.end(); ci++)
	ev << (*ci).second.getID() << " = " << (*ci).second.getValue() << ", "; }
    ev << endl;
#endif
    u.simulate(old_simtime, host().getLoad());
#if VERBOSITY >= 3
    { cAttributeSet as= u.getAttributes(); cAttributeSet::const_iterator ci;
      for (ci=as.begin(); ci != as.end(); ci++)
	ev << (*ci).second.getID() << " = " << (*ci).second.getValue() << ", "; }
    ev << endl;
#endif
	}

    if (!isNew && !isSelfMessage)
      host().getDisco().computeDisco(*old_inst, u);

    if (!isSelfMessage || isNew)
      {
      cInstance &u2(m_ghost.getInstance(msg->getObjectID()));

      u2.setSimTime(u.simTime());
      u2.setUpdateTime(u.updateTime());
      u2.setAttributes(u.getAttributes());

      host().getLoad().addCPU(5);
      }
    }
    break;
    default:
      ev  << "Unknown message received, type " << msg->getType()
          <<"\n";
      break;
  }
}
