/*
Copyright (C) 2007, 2008  Condel project, TU-Berlin.
Written by Jan Sablatnig, Sven Grottke, Jiehua Chen.

This file is part of Adam, the DVE testbed.

Adam is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3, or (at your option) any later
version.

Adam is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Adam; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.
*/

#ifndef __CONSISTENCY_LOOSE_H
#define __CONSISTENCY_LOOSE_H

#include <map>
#include "gen.hh"
#include "consistency.hh"
#include "message_hist.hh"



namespace Condel {

/// Loose Consistency.
/// This is the consistency algorithm for loose consistency.
///
/// Loose consistency works (roughly) as follows:
/// <ul><li> Whenever a local action is to be performed, do it immediately.
///     <li> At regular intervals, inform all other hosts of the current
///          state of all owned instances.
///     <li> Also, keep a ghost instance for every visible instance.
///          Whenever the distance between the visible and the ghost becomes
///          too large, send an update for that instance.
///          Whenever an update is sent out for an instance, copy the visible
///          to the ghost.
///     <li> When an update arrives for an instance, simply overwrite the
///          local state.</ul>
///
/// There is a bit of science involved to determine which host @em owns which
/// object.
///
/// This kind of consistency is really not consistent at all.  In particular,
/// there is no global truth that should be reached, the aggreed-upon truth
/// hinges closely on the network delays.  Nevertheless, this is a simple
/// and robust approach that allows simple DVEs to be run.
///
/// Read more about the algorithms in TKN-07-005.

class cLooseConsistency: public cConsistency {
protected:
  cWorld m_ghost; // The ghost world, used to see if player/ghost
                        // distance requires an update-message
  
  double m_regularUpdates; // interval between regular updates, in seconds

  double m_updateTimer;  // This time has gone by w/o regular update

  map<const OBJECT_ID, bool> m_isOwner;  // if true, this host owns the object
  map<const OBJECT_ID, double> m_maxDistance;  // maximum tolerated distance
					       // before update is forced

  double m_timeout;   // timeout as specified in ini file

public:
  cLooseConsistency(cHost& host);
  virtual ~cLooseConsistency(void) {}

  virtual void initialize(void);

  /// Do @em not use this call.
  /// This is the inherited call from cConsistency, but you need to use
  /// addInstance(const cInstance &, double) instead.
  virtual void addInstance(const cInstance &inst);

  /// Add an instance with its max-player-ghost distance.
  /// Use this call to add an instance instead of the inherited
  /// addInstance(const cInstance &).
  /// @param inst The new instance.  This will be copied, you retain ownership
  /// of the specified one.
  /// @param maxdist The maximum player/ghost distance before an update will
  /// be forced for this instance.  Unit depends on the scenario.  This value
  /// is compared to cInstance::getDistance(const cInstance &).
  /// @sa setMaxDist()
  virtual void addInstance(const cInstance &inst, double maxdist);

  // inherited from cConsistency
  /// Called every tick.
  /// In loose consistency, this advances (calls cWorld::simulate()) both the
  /// real and the ghost world, then calls the AI.  Finally checks for regular
  /// or indivual updates and sends them if necessary.
  virtual void handleTick(void);

  /// Receive a message.
  /// In loose consistency, if the message is for an object not yet
  /// on this host, it is replicated.  Then, if this host is not the owner,
  /// read the data given in the message and overwrite the corresponding
  /// instances (both visible and ghost).  During this change, discontinuity
  /// is also measured.
  /// This is called from cRouter::handleMessage(cMsg *).
  /// @param msg A pointer to the new message.  The host retains ownership of
  /// the message, do NOT delete it.  If you need your own copy, you'll have to
  /// cMsg::dup() it.
  virtual void handleMessage(cMsg *msg);

  /// Change an (existing) instance's attribute.
  /// In loose consistency, this simply performs the change, but only on
  /// the visible instance.
  /// @param obj The ID of the object whose instance is to be changed.
  /// @param attr The ID of the attribute to change in the instance.
  /// @param val The new value to set.
  /// @param preliminary Is ignored for loose consistency.
  virtual void changeAttribute(OBJECT_ID obj, ATTRIBUTE_ID attr, double val,
                                bool preliminary=false);

  /// Send an update for an object.
  /// Finds the corresponding instance in the visible world, packs it into
  /// a message, sends that message off and copies the visible state onto
  /// the corresponding instance in the ghost world.
  /// @param obj The object ID of the object to send an update about.  If
  /// this host does not own that object, nothing happens.
  virtual void updateInstance(const OBJECT_ID &obj);

  /// Am I the owner of this object?
  /// @param obj The object in question.
  /// @return True if this host is considered to be the owner of this object,
  /// i.e. the one who may send updates for it.
  virtual bool isOwner(const OBJECT_ID& obj)
	{ return m_isOwner[obj]; }
	
  /// Specify ownership of an object.
  /// @param obj The object this information pertains to.  The object must have
  /// been registered globally, but it need not yet have been instantiated
  /// anywhere.
  /// @param nval True if this host is to be the owner of the object, false
  /// otherwise.
  virtual void setOwner(const OBJECT_ID& obj, bool nval=true)
	{ m_isOwner[obj]= nval; }

  /// Reset maximum player/ghost distance.
  /// @param obj The object this information pertains to.  The object must have
  /// been registered globally, but it need not yet have been instantiated
  /// or added here.  Note, however, that if you use the recommended adding
  /// function addInstance(const cInstance &, double), the value given here
  /// will be overwritten.
  /// @param l The maximum player/ghost distance before an update will
  /// be forced for this instance.  Unit depends on the scenario.  This value
  /// is compared to cInstance::getDistance(const cInstance &).
  virtual void setMaxdist(const OBJECT_ID& obj, double l)
	{ m_maxDistance[obj]= l; }
};
  
};


#endif // __CONSISTENCY_LOOSE_H
