/*
Copyright (C) 2007, 2008  Condel project, TU-Berlin.
Written by Jan Sablatnig, Jiehua Chen, Sven Grottke.

This file is part of Adam, the DVE testbed.

Adam is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3, or (at your option) any later
version.

Adam is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Adam; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.
*/
#include "consistency_opt.hh"
#include <omnetpp.h>
#include "instance.hh"
#include "instance_factory.hh"
#include "scenario.hh"
#include "message_router.hh"
#include "host.hh"



cOptConsistency::cOptConsistency(cHost &Host)
: cConsistency(Host), m_history(),
  m_trailing(new double [sizeof(m_pState)/sizeof(scoped_ptr<cOptWorld>)]),
  m_dirty(new bool [sizeof(m_pState)/sizeof(scoped_ptr<cOptWorld>)]),
  m_dirtyVisible(false), m_pState(), m_localDelay(0.0), m_timeout(0.0),
  m_batching(0.0)
{
  char name[100];

  // require at least ONE backtracking state
  assert(sizeof(m_pState)/sizeof(scoped_ptr<cOptWorld>) > 0);
  
  for (unsigned i=0; i< sizeof(m_pState)/sizeof(scoped_ptr<cOptWorld>);i++)
    {
    unsigned long ms= 100UL<<i;
    snprintf(name, sizeof(name), "trailing state -%lums", ms);
    m_trailing[i]= ms/1000.0;
    m_dirty[i]= false;
    m_pState[i].reset(new cOptWorld(host(), name, m_history));
    }
}

void cOptConsistency::initialize(void)
{
  cConsistency::initialize();

  m_localDelay= host().par("local_lag");
  host().getScenario().setDelayHint(m_localDelay);

  m_timeout= host().par("timeout");

  m_batching= host().par("batching");

  m_history.initialize();

  for (unsigned i=0; i< sizeof(m_pState)/sizeof(scoped_ptr<cOptWorld>);i++)
    {
    m_dirty[i]= false;
    m_pState[i]->initialize();
    }

  m_dirtyVisible= false;
}

void cOptConsistency::addInstance(const cInstance &inst)
{
  // Create an insertion-message
  scoped_ptr<cMsg> msg(host().getMsgFactory().createMsg("AddInstance",
                                                       m_timeout, m_batching));

  msg->setObjectID(inst.getObjectID());
  msg->setObjectName(inst.getName().c_str());

  msg->setAttributes(inst.getAttributes());

  // Now, send that message to myself and to the routing layer
  host().getRouter().send(msg.get());

  host().getLoad().addCPU(10);
}

// This doesn't change the world-instance at all, it only sends off a message
void cOptConsistency::changeAttribute(OBJECT_ID obj, ATTRIBUTE_ID attr,
        double val, bool preliminary)
{
  // What I SHOULD do with prelim data is make a preliminary message that
  // is only sent to the local host and then discarded when the actual
  // message with the analogous information arrives.  That's extremely
  // difficult though (e.g. because I don't know how to match info).
  // So I just discard preliminary info, which works well in an optimistic
  // setup (sooner or later the actual information will arrive and then
  // everything will make sense).
  // This isn't quite the same as in the loose case (where preliminary data
  // is actually used in the game in some cases).
  if (preliminary) return;

  // Create an update-message
  scoped_ptr<cMsg> msg(host().getMsgFactory().createMsg("UpdateAttribute",
                                                       m_timeout, m_batching));

  msg->setTimestamp(host().now() +m_localDelay);
  msg->setObjectID(obj);
  msg->setObjectName(host().getWorld().getInstance(obj).getName().c_str());

  cAttribute cat(attr, val);
  cAttributeSet as;
  as.add(cat);

  msg->setAttributes(as);

  // Now, send that message to myself and to the routing layer
  host().getRouter().send(msg.get());

  host().getLoad().addCPU(12);
}

// not needed currently
void cOptConsistency::updateInstance(const OBJECT_ID &obj)
{
  assert(0);
}

void cOptConsistency::update(cOptWorld &rep, double newtime,
        bool &dirty, const cOptWorld *prev, bool measure_disco)
{
  if (dirty)
    {
    scoped_ptr<cWorld> old;
    if (measure_disco)
      // For measurement only, we need to copy the currently viewable world
      old.reset(new cWorld(rep));         // expensive copy

    // Backtrack
    double ost= rep.simTime();
    rep= *prev;
//    cerr << "BACKTRACKING: " << &rep << " @ " << newtime << endl;
    rep.simulate(ost,host().getLoad());
    host().getLoad().addCPU(rep.size());
    dirty= false;

    if (measure_disco)
      {
      // Now do the actual measuring
      cWorld::const_iterator iter;
      for (iter = old->begin(); iter != old->end(); iter++)
        {
        const cInstance &o= *(*iter).second;
        if (!rep.isReplicated(o.getObjectID()))
          continue;
        const cInstance &n= rep.getInstance(o.getObjectID());
        host().getDisco().computeDisco(o, n);
        }
      }
   }
  
  newtime= cHost::roundTime(newtime);
  if (newtime > rep.simTime())
    rep.simulate(newtime, host().getLoad());

  host().getLoad().addCPU(3);
}


void cOptConsistency::handleTick(void)
{
#if VERBOSITY >=2
  ev << "tick on host " << host().name() << " @ " << host().now() << ".\n";
#endif

  unsigned long long mc= host().getWorld().getTotalLength();
  for (unsigned i=0; i< sizeof(m_pState)/sizeof(scoped_ptr<cOptWorld>);i++)
    mc += m_pState[i]->getTotalLength();

  host().getLoad().addMem(mc);

  // Find trailing states that have been outdated, they are marked dirty
  for (int i=sizeof(m_pState)/sizeof(scoped_ptr<cOptWorld>) -1; i>=0;i--)
    {
    // Does this TS exist?
    if (!m_pState[i]) return;

    update(*m_pState[i], host().now()-m_trailing[i],
                m_dirty[i], m_pState[i+1].get());
    }

  cOptWorld &visible= *(cOptWorld *) &(host().getWorld());

  update(visible, host().now(), m_dirtyVisible, m_pState[0].get(), true);

  // Clean out some way-too-old messages from the message history
  // Problem:  * Each world keeps an iterator into the message history.
  //   * If the referenced event is deleted, I'll get a SEGV later.
  //   * If events are very sparse (like only 1 event), the newer worlds
  //     might still point to a really old event as last event processed.
  //   * Therefore, I SHOULD check each world's iterator before I discard.
  // I'm not doing that, I'm only checking the last one (usually sufficient).
  m_history.discard(host().now() - m_trailing[TRAILING_STATES -1] -cHost::TICK_TIME,
     m_pState[TRAILING_STATES -1]->current());
  
  // This must be last.  That isn't entirely logical, but it's the same
  // as in loose consistency...
  host().getScenario().ai(host().now());
}

void cOptConsistency::handleMessage(cMsg *msg)
{
  // Add the message to the history
  cMessageHistory::iterator n= m_history.add(*msg);
  if (n == m_history.end())
    return;

#if VERBOSE >=3
  cerr << "MSG: " << host().simTime() << " @ " << msg->timestamp() << " from "
       << msg->getHostID() << " to " << host().id() << endl;
#endif

  bool is_dirty= false;

  // Find trailing states that are now outdated
  for (int i=sizeof(m_pState)/sizeof(scoped_ptr<cOptWorld>) -1; i>=0;i--)
    {
    if (is_dirty || msg->timestamp() < m_pState[i]->simTime() ||
        *msg < *(m_pState[i]->current()))
      {
      if (i==sizeof(m_pState)/sizeof(scoped_ptr<cOptWorld>) -1)
        {
#if VERBOSE >=0
        cerr << "Received message beyond simulation horizon. Skipping it." <<endl;
#endif
        // Would now need to do a network split and rejoin.
        // Perhaps exit(), at least?
        host().split(msg->getHostID());
        return;
        }
      m_dirty[i]= true;
      is_dirty= true;
      }
    host().getLoad().addCPU(2);
    }

  cOptWorld &visible= *(cOptWorld *) &(host().getWorld());

  // backtracking for current.
  if (is_dirty || msg->timestamp() < visible.simTime() ||
      *msg < *(visible.current()))
    m_dirtyVisible= true;

  host().getLoad().addCPU(2);
}
