/*
Copyright (C) 2007, 2008  Condel project, TU-Berlin.
Written by Jan Sablatnig, Sven Grottke, Jiehua Chen.

This file is part of Adam, the DVE testbed.

Adam is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3, or (at your option) any later
version.

Adam is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Adam; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.
*/

#ifndef __CONSISTENCY_OPT_H
#define __CONSISTENCY_OPT_H

#include <map>
#include "gen.hh"
#include "consistency.hh"
#include "message_hist.hh"
#include "world_opt.hh"

/// @file consistency_opt.hh
/// This file contains the description of cOptConsistency.

/// Sets the # of trailing states to use.
/// This could easily be increased or decreased.  If everything works
/// correctly, this should @em not change anything, except for the
/// complexity.  If the # of TSSes is too low, backtracks will waste
/// too much calculation time.  If the # of TSSes is too high, idle
/// calculation of all those engines will waste too much calculation time.
#define TRAILING_STATES 8

namespace Condel {

/// Optimistic Consistency.
/// This is the consistency algorithm for optimistic consistency.
///
/// Optimistic consistency works (roughly) as follows:
/// <ul><li> Whenever a local action is to be performed, do it immediately.
///     <li> Send the @em action to all other hosts.
///     <li> When an action is received from another host, backtrack to
///          a time just before that, add it to the message history and
///          recap everything to the present time.</ul>
///
/// For this to practicable, messages need to be transferred in a secure
/// way (using a selective repeat in our case).
///
/// While this kind of consistency behaves just like loose consistency at
/// first, the conflict resolution is much smarter and leads to an eventual
/// consistency (i.e. past states are consistent) of all hosts with a
/// global history.  In addition, the global history would be logical even
/// if this was a single-user system.  Note that the price for these advantages
/// is significantly higher CPU load and the tendency to break if any unforseen
/// errors occur (like different rounding on different machines etc.).
///
/// Read more about the algorithms in TKN-07-005.

class cOptConsistency: public cConsistency
{
protected:
  cMessageHistory m_history;      // The event queue
  scoped_array<double> m_trailing;
  scoped_array<bool> m_dirty;   // true if rollback is necessary
  bool m_dirtyVisible;          // same, for the visible world
  scoped_ptr<cOptWorld> m_pState[TRAILING_STATES];  // States from the past
  double m_localDelay;          // how long before an action is shown locally

  double m_timeout;             // timeout as specified in ini-file
  double m_batching;            // batching as specified in ini-file

  // Updates rep to newtime, perhaps rolling back to prev if dirty.
  // If measure_disco is given, will also use the host's DiscoMeter to
  // measure effect of rollback
  void update(cOptWorld &rep, double newtime,
        bool &dirty, const cOptWorld *prev, bool measure_disco=false);
public:
  cOptConsistency(cHost& host);
  virtual ~cOptConsistency(void) {}

  virtual void initialize(void);

  /// Gets you the message history.
  /// @return A message/event list that defines the entire visible world.
  /// This is the only one this host is using.  Note that all TSS'es are using
  /// that same queue.
  cMessageHistory &getMessageHistory(void) { return m_history; }

  /// Add an instance.
  /// The instance is not actually added here, rather an AddInstance
  /// message is generated which is then sent to all hosts and myself.
  /// When that message is evaluated, the instance is added.
  /// @param inst A new instance.  The values from this are copied, you
  /// retain ownership of it.
  virtual void addInstance(const cInstance &inst);

  /// Do scheduled work for each tick.
  /// In optimistic consistency, this means backtracking/updating the TSSes,
  /// cropping the message history, and finally calling the scenario's AI.
  virtual void handleTick(void);

  /// Receive a new message from the network.
  /// This is called from cRouter::handleMessage(cMsg *).
  /// The message is simply added to the message history.  Some of the trailing
  /// states may be marked as dirty, indicating they need to backtrack.
  /// @param msg A pointer to the new message.  The host retains ownership of
  /// the message, do NOT delete it.  If you need your own copy, you'll have to
  /// cMsg::dup() it.
  virtual void handleMessage(cMsg *msg);

  /// Update an instance.
  /// Has no effect for this consistency.
  /// @param obj The object whose instance's state is to be sent out to other
  /// hosts.
  virtual void updateInstance(const OBJECT_ID &obj);

  /// Change an attribute on an instance.
  /// This is assumed to be a nondeterministic event, so this entire effect
  /// is wrapped into a message, sent to all hosts and myself.
  /// @param obj The ID of the object whose instances are to be changed.
  /// @param attr The ID of the attribute to change in the instances.
  /// @param val The new value to set.
  /// @param preliminary If preliminary is true, this call returns immediately
  /// for this consistency.
  virtual void changeAttribute(OBJECT_ID obj, ATTRIBUTE_ID attr, double val,
                                bool preliminary=false);
};
  
};


#endif // __CONSISTENCY_OPT_H
