#!/bin/sh
# Copyright (C) 2007, 2008  Condel project, TU-Berlin.
# Written by Sven Grottke, Jan Sablatnig, Jiehua Chen.
# 
# This file is part of Adam, the DVE testbed.
# 
# Adam is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3, or (at your option) any later
# version.
# 
# Adam is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
# 
# You should have received a copy of the GNU General Public License
# along with Adam; see the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.
# 
# Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.


# please note that by themselves, NONE of these functions are thread-safe; if
# you want to use them in a parallel environment, use the semaphores provided
# by gen.sh; the only execption as ak_run(), which uses semaphores to protect
# the host pool

# setup essential variables
# args: -
# sets AK_BASE, AK_MASTER, AK_SLAVE, AK_RUN, AK_STAT, AK_HOSTS_FILE,
# AK_TMP_BASE, AK_SLAVE_SEM; these are all exported
ak_setup_vars() {
    AK_BASE="${CONDEL_ROOT}/akaroa/install"
    AK_MASTER="${AK_BASE}/bin/akmaster"
    AK_SLAVE="${AK_BASE}/bin/akslave"
    AK_RUN="${AK_BASE}/bin/akrun"
    AK_STAT="${AK_BASE}/bin/akstat"
    AK_HOSTS_FILE="${CONDEL_ROOT}/experiments/common_files/akhosts"
    AK_TMP_BASE="/home/condel_store"
    AK_SLAVE_SEM="${CONDEL_ROOT}/experiments/common_files/slave_sem"

    export AK_BASE AK_MASTER AK_SLAVE AK_RUN AK_STAT AK_HOSTS_FILE AK_TMP_BASE AK_SLAVE_SEM
}

# initialize the akaroa engine
# args: -
ak_init() {
    # log it
    log "ak_init() called"

    # setup variables
    ak_setup_vars

    # check for existence of Akaroa
    [ -d "${AK_BASE}" ] || error "Akaroa not installed, aborting distributed simulation run."
    # start master
    ak_start_master

    # start slaves
    #ak_start_slaves    
}

# shutdown the akaroa server and clients; called by the cleanup handler
ak_cleanup() {
    [ "${AK_MASTER_PID}" ] && kill "${AK_MASTER_PID}" && {
        log "akaroa master killed successfully"
    } || {
        log "could not shutdown akaroa"
    }
    [ -e "${AK_SLAVE_SEM}" -a -d "${AK_SLAVE_SEM}" ] && {
        sem_destroy "${AK_SLAVE_SEM}" || error "could not destroy ${AK_SLAVE_SEM}"
        log "${AK_SLAVE_SEM} destroyed successfully"
    } || {
        log "WARNING: ${AK_SLAVE_SEM} does not seem to exists anymore"
    }
}

# start the Akaroa master
# sets AK_MASTER_PID and AK_MASTER_HOST
# sets AK_MASTER_RUNNING to "yes" in case of success
# these are all exported
# also sets AK_BOSS to "yes"; this one is not exported and allows the
# script level that actually started Akaroa to commence a shutdown
ak_start_master() {
    # do we have a working akaroa?
    [ -f "${AK_MASTER}" -a -x "${AK_MASTER}" ] || {
        error "${AK_MASTER} does not exist or is not executable"
    }
    
    # if we started a master before, do nothing
    [ "${AK_MASTER_RUNNING}" = "yes" ] && {
        [ "${AK_MASTER_PID}" ] || error "AK_MASTER_RUNNING is set, but AK_MASTER_PID is not"
        [ -f "${HOME}/.akmaster" ] || error "AK_MASTER_RUNNING is set, but ~/.akmaster is missing"

        # everything is fine
        log "akmaster already running, fine."
        return
    }

    # check if an unregistered akmaster is running
    [ -f "${HOME}/.akmaster" ] && error "Another akmaster process is already running."

    # start new master; use ssh to avoid wait conflicts
    echo "${AK_MASTER} 1>/dev/null 2>/dev/null &" | ssh -x "${HOST}" "/bin/sh" || error "could not start akmaster ${AK_MASTER} on ${HOST}"
    sleep 1

    [ -f "${HOME}/.akmaster" ] || error "could not start akmaster."
    "${AK_STAT}" 2>&1 1>/dev/null || error "could not start akmaster."
    AK_MASTER_PID=`cat "${HOME}/.akmaster" | cut -f1 -d"@"`
    AK_MASTER_HOST=`cat "${HOME}/.akmaster" | sed -e "s/^[[:digit:]]*@//g" -e "s/:[[:digit:]]*$//g"`
    log "akmaster started on ${AK_MASTER_HOST} with PID ${AK_MASTER_PID}"
    AK_MASTER_RUNNING="yes"

    # register cleanup handler
    add_exit_handler ak_cleanup

    # create AK_SLAVE_SEM
    sem_create "${AK_SLAVE_SEM}" || error "could not create the slave status access semaphore ${AK_SLAVE_SEM}"
    
    export AK_MASTER_PID AK_MASTER_HOST AK_MASTER_RUNNING
}

# get the list of potential hosts from $AK_HOSTS_FILE
get_potential_hosts() {
    grep -v "^[[:space:]]*#" "${AK_HOSTS_FILE}"
}

# get the number of potential hosts on stdout
get_number_of_hosts() {
    get_size `get_potential_hosts`
}

# get list of active slaves as a string on stdout
get_active_slaves() {
    "${AK_STAT}" -H | tail +2 | sed -r -e 's/^[[:space:]]*[[:digit:]]+[[:space:]]+([[:alnum:]]+)[[:space:]]+.*$/\1/'
}

# get list of active, but busy slaves
get_busy_slaves() {
    "${AK_STAT}" -H | tail +2 | grep -v "0$" | sed -r -e 's/^[[:space:]]*[[:digit:]]+[[:space:]]+([[:alnum:]]+)[[:space:]]+.*$/\1/'
}

# get list of active, but idle slaves
get_idle_slaves() {
    "${AK_STAT}" -H | tail +2 | grep "0$" | sed -r -e 's/^[[:space:]]*[[:digit:]]+[[:space:]]+([[:alnum:]]+)[[:space:]]+.*$/\1/'
}

# start Akaroa slaves; this is done in the following way:
# for each host in the list of potential hosts, it is checked whether a
# slave process is already running (using akstat); if not, the program tries
# to start one
ak_start_slaves() {
    # build potential host list, excluding the master host
    akss_host_list=""
    for akss_host in `get_potential_hosts`; do
        [ "${akss_host}" = "${AK_MASTER_HOST}" ] && continue
        akss_host_list="${akss_host_list} ${akss_host}"
    done

    # build list of running slaves
    akss_running_slaves=`get_active_slaves`
    
    # for each host, check if a slave is running, try to start one if not
    for akss_host in ${akss_host_list}; do
        echo "${akss_running_slaves}" | egrep "\b${akss_host}\b" >/dev/null && {
            #log "slave already running."
            continue
        }
        log "starting Akaroa slave on ${akss_host}..."
        ssh -x "${akss_host}" "test -d ${AK_TMP_BASE} -a -w ${AK_TMP_BASE}" || {
            log "no write access possible to temporary base directory ${AK_TMP_BASE} on ${akss_host}"
            continue
        }
        ssh -x "${akss_host}" "/usr/bin/nice -n 10 ${AK_SLAVE} &" || {
            log "could not start akslave on ${akss_host}"
            continue
        }
    done

    # wait
    sleep 2
}

# This is the directory where the locks are looked for which are used in
# ak_cleanup
unset AKC_TMP_RESULTS

# Cleans up the temporary files on the slave-machines.
ak_cleanup_slaves() {
    log "ak_cleanup_slaves called"
    [ ! -d ${AKC_TMP_RESULTS} ] && return

    for akc_host_lock in ${AKC_TMP_RESULTS}/slave.*.lock; do
        akc_tmp_dir=`cat "${akc_host_lock}"` ||
        log "WARNING: could not read ${akc_host_lock}"
        akc_host=`basename "${akc_host_lock}" | sed -e "s/^slave\.\(..*\)\.[[:digit:]]*\.lock$/\1/g"`
        akc_pid=`basename "${akc_host_lock}" | sed -e "s/^slave\...*\.\([[:digit:]]*\)\.lock$/\1/g"`
        
        log "deleting ${akc_tmp_dir} on ${akc_host}"
        ssh -x "${akc_host}" rm -Rf "${akc_tmp_dir}" >/dev/null 2>&1 || :
        
        rm $akc_host_lock >/dev/null 2>&1 || :
    done
    
    unset AKC_TMP_RESULTS
}

# execute the simulation; then, wait till all .sca files have been written
# by the slaves
# arguments: 1) output directory
#            2) name of the Akaroa environment file
#            3) .ini file template (missing path to .vec and .sca)
#            4) number of engines to use
#            rest) command line
# returns 0 on success, and 1 if anything (non-critical) went wrong
ak_run() {
    # wrapper script
    akr_wrapper="${CONDEL_ROOT}/experiments/common_files/slave_wrapper.sh"
    
    # set timeout limit
    akr_timeout=2
    
    # sanity check
    [ $# -ge 4 ] || error "ak_run() expects at least 4 arguments; these were given: $@"

    # store the result path and .ini file stub, then start the simulation
    akr_output_dir="$1"
    shift
    akr_env_file="$1"
    shift
    akr_inifile_stub="$1"
    shift
    akr_num_engines="$1"
    shift
    [ "${akr_num_engines}" -ge 1 ] || error "ak_run() needs at least one engine to run"

    # check number of potential hosts
    [ `get_number_of_hosts` -ge "${akr_num_engines}" ] || error "too few hosts, at least ${akr_num_engines} are required"
    
    # gain exclusive access to the slave pool
    sem_lock "${AK_SLAVE_SEM}"

    # restart dead slaves
    ak_start_slaves

    # wait until enough hosts are available
    akr_slave_restart_interval=300
    akr_slave_restart_timer=0
    akr_wait_interval=5
    until akr_idle_slaves=`get_idle_slaves`; [ `get_size ${akr_idle_slaves}` -ge ${akr_num_engines} ]; do
        log "waiting for enough slaves, restart timer at ${akr_slave_restart_timer}"
        [ "${akr_slave_restart_timer}" -ge "${akr_slave_restart_interval}" ] && {
            # restart dead slaves
            ak_start_slaves
            akr_slave_restart_timer=$((${akr_slave_restart_timer} - ${akr_slave_restart_interval}))
        }
        sleep "${akr_wait_interval}"
        akr_slave_restart_timer=$((${akr_slave_restart_timer} + ${akr_wait_interval}))
    done
    
    akr_failure_list=""
    AKC_TMP_RESULTS=$akr_output_dir
    add_exit_handler ak_cleanup_slaves
    log "starting simulation"
    "${AK_RUN}" -n "${akr_num_engines}" -f "${akr_env_file}" -- "${akr_wrapper}" "${akr_output_dir}" "${AK_TMP_BASE}" "${akr_inifile_stub}" "$@" 1>/dev/null 2>&1 &
    akr_worker_pid=$!
    # TODO: do this in a smarter way; but how?
    sleep 1
    sem_unlock "${AK_SLAVE_SEM}"
    wait "${akr_worker_pid}"
    # TODO: check simulation results
    log "simulation finished"

    # find out which hosts were engaged in the simulation
    log "fetching results from participating hosts"
    akr_lock_files=`glob "${akr_output_dir}"/slave.*.lock`
    for akr_host_lock in ${akr_output_dir}/slave.*.lock; do
        # parse lock file to determine host name and temp directory
        akr_tmp_dir=`cat "${akr_host_lock}"` || {
            log "WARNING: could not read ${akr_host_lock}"
            akr_failure_list="${akr_failure_list} '${akr_host_lock} does not exist"
            continue
        }
        akr_host=`basename "${akr_host_lock}" | sed -e "s/^slave\.\(..*\)\.[[:digit:]]*\.lock$/\1/g"`
        akr_pid=`basename "${akr_host_lock}" | sed -e "s/^slave\...*\.\([[:digit:]]*\)\.lock$/\1/g"`
        # log "host ${akr_host} participated, pid is ${akr_pid}, temporary directory on slave is ${akr_tmp_dir}"

        # copy experimental data to actual result directory
#        scp "${akr_host}:${akr_tmp_dir}/omnetpp.vec*" "${akr_output_dir}" 2>&1 1>/dev/null || {
#            log "WARNING: could not copy .vec file from ${akr_host}"
#            akr_failure_list="${akr_failure_list} 'no .vec on ${akr_host}'"
#        }
        scp "${akr_host}:${akr_tmp_dir}/omnetpp.sca*" "${akr_output_dir}" 2>&1 1>/dev/null || {
            log "WARNING: could not copy .sca file from ${akr_host}"
            akr_failure_list="${akr_failure_list} 'no .sca on ${akr_host}'"
        }
        
        # cleanup: delete temporary directory on host
        #log "deleting ${akr_tmp_dir} on ${akr_host}"
        ssh -x "${akr_host}" rm -Rf "${akr_tmp_dir}" >/dev/null 2>&1 || error "failed to delete ${akr_tmp_dir} on ${akr_host}"
        rm $akr_host_lock >/dev/null 2>&1 || :
    done
    remove_exit_handler ak_cleanup_slaves
    unset AKC_TMP_RESULTS
    [ -z "${akr_failure_list}" ] || {
        log "WARNING: collecting the results failed for the following reasons: ${akr_failure_list}"
        return 1
    }
    
    # wait for the system to settle down; for this purpose, a simple heuristic
    # is used: if there are at least ${akr_num_engines} result files, and they
    # are all of the same length, then we are finished; otherwise, wait one
    # second and try again; after $akr_timeout seconds, time out
    log "simulation finished, waiting for the system to settle down..."
    akr_count=0
    while [ "${akr_count}" -lt "${akr_timeout}" ]; do
        # wait and count
        sleep 1
        akr_count=`echo "${akr_count} 1 + p q" | dc`

        # check number of files
        akr_sca_files=`glob "${akr_output_dir}/omnetpp.sca*"`;
        [ "${akr_num_engines}" = `get_size ${akr_sca_files}` ] || continue

        # check file lengths; TODO: this is buggy for akr_num_engines = 1
        # because of the wc output format
        akr_tmp=`get_nth 1 ${akr_sca_files}`
        akr_tmp=`wc -l "${akr_tmp}" | cut -f 1 -d" "`
        akr_tmp=`echo "${akr_tmp} ${akr_num_engines} * p q" | dc`
        akr_total_length=`wc -l ${akr_sca_files} | tail -1 | cut -f 3 -d" "`
        [ "${akr_tmp}" = "${akr_total_length}" ] || continue
        
        # all fine -> done
        break
    done

    # timeout?
    [ "${akr_count}" -lt "${akr_timeout}" ] || {
        log "after ${akr_count} seconds, at least one result file is still missing or incomplete; this is the content of the directory:"
        ls -l "${akr_output_dir}"
        log "wrong number of result files, expecting ${akr_num_engines} .sca files, found " `get_size ${akr_sca_files}`
        
        return 1
    }
    #log "cooldown took ${akr_count} seconds."

    return 0
}
