#!/bin/sh
# Copyright (C) 2007, 2008  Condel project, TU-Berlin.
# Written by Sven Grottke, Jan Sablatnig, Jiehua Chen.
# 
# This file is part of Adam, the DVE testbed.
# 
# Adam is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3, or (at your option) any later
# version.
# 
# Adam is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
# 
# You should have received a copy of the GNU General Public License
# along with Adam; see the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.
# 
# Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.

# run an experiment with a given parameter set using Akaroa
# usage: distrib_run.sh base_dir sim_time parm_set output_file

# try to set CONDEL_ROOT
[ -z "${CONDEL_ROOT}" ] && {
    pushd ../.. 2>&1 1>/dev/null
    CONDEL_ROOT=`pwd`
    export CONDEL_ROOT
    popd 2>&1 1>/dev/null
}

# include common functions
source "${CONDEL_ROOT}/experiments/common_files/gen.sh"
source "${CONDEL_ROOT}/experiments/common_files/akaroa.sh"


unset TMP_RESULTS
# cleanup, will be called automagically on exit (signal 0)
dr_cleanup() {
    # do not remove the temporary files if an error occured
    [ "${sim_err_code}" ] && return
    
    # remove config files
    [ -f "${INIFILE_STUB}" ] && rm "${INIFILE_STUB}"
    [ -f "${NEDFILE}" ] && rm "${NEDFILE}"
    [ -f "${AKENV_FILE}" ] && rm "${AKENV_FILE}"

    # remove temporary results
    [ "${TMP_RESULTS}" -a -d "${TMP_RESULTS}" -a -w "${TMP_RESULTS}" ] && {
        rm -Rf "${TMP_RESULTS}" || log "could not remove ${TMP_RESULTS}"
    }
}
add_exit_handler dr_cleanup



# sanity check
[ $# = 4 ] || {
    error "usage: distrib_run.sh base_dir sim_time parm_set output_file"
}
    
# set some paths and variables
EXP_BASE="$1"
BINARY="${EXP_BASE}/bin/simulation.sh"
COMPUTE_AVG="${CONDEL_ROOT}/experiments/common_files/compute_avg_results.pl"
CONF_EXP="${EXP_BASE}/scripts/conf_exp.sh"
RESULTS="${EXP_BASE}/results"
RESULT_FILE="$4"
INIFILE_STUB=`mktemp "${EXP_BASE}/omnetpp_dyna.ini.XXXXXX"` || error "could not create temporary .ini file"
NEDFILE=`mktemp "${EXP_BASE}/omnetpp_dyna.ned.XXXXXX"` || error "could not create temporary .ned file"

# setup akaroa; number of hosts must not be 1 or ak_run() will fail :-(
# (the correctness checker is buggy)
log "starting akaroa"
ak_init
log "done."

# create NED file
"${CONF_EXP}" --create-ned $3 > "${NEDFILE}" || error "could not create NED file ${NEDFILE}"

# set Akaroa parameters
# set a longer transient phase, and disable the termination condition
# (hopefully) 
AKENV_FILE=`mktemp "${EXP_BASE}/akaroa.env.XXXXXX"` || error "could not create Akaroa environment file"
echo "MaxTransientObs = 10000000"               >  "${AKENV_FILE}"
#echo "MaxSchreubenHeuristicObs = 10000"        >> "${AKENV_FILE}"
echo "Precision = 0.0001"                       >> "${AKENV_FILE}"
echo "Confidence = 0.99"                        >> "${AKENV_FILE}"

# measure execution time
START_TIME=`get_time`

# retry until success (10 times at most)
MAX_ATTEMPTS=1
ATTEMPTS=0
unset TMP_RESULTS
while ATTEMPTS=`echo "${ATTEMPTS} 1 + p q" | dc`; [ "${ATTEMPTS}" -le "${MAX_ATTEMPTS}" ]; do
    
    # delete any old TMP_RESULTS directory if it exists
    [ "${TMP_RESULTS}" -a -e "${TMP_RESULTS}" ] && {
        if [ -w "${TMP_RESULTS}" ]; then
            rm -Rf "${TMP_RESULTS}"
            unset TMP_RESULTS
        else
            error "TMP_RESULTS already exists and is set to ${TMP_RESULTS}, but I cannot remove it."
        fi
    }
    
    # create results directory
    TMP_RESULTS=`mktemp -d "${RESULTS}/tmp.XXXXXXXX"` || {
        error "could not create temporary directory"
    }
    log "collecting temporary results in ${TMP_RESULTS}" 
    ${CONF_EXP} --write-parms $3 > "${TMP_RESULTS}/parm"
    
    # create .ini file
    cat >"${INIFILE_STUB}" <<EOF
# omnetpp init file

include ${CONDEL_ROOT}/omnetpp.ini

[General]
preload-ned-files=${NEDFILE}
fname-append-host=yes
rng-class="cAkaroaRNG"
outputvectormanager-class="cAkOutputVectorManager"
sim-time-limit=$2

EOF
    ${CONF_EXP} --create-ini $3 >>"${INIFILE_STUB}" || error "could not write experiment-specific parameters to ${INIFILE_STUB}"

    # execute
    ak_run "${TMP_RESULTS}" "${AKENV_FILE}" "${INIFILE_STUB}" 7 "${BINARY}" -u Cmdenv || {
        log "ak_run() failed, starting new attempt."
        continue
    }

    # collect results
    log "collecting results..."
    SCA_FILES=`glob "${TMP_RESULTS}/omnetpp.sca*"`
    SCA_MASTER=`get_nth 1 ${SCA_FILES}`
    SCA_OTHERS=`shift_n 1 ${SCA_FILES}`

    combine_sca_files ${SCA_OTHERS} < "${SCA_MASTER}" > "${RESULT_FILE}" || {
        log "combine_sca_files() failed, starting new attempt."
        continue
    }

    # everything is fine, abort
    break
done

# timeout?
[ "${ATTEMPTS}" -le "${MAX_ATTEMPTS}" ] && {
    log "distributed run completed successfully after ${ATTEMPTS} attempts."
} || {
    error "distributed run failed after ${MAX_ATTEMPTS} attempts."
}

# measure execution time, write it to the .sca file
END_TIME=`get_time`
EXECUTION_TIME=`echo "20 k ${END_TIME} ${START_TIME} - p q" | dc`

# calculate avg. value for each measurename in sca
# echo -e "${COMPUTE_AVG} ${RESULT_FILE}\n"
${COMPUTE_AVG} ${RESULT_FILE}
echo -e "scalar \"parm\"\t\"execution time\"\t${EXECUTION_TIME}" >> "${RESULT_FILE}"

# write pseudo module containing the run parameters
${CONF_EXP} --write-parms $3 >> "${RESULT_FILE}" || error "could not write run parameters to ${RESULT_FILE}"

# write pseudo module containing variance and disco weights
VAR_WEIGHTS=`${CONF_EXP} --get-variance-weights` || error "could not get variance weights"
echo -e "scalar \"variance weights\"\t\"${VAR_WEIGHTS}\"\t0" >> "${RESULT_FILE}"
DISCO_WEIGHTS=`${CONF_EXP} --get-disco-weights` || error "could not get disco weights"
echo -e "scalar \"disco weights\"\t\"${DISCO_WEIGHTS}\"\t0" >> "${RESULT_FILE}"

# write pseudo value signalling that the experiment finished successfully
echo -e "scalar \"status\"\t\"success\"\t0" >> "${RESULT_FILE}"
