#!/bin/sh
# Copyright (C) 2007, 2008  Condel project, TU-Berlin.
# Written by Sven Grottke, Jiehua Chen, Jan Sablatnig.
# 
# This file is part of Adam, the DVE testbed.
# 
# Adam is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3, or (at your option) any later
# version.
# 
# Adam is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
# 
# You should have received a copy of the GNU General Public License
# along with Adam; see the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.
# 
# Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.

# must not be called directly, but sourced from one of the actual scripts

# logging to stderr
log() {
    log_date=`date --iso-8601="seconds"`
    echo "${HOST} ${log_date}: $@" >&2
}

# error function
error() {
    log "ERROR: $@"
    log "exiting"

    # set error code; this can be used by scripts that trap the 0 signal to
    # make some decisions at cleanup time
    sim_err_code=1
    
    exit 1
}

# exit handler
cleanup_on_exit() {
    for coe_h in $exit_cleanup_list; do
#        log "exit handler calling $coe_h"
        $coe_h
    done
}

[ "${exit_cleanup_init}" = "yes" ] || {
    exit_cleanup_init="yes"
    exit_cleanup_list=""
    trap cleanup_on_exit 0
}

# new exit handler will be added to the front to the list, so handlers are
# called in LIFO order
add_exit_handler() {
    # already there?
    for aeh_t in $exit_cleanup_list; do
        [ "${aeh_t}" = "$1" ] && {
            log "$1 already in cleanup list, skipping it"
            return 0
        }
    done

    # add it
    exit_cleanup_list="$1 ${exit_cleanup_list}"

    return 0
}

remove_exit_handler() {
    # exactly one argument?
    [ $# = 1 ] || {
        log "remove_exit_handler needs exactly one argument";
        return 1
    }
    
    # is it there?
    echo "${exit_cleanup_list}" | grep "\b$1\b" >/dev/null || {
        log "$1 is not a registered cleanup handler, can't remove it"
        return 1
    }

    # remove it
    exit_cleanup_list=`echo "${exit_cleanup_list}" | sed -e "s/\b$1\b//g"`

    return 0
}

# get_nth
get_nth() {
    shift $1
    echo "$1"
}

# shift the remaining arguments by the value specified by the first argument
shift_n() {
    shift $1
    shift
    echo "$@"
}

# perform filename expansion
glob() {
    set -- "$@"
    echo $*
}

# get number of arguments
get_size() {
    echo $#
}

# get time in second.nanoseconds
get_time() {
    date -u +"%s.%N"
}

# writes a random number between 0 and ($1 - 1) to stdout
random() {
    expr `od -N 4 -l </dev/urandom | head -1 | sed -e 's/.* \([0-9]\+\)$/\1/g'` % "$1"
}

# create a new semaphore
# arg: 1) name of the semaphore; this must be a valid file name
sem_create() {
    [ $# = 1 ] || error "no semaphore specified"
    mkdir -m 700 "$1" 1>/dev/null 2>&1 || return 1
    touch "$1/exists"

    return 0
}

# destroy a semaphore
sem_destroy() {
    [ $# = 1 ] || error "no semaphore specified"
    [ -d "$1" ] || return 1
    [ -f "$1/exists" ] || return 1
    [ -e "$1/lock" ] && {
        log "WARNING: semaphore $1 is still locked"
        rmdir "$1/lock"
    }
    rm -Rf "$1" 1>/dev/null 2>&1 || error "could not remove semaphore $1"

    return 0
}

# try to lock a specific semaphore once; this will NOT block, but return
# with an exit code of 1 if the semaphore is already locked; it relies
# on mkdir being atomic
sem_try_lock() {
    # must not use global vars to remain thread safe!
    [ $# = 1 ] || error "no semaphore specified"
    [ -f "$1/exists" ] || error "$1 is not a semaphore"
    mkdir -m 000 "$1/lock" 1>/dev/null 2>&1 || return 1

    return 0
}
    
# try to lock a specific semaphore; this will block until success
# it relies on mkdir being atomic
sem_lock() {
    # must not use global vars to remain thread safe!
    [ $# = 1 ] || error "no semaphore specified"
    local sl_wait=1
    local sl_max_wait=5
    until sem_try_lock "$1"; do
#        log "waiting $sl_wait seconds"
        sleep "${sl_wait}"
        sl_wait=`expr ${sl_wait} + 1`
        [ "${sl_wait}" -gt "${sl_max_wait}" ] && sl_wait="${sl_max_wait}"
    done

    return 0
}

# unlock a specific semaphore
sem_unlock() {
    # must not use global vars to remain thread safe!
    [ $# = 1 ] || error "no semaphore specified"
    [ -e "$1/lock" -a -d "$1/lock" ] || error "$1/lock does not exist"
    rmdir "$1/lock" || error "could not unlock $1"

    return 0
}

# combine the results from several .sca files; the master file is stdin
combine_sca_files() {
    # scan for header
    while [ "${cr_magic}" != "run" ]; do
        read cr_magic cr_run_id cr_run_name
    done
    [ "${cr_magic}" != "run" ] && error "run header missing from master .sca file"
    echo "run ${cr_run_id} ${cr_run_name}"

    # set number of sources
    cr_num_sources=`echo "$# 1 + p q" | dc`
    
    # average every single value over all the files
    cr_magic="bla"
    while [ "${cr_magic}" ]; do
        # read next value
        read cr_tmp
        cr_magic=`echo "${cr_tmp}" | cut -f 1 | cut -f 1 -d " "`
        [ "${cr_magic}" = "scalar" ] || continue
        cr_module=`echo "${cr_tmp}" | cut -f 1 | cut -f 2 -d " "`
        cr_attr_id=`echo "${cr_tmp}" | cut -f 2`
        cr_value=`echo "${cr_tmp}" | cut -f 3`
        cr_value=`printf "%.20f" "${cr_value}"`
        # quote regexp control characters in cr_attr_id and cr_module
        cr_attr_id_quoted=`echo "${cr_attr_id}" | sed -e 's/\([$*.[\^"]\)/\\\\\1/g'`
        cr_module_quoted=`echo "${cr_module}" | sed -e 's/\([$*.[\^"]\)/\\\\\1/g'`

        # search in all other files
        for cr_file in "$@"; do
            cr_tmp=`grep -e "^scalar ${cr_module_quoted}[[:space:]]*${cr_attr_id_quoted}" "${cr_file}"`
            [ "${cr_tmp}" ] || error "${cr_attr_id} at module ${cr_module} is missing from ${cr_file}"

            cr_tmp_value=`echo "${cr_tmp}" | cut -f 3`
            cr_tmp_value=`printf "%.20f" "${cr_tmp_value}"`
            cr_value=`echo "scale=20; ${cr_value} + ${cr_tmp_value}" | bc` || error "bc failed"
        done

        # divide by number of sources
        cr_value=`echo "scale=20; ${cr_value} / ${cr_num_sources}" | bc` || error "bc failed"

        # write result
        echo -e "scalar ${cr_module}\t${cr_attr_id}\t${cr_value}"
    done
}

