#!/bin/sh
# Copyright (C) 2007, 2008  Condel project, TU-Berlin.
# Written by Sven Grottke, Jan Sablatnig, Jiehua Chen.
# 
# This file is part of Adam, the DVE testbed.
# 
# Adam is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3, or (at your option) any later
# version.
# 
# Adam is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
# 
# You should have received a copy of the GNU General Public License
# along with Adam; see the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.
# 
# Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.

# write the NED file for pong_opt to stdout, gain information about
# run-specific parameters, write the application-specific part of an
# INI file, or write the run parameters into an OMNet++ .sca file
# usage: 1) write a .ned file to stdout using the specified delay parameter
#         conf_exp.sh --create-ned loss delay reception_delay local_lag timeout batching
#        2) write variance weights to stdout
#         conf_exp.sh --get-variance-weights
#        3) write disco weights to stdout
#         conf_exp.sh --get-disco-weights
#        4) write the application-specific portion of an INI file to stdout
#         conf_exp.sh --create-ini loss delay reception_delay local_lag timeout batching
#        5) write the run parameters as a pseudo-module in OMNet++ .sca file
#           syntax
#         conf_exp.sh --write-parms loss delay reception_delay local_lag timeout batching
# will be called by both, single_run.sh and distrib_run.sh


usage() {
    echo "possible usages: conf_exp.sh --create-ned loss delay reception_delay local_lag timeout batching" >&2
    echo "                 conf_exp.sh --get-variance-weights" >&2
    echo "                 conf_exp.sh --get-disco-weights" >&2
    echo "                 conf_exp.sh --create-ini loss delay reception_delay local_lag timeout batching" >&2
    echo "                 conf_exp.sh --write-parms loss delay reception_delay local_lag timeout batching" >&2
}


# set weight strings
VAR_WEIGHTS="cBall::positionX 1 cBall::positionY 1 cBall::velocityX 0.5 cBall::velocityY 0.5 cRacket::positionX 1 cRacket::positionY 1 cRacket::velocityX 1 cRacket::velocityY 0 cScore::score 1"
DISCO_WEIGHTS="cBall::positionX 1 cBall::positionY 1 cBall::velocityX 0.5 cBall::velocityY 0.5 cRacket::positionX 1 cRacket::positionY 1 cRacket::velocityX 1 cRacket::velocityY 0 cScore::score 1"

# check for calls with one parameter
[ $# -ge 1 ] || {
    usage
    exit 1
}
[ "$1" = "--get-variance-weights" ] && {
    echo "${VAR_WEIGHTS}"
    exit 0
}
[ "$1" = "--get-disco-weights" ] && {
    echo "${DISCO_WEIGHTS}"
    exit 0
}

# calls with full parameters
[ $# = 7 ] || {
    usage
    exit 1
}


# create NED file
[ "$1" == "--create-ned" ] && {
    shift
    
    MEAN_DELAY="$2"
    cat <<EOF
// NED file for the Pong experiment using optimistic consistency

simple cPongOptHost
    parameters:
        reception_delay: const,
        local_lag: const,
        timeout: const,
        loss_rate: const,
        batching: const,
	disco_weights: string,
	render: bool;
    gates:
        in: in;
        out: out;
endsimple

simple cEye
    parameters:
        variance_weights: string;
endsimple


module pongOpt
    submodules:
        player_left:  cPongOptHost;
		parameters: disco_weights=
			"$DISCO_WEIGHTS";
                display: "i=device/pc,cyan";

        player_right: cPongOptHost;
		parameters: disco_weights=
			"$DISCO_WEIGHTS";
                display: "i=device/pc,gold";

        eye: cEye;
		parameters: variance_weights=
			"$VAR_WEIGHTS";
                display: "i=block/eyes";

    connections:
        player_left.out --> delay spike ($MEAN_DELAY) --> player_right.in;
        player_left.in <-- delay spike ($MEAN_DELAY) <-- player_right.out;
endmodule

network pong_opt : pongOpt
endnetwork
EOF

    exit 0
}

# create .ini file section
[ "$1" == "--create-ini" ] && {
    shift

    cat <<EOF
[General]
network=pong_opt

[Parameters]
pong_opt.*.reception_delay = $3;
pong_opt.*.local_lag = $4;
pong_opt.*.loss_rate = $1;
pong_opt.*.timeout = $5;
pong_opt.*.render = false;
pong_opt.*.batching = $6;

EOF

    exit 0
}

# write parameters to a .sca file
[ "$1" = "--write-parms" ] && {
    shift

    echo -e "scalar \"parm\"\t\"loss rate\"\t$1"
    echo -e "scalar \"parm\"\t\"avg. delay\"\t$2"
    echo -e "scalar \"parm\"\t\"reception delay\"\t$3"
    echo -e "scalar \"parm\"\t\"local lag\"\t$4"
    echo -e "scalar \"parm\"\t\"timeout\"\t$5"
    echo -e "scalar \"parm\"\t\"batching\"\t$6"
    
    exit 0
}


# no legal parameters
usage
exit 1
