#!/bin/sh
# Copyright (C) 2007, 2008  Condel project, TU-Berlin.
# Written by Jan Sablatnig, Jiehua Chen, Sven Grottke.
# 
# This file is part of Adam, the DVE testbed.
# 
# Adam is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3, or (at your option) any later
# version.
# 
# Adam is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
# 
# You should have received a copy of the GNU General Public License
# along with Adam; see the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.
# 
# Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.

# write the NED file for swarm_loose to stdout, gain information about
# run-specific parameters, write the application-specific part of an
# INI file, or write the run parameters into an OMNet++ .sca file
# usage: 1) write a .ned file to stdout using the specified delay parameter
#         conf_exp.sh --create-ned nr_bees loss delay update queen_dist bee_dist
#        2) write variance weights to stdout
#         conf_exp.sh --get-variance-weights
#        3) write disco weights to stdout
#         conf_exp.sh --get-disco-weights
#        4) write the application-specific portion of an INI file to stdout
#         conf_exp.sh --create-ini nr_bees loss delay update queen_dist bee_dist
#        5) write the run parameters as a pseudo-module in OMNet++ .sca file
#           syntax
#         conf_exp.sh --write-parms nr_bees loss delay update queen_dist bee_dist
# will be called by both, single_run.sh and distrib_run.sh

usage() {
    echo "possible usages: conf_exp.sh --create-ned nr_bees loss delay update queen_dist bee_dist" >&2
    echo "                 conf_exp.sh --get-variance-weights" >&2
    echo "                 conf_exp.sh --get-disco-weights" >&2
    echo "                 conf_exp.sh --create-ini nr_bees loss delay update queen_dist bee_dist" >&2
    echo "                 conf_exp.sh --write-parms nr_bees loss delay update queen_dist bee_dist" >&2
}

# set weight strings
DISCO_WEIGHTS="cQueen::positionX 1 cQueen::positionY 1 cQueen::velocityAbs 0.5 cQueen::velocityPhi 0.0 cQueen::accelerationAbs 0.2 cQueen::accelerationPhi 0.0 cBee::positionX 1 cBee::positionY 1 cBee::velocityAbs 0.5 cBee::velocityPhi 0.0 cBee::accelerationAbs 0.2 cBee::accelerationPhi 0.0 cScore::score 1 cBee::dead 1"
VAR_WEIGHTS="cQueen::positionX 1 cQueen::positionY 1 cQueen::velocityAbs 1 cQueen::velocityPhi 0 cQueen::accelerationAbs 1 cQueen::accelerationPhi 0 cBee::positionX 1 cBee::positionY 1 cBee::velocityAbs 1 cBee::velocityPhi 0 cBee::accelerationAbs 1 cBee::accelerationPhi 0 cScore::score 1 cBee::dead 1"

# check for calls with one parameter
[ $# -ge 1 ] || {
    usage
    exit 1
}
[ "$1" == "--get-variance-weights" ] && {
    echo "${VAR_WEIGHTS}"
    exit 0
}
[ "$1" == "--get-disco-weights" ] && {
    echo "${DISCO_WEIGHTS}"
    exit 0
}

# calls with full parameters
[ $# = 7 ] || {
    usage
    exit 1
}

# get number of bees
NR_BEES=$2
LAST_BEE_NR=`echo "8 k ${NR_BEES} 1 - p q" | dc`

# create NED file
[ "$1" == "--create-ned" ] && {
    shift
    
    MEAN_DELAY="$3"
    cat <<EOF
// NED file for the Swarm experiment using loose consistency

simple cSwarmLooseHost
    parameters:
        regular_updates: const,
        queen_distance: const,
        bee_distance: const,
        timeout: const,
        loss_rate: const,
	disco_weights: string,
        start_queen: bool,
        render: bool;
    gates:
        in: in[];
        out: out[];
endsimple

simple cEye
    parameters:
        variance_weights: string;
endsimple


module swarmLoose
    submodules:
        player: cSwarmLooseHost[${NR_BEES}];
		parameters: disco_weights="$DISCO_WEIGHTS",
                            start_queen = false,
                            render = false;
                            parameters if index <= 0:
                            start_queen = true;
                display: "i=device/pc,azure";

        eye: cEye;
		parameters: variance_weights="$VAR_WEIGHTS";
                display: "i=block/eyes";

    connections:
    for i = 0..${LAST_BEE_NR}, j=i+1..${LAST_BEE_NR} do
        player[i].out++ --> delay spike ($MEAN_DELAY) --> player[j].in++;
        player[i].in++ <-- delay spike ($MEAN_DELAY) <-- player[j].out++;
    endfor
endmodule

network swarm_loose : swarmLoose
endnetwork
EOF

    exit 0
}

# create .ini file section
[ "$1" == "--create-ini" ] && {
    shift

    cat <<EOF
[General]
network=swarm_loose

[Parameters]
swarm_loose.*.regular_updates = $4;
swarm_loose.*.queen_distance = $5;
swarm_loose.*.bee_distance = $6;
swarm_loose.*.timeout = 1.0;
swarm_loose.*.loss_rate = $2;

EOF

    exit 0
}

# write parameters to a .sca file
[ "$1" = "--write-parms" ] && {
    shift

    echo -e "scalar \"parm\"\t\"nr. of bees\"\t$1"
    echo -e "scalar \"parm\"\t\"loss rate\"\t$2"
    echo -e "scalar \"parm\"\t\"avg. delay\"\t$3"
    echo -e "scalar \"parm\"\t\"regular update rate\"\t$4"
#    echo -e "scalar \"parm\"\t\"queen update distance\"\t$5"
    echo -e "scalar \"parm\"\t\"bee update distance\"\t$6"

    exit 0
}


# no legal parameters
usage
exit 1
