/*
Copyright (C) 2007, 2008  Condel project, TU-Berlin.
Written by Jan Sablatnig, Sven Grottke, Jiehua Chen.

This file is part of Adam, the DVE testbed.

Adam is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3, or (at your option) any later
version.

Adam is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Adam; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.
*/
#include "swarm_loose_host.hh"
#include <stdio.h>


/// @file
/// The implementation of the queen ownership passing scheme for loose
/// consistency and Swarm.

/// Switch time.
/// Every SWITCHOVER_TIME seconds, the ownership of the queen transfers to the
/// next host.
#define SWITCHOVER_TIME 2.0


#ifndef DOXYGEN_SKIP

#define TIME_2ND_INIT 0.0001


Define_Module(cSwarmLooseHost);

#endif // DOXYGEN_SKIP

cSwarmLooseHost::cSwarmLooseHost()
: cHost(), m_p2ndInitEvent(NULL), m_hindex(0), m_schid(0), m_hcount(0)
{
  m_factory.reset(new cMsgFactory(*this));

  m_router.reset(new cMessageRouter(*this));

  m_consistency.reset(new cLooseConsistency(*this));

  m_world.reset(new cWorld(*this, "visible world"));

  m_scenario.reset(new cSwarmLooseScenario(*this));
}

void cSwarmLooseHost::initialize(void)
{
  ((cSwarmScenario *) &getScenario())->registerHost(id());

  cHost::initialize();


  // set up a 2nd initialization phase event
  if (m_p2ndInitEvent)
    cancelAndDelete(m_p2ndInitEvent);
  m_p2ndInitEvent= new cMessage("2nd initialization", 2);
  scheduleAt(TIME_2ND_INIT, m_p2ndInitEvent);
}

void cSwarmLooseHost::finish(void)
{
  if (m_p2ndInitEvent)
    {
    cancelAndDelete(m_p2ndInitEvent);
    m_p2ndInitEvent= NULL;
    }
  cHost::finish();
}

void cSwarmLooseHost::handleMessage(cMessage* pMsg)
{
  if (pMsg == m_p2ndInitEvent)
    {
    initialize_2nd();
    cancelAndDelete(m_p2ndInitEvent);
    m_p2ndInitEvent= NULL;
    }

  else
    cHost::handleMessage(pMsg);
}

void cSwarmLooseHost::initialize_2nd(void)
{
  cLooseConsistency &con= *((cLooseConsistency *) &getConsistency());
  map<unsigned long, string> &s_beeName= cSwarmScenario::s_beeName;
  map<unsigned long, string> &s_scoreName= cSwarmScenario::s_scoreName;

  double queen_distance= 0.0;
  if (hasPar("queen_distance"))
    queen_distance= par("queen_distance");
  con.setMaxdist(cSwarmScenario::s_queenID, queen_distance);

  double bee_distance= 0.0;
  if (hasPar("bee_distance"))
    bee_distance= par("bee_distance");

  list<unsigned long>::const_iterator hit;
  for(hit= getRouter().beginHID(); hit!= getRouter().endHID(); hit++)
    if (s_beeName.find(*hit) != s_beeName.end())
      con.setMaxdist(cObjectFactory::nameToID(s_beeName[*hit]), bee_distance);


  // EVIL MAGIC AT WORK:
  // Currently, the ownership over the queen is bounced from host to host in a
  // round-robin-manner.  For this to work, I need to know which of the hosts
  // I AM.
  bool found=false;
  m_hcount=0;
  map<unsigned long, string>::const_iterator bee_it;
  for (m_schid=0, bee_it= s_beeName.begin(); bee_it != s_beeName.end();
        bee_it++, m_hcount++)
    m_schid= (*bee_it).first;

  for (m_hindex=0, bee_it= s_beeName.begin(); bee_it != s_beeName.end();
        m_hindex++, bee_it++)
    {
    if ((*bee_it).first == (unsigned) id())
      {
      found= true;
      break;
      }
    m_schid= (*bee_it).first;
    }

  assert(found);

  // Take control of the score for the host BEFORE me, lose my own.
  con.setOwner(cObjectFactory::nameToID(s_scoreName[m_schid]));
  con.setOwner(cObjectFactory::nameToID(s_scoreName[id()]), false);
}

void cSwarmLooseHost::handleTick(void)
{
  cLooseConsistency &con= *((cLooseConsistency *) &getConsistency());

  cHost::handleTick();

  unsigned long oldowner= lrint(floor(now()/SWITCHOVER_TIME))
                                %m_hcount;
  unsigned long newowner= lrint(floor((now()+TICK_TIME)/SWITCHOVER_TIME))
                                %m_hcount;

  if (newowner == oldowner)
    return;


  if (m_hindex == newowner)
    {
    // assume ownership
#   if VERBOSITY >= 1
    ev << "Time is " << now() << ", " << name() << " taking ownership.\n";
#   endif

    con.setOwner(cSwarmScenario::s_queenID);
    }

  else          // oldowner, or might be start_queen host
    {
    // yield ownership
#   if VERBOSITY >= 1
    ev << "Time is " << now() << ", " << .name() << " yielding ownership.\n";
#   endif

    con.setOwner(cSwarmScenario::s_queenID, false);
    }
}

