/*
Copyright (C) 2007, 2008  Condel project, TU-Berlin.
Written by Jan Sablatnig, Sven Grottke, Jiehua Chen.

This file is part of Adam, the DVE testbed.

Adam is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3, or (at your option) any later
version.

Adam is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Adam; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.
*/

#ifndef __SWARM_LOOSE_HOST_H
#define __SWARM_LOOSE_HOST_H

#include <omnetpp.h>
#include "message_router.hh"
#include "swarm_loose_scenario.hh"
#include "host.hh"

namespace Condel {

/// The host for Swarm with loose consistency.
/// This is the class that creates the entire loose consistency/swarm system.
///
/// It also does quite a few extra things having to do with creating a complete
/// host list.  This host list ist used for:
/// <ul><li> Move the ownership of the score from each host to the next once.
///          Thus, no host owns its own score (fairness!).
///     <li> Move the ownership of the queen in a round-robin manner to the
///          next host every few seconds (SWITCHOVER_TIME).</ul>
/// To be able to construct such a list, all hosts must have already registered,
/// which they do during the initialize() call.  To construct the list, a
/// @em second initialize phase is added.
///
/// Note that the whole list is a cludge, but the addressing @em could be done
/// in a clean way without adding noticable network traffic, we just chose
/// to go the easy route.
  
class cSwarmLooseHost : public cHost
{
private:
  cMessage *m_p2ndInitEvent;
  unsigned m_hindex;            // MY index number in the hosts
  unsigned m_schid;             // IP of the host whose score I'm keeping
  unsigned m_hcount;            // number of hosts

public:
  /// The constructor.
  /// This sets cHost's component pointers to components instantiated here.
  /// The components chosen are cSwarmLooseScenario, cLooseConsistency, etc..
  cSwarmLooseHost(void);
  virtual ~cSwarmLooseHost(void) {}

  /// Reset everything.
  /// This registers the current host in cSwarmScenario::registerHost(),
  /// calls the parent's cHost::initialize(), Then sets up a second
  /// initialization phase.
  /// This member is called by OMNeT when the simulation is about to start.
  virtual void initialize(void);


  virtual void finish(void);

  /// A secondary initialization phase.
  /// This is called right after all hosts have initialized.
  /// This sets the player/ghost distances, then finds out where this host
  /// is in the host list, finally takes control of the prior host's score
  /// and yields its own score.
  virtual void initialize_2nd(void);

  /// Accept a message.
  /// This can either be a network message sent by one of the other simulated
  /// hosts or it could be some tick event.
  /// It is overloaded here so that the 2nd initialization event can be caught.
  /// Other than that, the parent's cHost::handleMessage() is called.
  /// This is called by OMNeT (and ONLY by OMNeT, please).
  /// @param pMsg A pointer to the message.  OMNeT (i.e. the caller of this
  /// function) retains ownership of the message, do NOT delete it.
  /// If you need your own copy, you'll have to dup() it.
  virtual void handleMessage(cMessage* pMsg);

  /// A frame has passed.
  /// Call cHost::handleTick(), then take/yield ownership of the queen in
  /// a round-robin scheme.  This only happens every SWITCHOVER_TIME seconds.
  /// This is called at the end of every frame, which are TICK_TIME seconds
  /// apart.  Called by handleMessage(cMessage *).
  virtual void handleTick(void);
};

};

#endif // __SWARM_LOOSE_HOST_H
