#!/bin/sh
# Copyright (C) 2007, 2008  Condel project, TU-Berlin.
# Written by Jiehua Chen, Sven Grottke, Jan Sablatnig.
# 
# This file is part of Adam, the DVE testbed.
# 
# Adam is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3, or (at your option) any later
# version.
# 
# Adam is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
# 
# You should have received a copy of the GNU General Public License
# along with Adam; see the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.
# 
# Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.

# write the NED file for timeout to stdout, gain information about
# run-specific parameters, write the application-specific part of an
# INI file, or write the run parameters into an OMNet++ .sca file
# usage: 1) write a .ned file to stdout using the specified parameter
#           conf_exp.sh --create-ned delay_fun rtt_expect [arg]
#                       data_expect [arg] loss_data loss_ack timeout             
#        2) write variance weights to stdout
#           conf_exp.sh --get-variance-weights
#        3) write disco weights to stdout
#           conf_exp.sh --get-disco-weights
#        4) write the application-specific portion of an INI file to stdout
#           conf_exp.sh --create-ini delay_fun rtt_expect [arg]
#                       data_expect [arg] loss_data loss_ack timeout 
#        5) write the run parameters as a pseudo-module in OMNet++ .sca file
#           conf_exp.sh --write-parms delay_fun rtt_expect [arg]
#                       data_expect [arg] loss_data loss_ack timeout 
# will be called by both, single_run.sh and distrib_run.sh
# delay_fun can be "step"(no arg), "exp"(arg: minimum) or "pareto"(arg: k)

usage() {
    echo "possible usages: conf_exp.sh --create-ned delay_fun rtt_expect [arg] data_expect [arg] loss_data loss_ack timeout" >&2
    echo "                 conf_exp.sh --get-variance-weights" >&2
    echo "                 conf_exp.sh --get-disco-weights" >&2
    echo "                 conf_exp.sh --create-ini delay_fun rtt_expect [arg] data_expect [arg] loss_data loss_ack timeout" >&2
    echo "                 conf_exp.sh --write-parms delay_fun rtt_expect [arg] data_expect [arg] loss_data loss_ack timeout" >&2
}

# set weight strings
VAR_WEIGHTS=""
DISCO_WEIGHTS=""

# check for calls with one parameter
[ $# -ge 1 ] || {
    usage
    exit 1
}
[ "$1" = "--get-variance-weights" ] && {
    echo "${VAR_WEIGHTS}"
    exit 0
}
[ "$1" = "--get-disco-weights" ] && {
    echo "${DISCO_WEIGHTS}"
    exit 0
}

# calls with full parameters
[ $# = 7 -o $# = 9 ] || {
    usage
    exit 1
}

# parse parameters
MODE=$1
shift

DELAY_FUN=$1
shift

RTT_EXPECT=$1
shift
case $DELAY_FUN in
    "step") ARG_RTT=0;
        PRINT_FUN="$EXPECT";
        DATA_EXPECT=$1;
        shift;;
    "exp") ARG_RTT=$1;
        shift;
        DATA_EXPECT=$1;
        shift;
        ARG=$1;
        PRINT_FUN="exponential_shifted (${DATA_EXPECT}, $ARG)";
        shift;;
    "pareto") ARG_RTT=$1;
        shift;
        DATA_EXPECT=$1;
        shift;
        ARG=$1;
        PRINT_FUN="pareto_extended (${DATA_EXPECT}, $ARG)";
        shift;;
esac

LOSS_DATA=$1
LOSS_ACK=$2
TIMEOUT=$3
    
# create NED file
[ "$MODE" == "--create-ned" ] && {
    
    cat <<EOF
// NED file for the timeout experiment

simple cSender
    parameters:
        loss_ack: const,
        timeout: const;
    gates:
        in: in;
        out: out;
endsimple

simple cReceiver
    parameters:
        loss_data: const;
    gates:
        in: in;
        out: out;
endsimple

module cTimeoutNetwork
    submodules:
        S: cSender;
		parameters: loss_ack=${LOSS_ACK},
                            timeout=$TIMEOUT;
                display: "i=device/pc,gold";

        R: cReceiver;
		parameters: loss_data=${LOSS_DATA};
                display: "i=device/pc,cyan";
      
    connections:
        S.out --> delay ${PRINT_FUN} --> R.in;
        S.in <-- delay ${PRINT_FUN} <-- R.out;
endmodule

network timeoutNetwork : cTimeoutNetwork
endnetwork
EOF
    exit 0
}

# create .ini file section
[ "$MODE" == "--create-ini" ] && {

    cat <<EOF
[General]
network=timeoutNetwork

EOF

    exit 0
}

# write parameters to a .sca file
[ "$MODE" = "--write-parms" ] && {
    shift

    echo -e "scalar \"parm\"\t\"delay_fun: ${DELAY_FUN}\"\t0"
    echo -e "scalar \"parm\"\t\"rtt_delay_expectation\"\t${RTT_EXPECT}"
    echo -e "scalar \"parm\"\t\"data_delay_expectation\"\t${DATA_EXPECT}"    
    echo -e "scalar \"parm\"\t\"arg_rtt\"\t${ARG_RTT}";
    echo -e "scalar \"parm\"\t\"loss_rate_data\"\t${LOSS_DATA}"
    echo -e "scalar \"parm\"\t\"loss_rate_ack\"\t${LOSS_ACK}"
    echo -e "scalar \"parm\"\t\"timeout\"\t$TIMEOUT"
    
    exit 0
}


# no legal parameters
usage
exit 1
