#!/bin/sh
# Copyright (C) 2007, 2008  Condel project, TU-Berlin.
# Written by Jiehua Chen, Sven Grottke, Jan Sablatnig.
# 
# This file is part of Adam, the DVE testbed.
# 
# Adam is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3, or (at your option) any later
# version.
# 
# Adam is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
# 
# You should have received a copy of the GNU General Public License
# along with Adam; see the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.
# 
# Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.

# masterscript for the timeout script(single_timeout.sh) and the analysis script(plot_timeout.pl)
# usage: timeout_master.sh min_timeout max_timeout stepsize delay_fun expectation [arg] loss_data loss_ack out_dir
# delay_fun can be "step"(no arg), "exp"(arg: x_min), "pareto"(arg: k)


# try to set CONDEL_ROOT
[ -z "${CONDEL_ROOT}" ] && {
    pushd ../.. 2>&1 1>/dev/null
    CONDEL_ROOT=`pwd`
    export CONDEL_ROOT
    popd 2>&1 1>/dev/null
}
# include common functions
source "${CONDEL_ROOT}/experiments/common_files/gen.sh"

# sanity check
[ $# = 8 -o $# = 9 ] || {
    error "usage: timeout_master.sh min_timeout max_timeout stepsize delay_fun expectation [arg] loss_data loss_ack output_directory"
}

# set some paths
TIMEOUT_BASE="${CONDEL_ROOT}/experiments/timeout"
TIMEOUT_RESULTS="${TIMEOUT_BASE}/results"
PLOT_BASE="${CONDEL_ROOT}/util/plot"

# set the executable measurement- and plotscript
SINGLE_TIMEOUT="${TIMEOUT_BASE}/scripts/single_timeout.sh"
SCA_FILE="${TIMEOUT_RESULTS}/timeout.sca"
PLOT_TIMEOUT="${PLOT_BASE}/plot_timeout.pl"
FAILED_FILE="${CONDEL_ROOT}/experiments/common_files/measurement_failed.ps"

# set some variables
MIN_TIMEOUT=$1
shift
MAX_TIMEOUT=$1
shift
STEPSIZE=$1
shift
DELAY_FUN=$1
shift
EXPECT=$1
shift
case $DELAY_FUN in
    "step") delay_parm_set="${DELAY_FUN} $EXPECT";;
    "exp") delay_parm_set="${DELAY_FUN} $EXPECT $1"; shift;;
    "pareto") delay_parm_set="${DELAY_FUN} $EXPECT $1"; shift;;
esac

delay_parm_set_fn=`echo "${delay_parm_set}" | sed -e "s/ /_/g"`

LOSS_DATA=$1
LOSS_ACK=$2
PLOT_OUT_BASE=$3

# create output directory
mkdir -p "${PLOT_OUT_BASE}" || {
    error "can not create ${PLOT_OUT_BASE}"
}

sender="timeoutNetwork.S"
receiver="timeoutNetwork.R"

# check if the output directory exists and is writable
[ -d "${PLOT_OUT_BASE}" -a -w "${PLOT_OUT_BASE}" ] || {
    error "${PLOT_OUT_BASE} does not exist or is not writable"
}

# set the filename from results and the ps_file for output
GRAPHIC_N="${PLOT_OUT_BASE}/N_timeout_${MIN_TIMEOUT}_${MAX_TIMEOUT}_${delay_parm_set_fn}_lossData_${LOSS_DATA}_lossAck_${LOSS_ACK}.ps"
GRAPHIC_A="${PLOT_OUT_BASE}/A_timeout_${MIN_TIMEOUT}_${MAX_TIMEOUT}_${delay_parm_set_fn}_lossData_${LOSS_DATA}_lossAck_${LOSS_ACK}.ps"
GRAPHIC_ERRORS="${PLOT_OUT_BASE}/errors_timeout_${MIN_TIMEOUT}_${MAX_TIMEOUT}_${delay_parm_set_fn}_lossData_${LOSS_DATA}_lossAck_${LOSS_ACK}.ps"

# execute
log "starting the measurement script..."
${SINGLE_TIMEOUT} ${MIN_TIMEOUT} ${MAX_TIMEOUT} ${STEPSIZE} ${delay_parm_set} ${LOSS_DATA} ${LOSS_ACK} || {
    # copy an error message postscript file
    for dest in "${GRAPHIC_N}" "${GRAPHIC_A}"; do
        cp "${FAILED_FILE}" "${dest}"
    done

    log "measurement failed, aborting"
    exit 0
}

# copy the sca file under ${PLOT_OUT_BASE}/
cp "${SCA_FILE}" "${PLOT_OUT_BASE}/timeout_${MIN_TIMEOUT}_${MAX_TIMEOUT}_${STEPSIZE}_${delay_parm_set_fn}_lossData_${LOSS_DATA}_lossAck_${LOSS_ACK}.sca"

# N( number of data packets sent )              
log "starting the plot for \"N\" at \"${sender}\""
${PLOT_TIMEOUT} ${SCA_FILE} ${MIN_TIMEOUT} ${MAX_TIMEOUT} ${sender} "N" ${sender} "E_N_PREC" ${sender} "E_N_APPROX" > "${GRAPHIC_N}" || {
    cp "${FAILED_FILE}" "${GRAPHIC_N}"
    log "WARNING: plot_timeout failed with the following parameters: ${PLOT_TIMEOUT} ${SCA_FILE} ${MIN_TIMEOUT} ${MAX_TIMEOUT} ${delay_parm_set} $sender \"N\" $sender \"E_N_PREC\"  ${sender} \"E_N_APPROX\" > ${GRAPHIC_N}"
}

# A( number of acknowledgements sent )              
log "starting the plot for \"A\" at \"${receiver}\""
${PLOT_TIMEOUT} ${SCA_FILE} ${MIN_TIMEOUT} ${MAX_TIMEOUT} ${receiver} "A" ${receiver} "E_A_PREC" ${receiver} "E_A_APPROX" > "${GRAPHIC_A}" || {
    cp "${FAILED_FILE}" "${GRAPHIC_A}"
    log "WARNING: plot_timeout failed with the following parameters: ${PLOT_TIMEOUT} ${SCA_FILE} ${MIN_TIMEOUT} ${MAX_TIMEOUT} ${delay_parm_set} ${receiver} \"A\" ${receiver} \"E_A_PREC\" ${receiver} \"E_A_PREC\" > ${GRAPHIC_A}"
}

# P_delay_error
log "starting the plot for \"P_delay_errors\" at \"${receiver}\""
${PLOT_TIMEOUT} ${SCA_FILE} ${MIN_TIMEOUT} ${MAX_TIMEOUT} ${receiver} "P_delay_error_prec" ${receiver} "P_delay_error_inter" ${receiver} "P_delay_error_hybrid" > "${GRAPHIC_ERRORS}" || {
    cp "${FAILED_FILE}" "${GRAPHIC_ERRORS}"
    log "WARNING: plot_timeout failed with the following parameters: ${PLOT_TIMEOUT} ${SCA_FILE} ${MIN_TIMEOUT} ${MAX_TIMEOUT} ${delay_parm_set} ${receiver} \"P_delay_error_prec\" ${receiver}  \"P_delay_error_inter\" ${receiver}  \"P_delay_error_hybrid\" > ${GRAPHIC_ERRORS}"
}

# remove tmp scafile
rm ${SCA_FILE}
