/*
Copyright (C) 2007, 2008  Condel project, TU-Berlin.
Written by Jiehua Chen, Sven Grottke, Jan Sablatnig.

This file is part of Adam, the DVE testbed.

Adam is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3, or (at your option) any later
version.

Adam is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Adam; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.
*/
#include <stdio.h>
#include <string.h>
#include <omnetpp.h>
#include <assert.h>

#define MAX_STEPS 100000

class cSender: public cSimpleModule{
private:
  double timeout;
  double loss_ack;
  cMessage* timeoutEvent; // holds pointer to the timeout self-message
  int seq;                // packets sequence number per iteration
  int steps;              // iteration number
  cMessage* message;      // message that has to be re-sent on timeout
  int N[MAX_STEPS];       // number of data packets sent 
  
public:
  cSender();
  virtual ~cSender();

protected:
  virtual cMessage* generateNewMessage();
  virtual void sendCopyOf(cMessage *msg);
  virtual void initialize();
  virtual void finish();
  virtual void handleMessage(cMessage *msg);
};

Define_Module(cSender);

cSender::cSender(){
  timeoutEvent = message = NULL;
}

cSender::~cSender(){
  cancelAndDelete(timeoutEvent);
  if( message ) delete message;
}

void cSender::initialize(){
  // Initialize variables.
  seq = 0;
  steps = 0;
  timeout = par("timeout");
  assert( timeout > 0 );
  loss_ack = par("loss_ack");
  timeoutEvent = new cMessage("timeoutEvent");

  // Generate and send initial message.
  ev << "Sending initial message\n";
  message = generateNewMessage();
  sendCopyOf(message);
  scheduleAt(simTime()+timeout, timeoutEvent);
}

void cSender::finish(){
  double sum_packets = 0;
  for( int i = 0; i < MAX_STEPS; i++ ){
    sum_packets += N[i];
  }

  double avg_datas = (double)( sum_packets ) / MAX_STEPS;
//   std::cout << "sum for N: " << sum_packets << std::endl;
//   std::cout << "average packets sent for data: "
//             << (double)( sum_packets ) / MAX_STEPS << std::endl;

  recordScalar("N", avg_datas);
  
  cSimpleModule::finish();
}

void cSender::handleMessage(cMessage *msg){  
  if( msg == timeoutEvent ){
    // If we receive the timeout event, that means the packet hasn't
    // arrived in time and we have to re-send it.
    ev << "Timeout expired, resending message and restarting timer\n";
    sendCopyOf(message);
    scheduleAt(simTime()+timeout, timeoutEvent);
  }else if( uniform(0,1) < loss_ack ){
    ev << "\"Losing\" acknowledgment packet " << msg << endl;
    bubble("acknowledgment packet lost");
    delete msg;
  }else{
    // message arrived
    // Acknowledgement received!
    ev << "Received: " << msg->name() << "\n";

    // check the acknowledgment for the data
    if( msg->priority() == steps ){
   
      // Also delete the stored message and cancel the timeout event.
      ev << "Timer cancelled.\n";
      cancelEvent(timeoutEvent);
      delete message;
      message = NULL;
      
      // Store the data packets send
      // for iteration: steps
      N[steps++] = seq;

      // Ready to send another one.
      if( steps < MAX_STEPS ){
        message = generateNewMessage();
        sendCopyOf(message);
        scheduleAt(simTime()+timeout, timeoutEvent);
      }
    }
    delete msg;
  }
}

cMessage* cSender::generateNewMessage(){
  // Generate a message with a different name every time.
  cMessage *msg = new cMessage("data");
  msg->setTimestamp();
  msg->setPriority(steps);
  seq = 0;
  return msg;
}

void cSender::sendCopyOf(cMessage *msg){
  // Duplicate message and send the copy.
  cMessage *copy = (cMessage *) msg->dup();
  copy->setKind(seq++);
  send(copy, "out");
}


class cReceiver : public cSimpleModule{
private:
  double loss_data;
  simtime_t delays[MAX_STEPS]; // T_delay
  int A[MAX_STEPS];            // number of acknowledgments
  const double t_max;                // maximum time measured
  const double delta_t;              // measurement precision 
  
protected:
  virtual void initialize();
  virtual void finish();
  virtual void handleMessage(cMessage *msg);

public:
  cReceiver();

};

Define_Module(cReceiver);
cReceiver::cReceiver(): cSimpleModule(),
                        t_max(10),
                        delta_t(0.001){}

void cReceiver::initialize(){
  // Initialize variables.
  loss_data = par("loss_data");
  
  // Initialize the delays with "-1"
  // and the number of acknowledgments with "0"
  for( int i = 0; i < MAX_STEPS; i++ ){
    delays[i] = -1;
    A[i] = 0;
  }

}

void cReceiver::finish(){
  long sum_packets = 0;

  // number of messages 
  int W = (int) ceil(t_max / delta_t);
  int nr_messages[W+1]; 
  for( int i = 0; i < W; i++ )
    nr_messages[i] = 0;
  
  for( int k = 0; k < MAX_STEPS; k++ ){
    sum_packets += A[k];
    nr_messages[(int) ceil( delays[k]/delta_t )]++;
  }

  // number of acknowledgements
  double avg_acks = (double)( sum_packets ) / MAX_STEPS;
  recordScalar("A", avg_acks);
  
  for( int i = 0; i < W; i++ ){
    nr_messages[i+1] += nr_messages[i];
    char temp[100];
    sprintf(temp, "T = %lf", i * delta_t);
    recordScalar(temp, ((double) nr_messages[i]) / MAX_STEPS );
  }
  
  //  double avg_delay = (double)( sum_delay ) / MAX_STEPS;
  
  //   std::cout << "sum for A: " << sum_packets << std::endl;
  //   std::cout << "average packets sent for acknowledgment: "
  //             << (double)( sum_packets ) / MAX_STEPS << std::endl;
  
//   std::cout << "sum for delay times: " << sum_delay << std::endl;
//   std::cout << "average delay: "
//             << (double)( sum_delay ) / MAX_STEPS << std::endl;
  
  
  cSimpleModule::finish();
}

void cReceiver::handleMessage(cMessage *msg){
  if( uniform(0,1) < loss_data ){
    ev << "\"Losing\" data packet " << msg << endl;
    bubble("data packet lost");
    delete msg;
  }else{
    ev << msg << " received, sending back an acknowledgement.\n";
    int steps = msg->priority();
    
    if( delays[steps] < 0 )
      delays[steps] = simTime() - msg->timestamp(); 
     
    // increment the number of acknowledgments.
    A[steps]++;

    send(new cMessage("ack", msg->kind(), 0, steps ), "out");
    delete msg;
  }
}
