/*
Copyright (C) 2007, 2008  Condel project, TU-Berlin.
Written by Sven Grottke, Jan Sablatnig, Jiehua Chen.

This file is part of Adam, the DVE testbed.

Adam is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3, or (at your option) any later
version.

Adam is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Adam; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.
*/

#ifndef __ATTRIBUTE_H
#define __ATTRIBUTE_H

#include "gen.hh"
#include "attribute_factory.hh"  // for ATTRIBUTE_ID

namespace Condel {

/// A general named variable class.
/// An attribute is a tuple @f$(n, v), n \in N, v \in V_n @f$ of a name and a
/// value, where @f$ N @f$ is the set of possible names and @f$ V_n @f$ the set
/// of possible values; for our purposes, @f$ V_n @f$ always equals @f$ R @f$
/// (in other words, it's type (double)).

class cAttribute {
private:
  const ATTRIBUTE_ID m_id;
  double m_value;
public:
  /// A constructor.
  /// The value of this attribute will be set to 0.
  /// @param id The created attribute will have this ID.  This must have
  /// been previously registered by
  /// cAttributeFactory::registerAttribute(const string &).
  cAttribute(const ATTRIBUTE_ID& id);
  /// A constructor.
  /// The value of this attribute will be set to 0.
  /// @param name The created attribute will have the ID corresponding to name.
  /// This must have been previously registered by
  /// cAttributeFactory::registerAttribute(const string &).
  cAttribute(const string& name);
  /// A constructor.
  /// @param id The created attribute will have this ID.  This must have
  /// been previously registered by
  /// cAttributeFactory::registerAttribute(const string &).
  /// @param value The value to initalize the attribute with.
  cAttribute(const ATTRIBUTE_ID& id, double value);
  /// A constructor.
  /// The value of this attribute will be set to 0.
  /// @param name The created attribute will have the ID corresponding to name.
  /// This must have been previously registered by
  /// cAttributeFactory::registerAttribute(const string &).
  /// @param value The value to initalize the attribute with.
  cAttribute(const string& name, double value);
  cAttribute(const cAttribute&);
  ~cAttribute(void) {}

  const ATTRIBUTE_ID& getID(void) const { return m_id; }
  double getValue(void) const { return m_value; }
  void setValue(double value) { m_value = value; }
};

};

/// A set of attributes.
/// This collects several attributes, usually all attributes of a given
/// object.
class cAttributeSet : private map<const ATTRIBUTE_ID, cAttribute> {
public:
  cAttributeSet(void);
  virtual ~cAttributeSet(void) {}

  /// Check whether a particular attribute exists.
  /// @param id The idea of the attribute in question.
  /// @return True if the attribute exists within the set.
  bool hasAttribute(const ATTRIBUTE_ID& id) const;

  /// Get an attribute.
  /// @param id The idea of the attribute in question.
  /// @return A const reference to the attribute.
  const cAttribute& get(const ATTRIBUTE_ID& id) const;
  /// Get an attribute.
  /// @param id The idea of the attribute in question.
  /// @return A reference to the attribute.
  cAttribute& get(const ATTRIBUTE_ID& id);

  /// Add a new attribute.
  /// This attribute will be copied.
  /// Fails if an attribute with the same name already exists.
  /// @param attr An attribute.  Ownership does not change.
  void add(const cAttribute& attr);

  /// Find out the size of the information in this set.
  /// @return The # of bytes the whole information in these attributes will
  /// take when coded.
  unsigned long getCodedLength(void) const { return size() * 8; }

  /// An iterator.
  typedef map<const ATTRIBUTE_ID, cAttribute>::const_iterator const_iterator;
  /// An iterator.
  typedef map<const ATTRIBUTE_ID, cAttribute>::iterator iterator;
  /// An iterator.
  typedef map<const ATTRIBUTE_ID, cAttribute>::value_type value_type;

  /// Start an iteration.
  const_iterator begin(void) const;
  /// Start an iteration.
  iterator begin(void);
  /// End of an iteration.
  const_iterator end(void) const;
  /// End of an iteration.
  iterator end(void);

  /// Find the size of this set.
  /// This function is slow, so cache the results if possible.
  /// @return The amount of attributes in this set.
  unsigned long size(void) const
	{ return map<const ATTRIBUTE_ID, cAttribute>::size(); }
};

#endif // __ATTRIBUTE__H
