/*
Copyright (C) 2007, 2008  Condel project, TU-Berlin.
Written by Jan Sablatnig, Sven Grottke, Jiehua Chen.

This file is part of Adam, the DVE testbed.

Adam is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3, or (at your option) any later
version.

Adam is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Adam; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.
*/

#ifndef __CONSISTENCY_H
#define __CONSISTENCY_H

#include <map>
#include "gen.hh"
#include "world.hh"
#include "host.hh"



namespace Condel {

// forward references
class cInstance;
class cMsg;

/// The basic consistency algorithm interface.
/// This is the common base class of all actual consistency schemes.
///
/// Consistency schemes decide on the strategy to update/receive updates
/// on local events/changes to local state.  This can be as simple as
/// loose (simply apply all local changes, send updates to the network,
/// solve conflicts by simple overwriting), or it could be a complicated
/// optimistic approach (apply local changes, send only nondeterministic
/// actions to other hosts, reassemble the same global event history on
/// all hosts, backtrack to synchronize the local state with the global).
  
class cConsistency : public cHostComponent
{
protected:
  
  double m_lastTickTime;	   ///< The host-time of the last tick handled.

  /// Prevent accidental usage of copy constructor.
  /// This is declared, but not defined.
  cConsistency(const cConsistency&);
public:
  cConsistency(cHost& host);
  virtual ~cConsistency(void) {}

  /// Reinitialize the object.
  /// Afterwards, same state as if freshly created.
  virtual void initialize(void);
  /// Object is done.
  /// Perform all pending calls to OMNeT.  Calculate and write final analysis.
  /// When this is called, all other objects (and OMNeT) are still existant.
  /// ~cConsistency() will be called soon, but by then, much of the environment
  /// will be gone.
  virtual void finish(void) {}

  /// Check for local replication.
  /// @param objectID The ID of the object to check for.
  /// @return True if an instance of that object replicated locally.
  bool isReplicated(const OBJECT_ID& objectID) const
        // @todo Currently, all worlds have the same replication list, so this
        // implementation is OK.  This may change.
	{ return host().getWorld().isReplicated(objectID); }
  
  /// Gets a locally replicated instance.
  cInstance& getInstance(const OBJECT_ID& objectID)
	{ return host().getWorld().getInstance(objectID); }
  /// Gets a locally replicated instance.
  const cInstance& getInstance(const OBJECT_ID& objectID) const
	{ return host().getWorld().getInstance(objectID); }
  
  /// Inserts a clone of this instance into the local world.
  /// Handles insertion into the whole world (meaning other hosts) if necessary.
  /// Plus insertion into local world-copies (like ghosts etc.).
  /// @param instance An instance.  This is copied.
  virtual void addInstance(const cInstance& instance) =0;

  /// Change an (existing) instance's attribute.
  /// This may or may not happen immediately, depending on the consistency used.
  /// This may or may not update other hosts, depending on the consistency used.
  /// @param obj The ID of the object whose instances are to be changed.
  /// @param attr The ID of the attribute to change in the instances.
  /// @param val The new value to set.
  /// @param preliminary If preliminary is true, the effect is predicted, but
  ///   the actual information on what actually happened will arrive later from
  ///   another host.
  virtual void changeAttribute(OBJECT_ID obj, ATTRIBUTE_ID attr, double val,
        bool preliminary=false) =0;

  /// Synchronizes an existing instance with other hosts.
  /// The update may or may not happen immediately (depending on the consistency
  /// used).
  /// @param obj The ID of the object whose instances are synchronized.
  virtual void updateInstance(const OBJECT_ID &obj) =0;

  /// Called every tick.
  /// This allow the consistency-layer to make automatic updates.  This in turn
  /// will call simulate on the world or instances.
  /// @sa TICK_TIME
  virtual void handleTick(void) =0;

  /// Receive a message.
  /// This is called from cRouter::handleMessage(cMsg *).
  /// @param pMsg A pointer to the new message.  The host retains ownership of
  /// the message, do NOT delete it.  If you need your own copy, you'll have to
  /// cMsg::dup() it.
  virtual void handleMessage(cMsg *pMsg) =0;
};
  
};


#endif // __CONSISTENCY_H
