/*
Copyright (C) 2007, 2008  Condel project, TU-Berlin.
Written by Jan Sablatnig, Sven Grottke, Jiehua Chen.

This file is part of Adam, the DVE testbed.

Adam is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3, or (at your option) any later
version.

Adam is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Adam; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.
*/
#include "disco_meter.hh"
#include "consistency.hh"
#include "instance.hh"
#include "scenario.hh"


// Any number absolutely less than this is considered 0.0
#define PRECISION 0.000001

cDiscoMeter::cDiscoMeter()
: m_attributeWeight(), m_attributeSum(), m_attributeCount(), m_pAttributeNorm(NULL), m_count(0), m_sum(0.0)
{ }

void cDiscoMeter::addAttribute(const ATTRIBUTE_ID& attribute, double weight)
{
#if DEBUG >= 1 
  // Make sure the attribute wasn't added before
  if (m_attributeWeight.find(attribute) != m_attributeWeight.end())
    condel_error("attribute weight registered multiple times");
#endif

  m_attributeWeight[attribute]= weight;
}

void cDiscoMeter::initWeights(void)
{
  map<const ATTRIBUTE_ID, double>::iterator w;
  double sum= 0.0;

  for (w= m_attributeWeight.begin(); w != m_attributeWeight.end(); w++)
    sum += (*w).second;
  for (w= m_attributeWeight.begin(); w != m_attributeWeight.end(); w++)
    (*w).second /= sum;
}

void cDiscoMeter::init(map<const ATTRIBUTE_ID, double> &attributeNorm)
{
  m_pAttributeNorm= &attributeNorm;

  m_count=0;
  m_sum=0.0;

  m_attributeWeight.erase(m_attributeWeight.begin(), m_attributeWeight.end());

  m_attributeSum.erase(m_attributeSum.begin(), m_attributeSum.end());

  m_attributeCount.erase(m_attributeCount.begin(), m_attributeCount.end());
}

void cDiscoMeter::computeDisco(const cInstance &old, const cInstance &updated)
{
#if DEBUG >= 1
  assert (old.getObjectID() == updated.getObjectID());
#endif

  const cAttributeSet os(old.getAttributes());
  const cAttributeSet ns(updated.getAttributes());

  cAttributeSet::const_iterator o;
  cAttributeSet::const_iterator n;
  for (o= os.begin(), n=ns.begin(); o != os.end(); o++, n++)
    {
    const cAttribute &oa= (*o).second;
    const cAttribute &na= (*n).second;
    const ATTRIBUTE_ID aid= oa.getID();
    double v= oa.getValue() - na.getValue();
    if (fabs(v) < PRECISION)
      continue;
    
#if DEBUG >=2
    assert (m_pAttributeNorm->find(aid) != m_pAttributeNorm->end());
#endif

    double n= (*m_pAttributeNorm->find(aid)).second;
    v *= n;

    if (fabs(v) < PRECISION)
      continue;

#if DEBUG >=2
    assert (m_attributeWeight.find(aid) != m_attributeWeight.end());
#endif

#if VERBOSE >= 3
    cerr << "DISCO: " << aid << " -> " << v << endl;
#endif
    if (isnan(v))
        {
        cerr << "DISCO: " << aid << " -> " << v <<
                ", objectID: " << old.getObjectID() << endl;
        }

    double w= (*m_attributeWeight.find(aid)).second;
    double v2= v* v;
	
    m_sum += w * v2;
    m_count++;

    if (m_attributeSum.find(aid) == m_attributeSum.end())
        {
        m_attributeSum[aid] = 0.0;
        m_attributeCount[aid] = 0;
        }
    (* m_attributeSum.find(aid)).second += v2;
    (* m_attributeCount.find(aid)).second ++;
    }
}


double cDiscoMeter::averageDisco(void) const
{
  return m_sum;
}

double cDiscoMeter::averageDisco(ATTRIBUTE_ID aid) const
{
  map<const ATTRIBUTE_ID, double>::const_iterator it(m_attributeSum.find(aid));

  if (it == m_attributeSum.end())
    return 0.0;

  return (*it).second;
}

double cDiscoMeter::countDisco(ATTRIBUTE_ID aid) const
{
  map<const ATTRIBUTE_ID, unsigned long>::const_iterator it(m_attributeCount.find(aid));

  if (it == m_attributeCount.end())
    return 0;

  return (*it).second;
}
