/*
Copyright (C) 2007, 2008  Condel project, TU-Berlin.
Written by Jan Sablatnig, Sven Grottke, Jiehua Chen.

This file is part of Adam, the DVE testbed.

Adam is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3, or (at your option) any later
version.

Adam is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Adam; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.
*/

#ifndef __DISCO_METER_H
#define __DISCO_METER_H

#include <list>

#include "gen.hh"
#include "attribute_factory.hh"

namespace Condel {

class cHost;
class cInstance;
  
/// Computes the discontinuity measure G.
/// This tool is activated whenever a message arrives that updates an
/// instance in the visible world.  It calculates the size of the change,
/// squares it, counts number and average sizes of such
/// changes internally, supplying them to cHost for final analysis.
///
/// Changes can affect position as well as speeds, even angles
/// and accelerations.  Since these are perceived differently by human
/// observers, each attribute needs to be weighted.
/// @sa cVarianceMeter

class cDiscoMeter
{
private:
  map<const ATTRIBUTE_ID, double> m_attributeWeight;  // list of attributes monitored
  map<const ATTRIBUTE_ID, double> m_attributeSum;
  map<const ATTRIBUTE_ID, unsigned long> m_attributeCount;
  map<const ATTRIBUTE_ID, double> *m_pAttributeNorm;

  unsigned long m_count;
  double m_sum;

public:
  cDiscoMeter(void);
  virtual ~cDiscoMeter(void) {}

  /// Add a new attribute to the watchlist.
  /// @param attribute The ID of the new attribute.
  /// @param weight The relative importance of the attribute to the overall
  ///   perception of disontinuity.  In definitionen.pdf, this would be
  ///   @f$ w_a' @f$.
  void addAttribute(const ATTRIBUTE_ID& attribute, double weight);

  /// Run this once after all attributes have been added.
  /// Only call this ONCE.  This fixes the weights to have
  /// @f$ \sum w_a' == 1 @f$.
  void initWeights(void);

  /// Tally a discontinuous change.
  /// When an instance is changed by an external message, this member is
  /// called to calculate the discontinuity that arises (the possibly
  /// visible jump or teleportation).   All attributes are looked at,
  /// each weighted with the weights set in
  /// addAttribute(const ATTRIBUTE_ID &, double).
  /// @param old The old version of the instance (before the change).
  /// @param updated The new version of the instance (after the change).
  void computeDisco(const cInstance &old, const cInstance &updated);

  /// Reinitialize this tool.
  /// After this call, the tool looks like right after its creation.
  /// @param attributeNorm A list of attribute NORMs to use for the attributes.
  ///   This list is NOT copied, the original version specified here is used.
  ///   The norms are independant from the weights, they are set by the
  ///   scenario so that by multiplying it with its attribute, the attribute
  ///   will have a natural domain of +/- 1.  This does not mean the product
  ///   cannot exceed this range, btw.  An example would be the x position,
  ///   where the weight is 1/width.  Another example is score, where weight
  ///   is 1.
  void init(map<const ATTRIBUTE_ID, double> &attributeNorm);

  /// Compute the average so far.
  /// @return The current average discontinuity.  This includes all changes
  /// until now, in all attributes, all weighted and added.
  double averageDisco(void) const;

  /// Count the changes so far.
  /// @return The # of discontinuous changes that have occurred until now.
  /// This is more or less just the # of calls to
  /// computeDisco(const cInstance &, const cInstance), except that if there
  /// appears to be no discernable difference between the instances, the
  /// call is not counted.
  double countDisco(void) const
	{ return m_count; }

  /// Compute the average change in ONE attribute so far.
  /// @param aid The ID of the attribute that for which changes are returned.
  /// @return The current average discontinuity for just this one attribute.
  /// This does not include the weight, but norm is already factored.
  double averageDisco(ATTRIBUTE_ID aid) const;

  /// Count the changes in ONE attribute so far.
  /// @param aid The ID of the attribute that for which the count is returned.
  /// @return The # of discontinuous changes that have occurred in the
  /// specified attribute until now.
  double countDisco(ATTRIBUTE_ID aid) const;
};
  
};

#endif // __DISCO_METER_H
