/*
Copyright (C) 2007, 2008  Condel project, TU-Berlin.
Written by Jiehua Chen, Jan Sablatnig, Sven Grottke.

This file is part of Adam, the DVE testbed.

Adam is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3, or (at your option) any later
version.

Adam is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Adam; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.
*/
#include <math.h>

// comparisons are made within this precision
#define PRECISION 0.000001
// comparisons are made within 50 steps
#define STEPS 50

#include <omnetpp.h>
#include "gen.hh"

// A random function(pdf) with Pareto Distribution with a cutoff.
// p(x) := (a * b^a) / (x-x_shift)^(a+1)
//      AND P(x) := 0 for x<b
//      AND P(x) := 0 for x>cutoff
//
// Parameters:
//      a:  [a.k.a. k], a shape parameter
//      b:  x_min, D(x < x_min) = 0
//      c:  x_shift
//      cutoff:  when p_c() > cutoff, rerun p_c()
double pareto_cutoff(double a, double b, double c, double cutoff)
{
  double r=cutoff+1.0;

  while (r > cutoff)
    r= pareto_shifted(a, b, c);

  return r;
}

Define_Function(pareto_cutoff,4)

// E formel is:
// ( (m/c) - (m/c)^k ) * c * k /(k-1)/( 1-(m/c)^k )
// m: x_min
// c: cutoff
// k: shape parameter
double calculate_expect(double m, double k, double c){
  double rate=m/c;
  double pow_of_rate=pow(rate,k);
  return ( (rate-pow_of_rate)*c*k/(1-pow_of_rate)/(k-1) );
}

// calculate the derivative of the expected value pareto_cutoff
// note: the function increases monotonously, so the derivative is positiv
// e: expectation from pareto
// k: shape parameter, in (1,2]
// c: cutoff
double calculate_dir_of_expect(double m, double k, double c){
  if( k <= 1 || k > 2 )
    condel_error("invalid range for shape parameter in pareto distribution");
  if( m <= 0 || m > 5  )
    condel_error("invalid range for x_min in pareto distribution");
  double rate=m/c;
  double pow_of_rate=pow(rate,k);
  double rest=1 - pow_of_rate;
  return ( k * (rest - k * (pow_of_rate / rate) + k * pow_of_rate) / rest / rest / (k - 1) );
}

// simple interpolation with newton method
// note: the function increases monotonously, so the derivative is positiv
// e: expectation from pareto
// k: shape parameter, in (1,2]
// c: cutoff
double calculate_x_min(double e, double k, double c){
  if( k <= 1 || k > 2 )
    condel_error("invalid range for shape parameter in pareto distribution");
  if( e < 0 || e > 4.5  )
    condel_error("invalid range for expectation value in pareto distribution");
  double x_new = 0.05;
  double e_new = calculate_expect(x_new, k, c);
  double dir_e_new;
  int count = 0;

  while( fabs(e_new-e) > PRECISION  && count < STEPS ){
    count++;
    dir_e_new=calculate_dir_of_expect(x_new, k, c);
    x_new=x_new + ( e - e_new )/dir_e_new;
    if( x_new < 0 ) x_new = PRECISION;
    else if( x_new > c ) x_new = c - PRECISION;
    e_new=calculate_expect(x_new,k,c);
  }

  //cout << count ;
  return x_new;
}

// simple interpolation with newton method
// note: the function increases monotonously, so the derivative is positiv
// e: expectation from pareto
// k: shape parameter, in (1,2]
// c: cutoff
double calculate_x_min_cached(double e, double k, double c){
  static double e_last=-1;
  static double k_last=-1;
  static double c_last=-1;
  static double result=-1;

  if( e_last != e || k_last != k || c_last != c ){
    e_last = e;
    k_last = k;
    c_last = c;
    result = calculate_x_min(e, k, c);
  }
  
  return result;
}

// simplified version of the Pareto distribution
// c will be set to constant
// a (aka k) can be varied from expectation and a
double pareto_simple(double e) {
  if( e < PRECISION ) return 0.0; 
  double k=1.5;
  double x_min=calculate_x_min_cached(e, k, 5);
  return pareto_cutoff(k, x_min, 0, 5);
}

// extended version of the Pareto distribution
double pareto_extended(double e, double k){
  if( e < PRECISION ) return 0.0;
  if( k <= 1 || k > 2 )
    condel_error("invalid range for k in pareto distribution");
  double x_min = calculate_x_min_cached(e, k, 5);
  return pareto_cutoff(k, x_min, 0, 5);
}

// shifted version of the exponential distribution
double exponential_shifted(double e, double x_min){
  if( e < PRECISION ) return 0.0;
  if( x_min < PRECISION )
    condel_error("invalid range for x_min in exponential distribution");
  if( e <= x_min )
    condel_error("expectation must be larger as x_min");
  double x = exponential(e-x_min);
  return x+x_min;
}

Define_Function(pareto_simple, 1)
Define_Function(pareto_extended, 2)
Define_Function(exponential_shifted, 2)


// The expected value is:
// E = 0.99 * T + 0.0099 * (T +0.01s) + 0.0001 * (T + 1.01s)
// => E = T + 0.0002

double spike(double e)
{
  double t= e - 0.0002;
  if (t <0) return 0.0;
  double u= uniform(0, 1);
  if (u <= 0.99)
    return t;

  // uniform between 0..20ms
  if (u <= 0.9999)
    return t+ (u-0.99)*2.0;

  // uniform between 0..2s
  return t+ 0.02 + (u-0.9999)* 19800.0;
}

Define_Function(spike, 1);


namespace Condel {
    void condel_error(const char* pWhat) {
      fprintf(stderr, "ERROR: %s, aborting\n", pWhat);
      exit(1);
    }
  };

/*int main(){

int c=0;
for( double k = 1.00001; k < 2; k+=0.001 )
for( double m = 0.001; m < 4; m+=0.005  ){
c++;
//cout << c << ". ";
double e = calculate_expect(m, k, 5);
double x = calculate_x_min(e, k, 5);
//cout << "\t" << "e:  " << e <<  "\t x:  " << x ;
//if( fabs(x-m) > PRECISION ) cout << "\t m: " << m;
cout << endl;
}
      
return 0;
}*/
