/*
Copyright (C) 2007, 2008  Condel project, TU-Berlin.
Written by Jan Sablatnig, Sven Grottke, Jiehua Chen.

This file is part of Adam, the DVE testbed.

Adam is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3, or (at your option) any later
version.

Adam is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Adam; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.
*/
#include "host.hh"
#include "msg.hh"
#include "scenario.hh"
#include "consistency.hh"
#include "router.hh"


#define WINDOW_SIZE (4.0)


// internal precision on time-stamps is ms, everything else will be rounded.
#define TIME_PRECISION 1000

const double cHost::TICK_TIME = 0.05;
const double cHost::CHECK_TIME = 0.001;

cHost::cHost(void)
: cSimpleModule(), m_disco(), m_attNames(), m_load(), m_gate(),
  m_pTickEvent(NULL), m_pCheckEvent(NULL),
  m_packetsSent(WINDOW_SIZE), m_bytesSent(WINDOW_SIZE),
  m_packetsReceived(WINDOW_SIZE, 0.0, false),
  m_bytesReceived(WINDOW_SIZE, 0.0, false),
  m_lossRate(0.0), m_drops(0), m_total(0),
  m_factory(), m_router(), m_consistency(), m_world(), m_scenario(),
  m_attributeNorm()
{ }

cHost::~cHost(void)
{ }

void cHost::initialize(void)
{
  char newname[100];
  snprintf(newname, 99, "player@%lx", (unsigned long) id());
  setName(newname);

  // set up self events for ticking away
  if (m_pTickEvent)
    cancelAndDelete(m_pTickEvent);
  else
    m_pTickEvent= new cMessage("tick", 2);
  scheduleAt(now()+ TICK_TIME, m_pTickEvent);
  if (m_pCheckEvent)
    cancelAndDelete(m_pCheckEvent);
  else
    m_pCheckEvent= new cMessage("check", 3);
  scheduleAt(roundTime(now()+ CHECK_TIME), m_pCheckEvent);

  m_lossRate= par("loss_rate");

  m_packetsSent.initialize(now());
  m_bytesSent.initialize(now());
  m_packetsReceived.initialize(now());
  m_bytesReceived.initialize(now());
  m_drops= 0;
  m_total= 0;

  // measurement code
  m_disco.init(m_attributeNorm);
  m_load.initialize();

  m_attNames.erase(m_attNames.begin(), m_attNames.end());
  m_attributeNorm.clear();

  cStringTokenizer vm(par("disco_weights")," \t\n");
  while (vm.hasMoreTokens())
    {
    string at= vm.nextToken();
    assert(vm.hasMoreTokens());

#if VERBOSITY >= 3
    cout << "Now parsing " << at << endl; 
#endif
    if (!cAttributeFactory::nameExists(at))
      {
      cerr << "Error in .ned file:  Attribute '" << at << "' not found.\n";
      exit(-1);
      }

    m_attNames.push_back(at);

    m_disco.addAttribute(cAttributeFactory::nameToID(at), atof(vm.nextToken()));
    }

  m_disco.initWeights();

  unsigned i;
  for(i=0; i<(unsigned) gates(); i++)
    {
    cGate *g= gate(i);
    if (g && g->type() == 'O')
      m_gate[g->destinationGate()->ownerModule()->id()]= g;
    }


  getMsgFactory().initialize();
  getRouter().initialize();
  getConsistency().initialize();
  getWorld().initialize();
  getScenario().initialize();
}

void cHost::finish(void)
{
#if VERBOSITY >= 1
  cout << "cHost::finish() called" << endl;
#endif
  
  if (m_pTickEvent)
    cancelAndDelete(m_pTickEvent);
  m_pTickEvent= NULL;
  if (m_pCheckEvent)
    cancelAndDelete(m_pCheckEvent);
  m_pCheckEvent= NULL;

  recordScalar("tick (in s)", TICK_TIME);
  recordScalar("avg. cpu ops /tick", m_load.averageCPU());
  recordScalar("max. cpu ops /tick", m_load.maxCPU());
  recordScalar("avg. memory (in bytes)", m_load.averageMem());
  recordScalar("max. memory (in bytes)", m_load.maxMem());
  
  recordScalar("#discontinuities /s", m_disco.countDisco() / now());
  recordScalar("avg. G", m_disco.averageDisco() / now());

  list<string>::const_iterator at;
  for (at= m_attNames.begin(); at != m_attNames.end(); at++)
    {
    char name[100];
    snprintf(name, sizeof(name), "#discontinuities /s (%s)", (*at).data());
    ATTRIBUTE_ID aid= cAttributeFactory::nameToID(*at);
    recordScalar (name, m_disco.countDisco(aid) / now());

    snprintf(name, sizeof(name), "avg. G(%s)", (*at).data());
    recordScalar (name, m_disco.averageDisco(aid) / now());
    }

  recordScalar("avg. packets sent", m_packetsSent.getTotalAverage());
  recordScalar("max. packets sent", m_packetsSent.getMax());
  recordScalar("avg. bytes sent", m_bytesSent.getTotalAverage());
  recordScalar("max. bytes sent", m_bytesSent.getMax());
  recordScalar("avg. packets received", m_packetsReceived.getTotalAverage());
  recordScalar("max. packets received", m_packetsReceived.getMax());
  recordScalar("avg. bytes received", m_bytesReceived.getTotalAverage());
  recordScalar("max. bytes received", m_bytesReceived.getMax());

  recordScalar("real loss rate",m_drops/m_total);

  getRouter().finish();
  getScenario().finish();
}

void cHost::handleMessage(cMessage* pMsg)
{
  // is it the tick-timer?
  if (pMsg == m_pTickEvent)
    {
    // reschedule
    scheduleAt(roundTime(now() + TICK_TIME), m_pTickEvent);
#if VERBOSITY >= 2
    ev << "TICK!\n";
#endif
    handleTick();
    }
  else if (pMsg == m_pCheckEvent)
    {
    // reschedule
    scheduleAt(roundTime(now() + CHECK_TIME), m_pCheckEvent);
    handleCheck();
    }
  else
    {
    // it's an actual message; pass to message router
#if VERBOSITY >= 2
    ev << "regular message received\n";
#endif
    // take ownership of the message
    scoped_ptr<cMsg> msg(check_and_cast<cMsg*>(pMsg));

    if (!pMsg->isSelfMessage())
      {
      m_total++;
      if ( uniform(0,1) < m_lossRate )
        {
        m_drops++;
#if VERBOSITY >= 3
        ev << "Dropping message due to lossy connection." << endl;
#endif
        return;
        }
    
      // This message is received OK; tally it
      m_packetsReceived.collect(now(), 1);
      m_bytesReceived.collect(now(), pMsg->byteLength());
      }

    getRouter().handleMessage(msg.get());
    }
}

void cHost::handleTick()
{
  getLoad().nextFrame();
  // MISSING:  getLoad().addMem(), has to be done in the specific consistency.

  // actually handle ticks
  // getMsgFactory().handleTick();              // unnecessary
  // getRouter().handleTick();                  // now done by check
  getConsistency().handleTick();
  // getWorld().handleTick();             // done from Consistency
  // getScenario().handleTick();             // done from Consistency
}

void cHost::handleCheck()
{
  getRouter().handleCheck();
}

void cHost::sendOut(cMsg *pMsg, unsigned long hid)
{
  m_packetsSent.collect(now(), 1);
  m_bytesSent.collect(now(), pMsg->byteLength());

  send(pMsg, m_gate[hid]);
}


void cHost::split(unsigned long id)
{
  // TODO:  No network split implemented as of yet.
  // we might use error(), as well.
  cerr << "Network split; ending simulation" << endl;
  endSimulation();
}

double cHost::roundTime(double in)
{
  return round(in * TIME_PRECISION) / TIME_PRECISION;
}

double cHost::now(void) const
{
   return roundTime(simTime());
}




