/*
Copyright (C) 2007, 2008  Condel project, TU-Berlin.
Written by Jan Sablatnig, Sven Grottke, Jiehua Chen.

This file is part of Adam, the DVE testbed.

Adam is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3, or (at your option) any later
version.

Adam is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Adam; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.
*/

#ifndef __HOST_H
#define __HOST_H

#include <omnetpp.h>

#include "gen.hh"
#include "disco_meter.hh"
#include "load.hh"
#include "sliding_window.hh"

namespace Condel {

// forward references
class cScenario;
class cRouter;
class cConsistency;
class cWorld;
class cMsg;  
class cMsgFactory;

/// The base class for simulation hosts.
/// cHost represents a physical computer, as such it also contains (manages)
/// all the other parts (called components in this context) that run on the
/// individual computers.
/// This class also provides the entire interface to OMNeT.

class cHost : public cSimpleModule
{
private:
  /// Do not use this.
  /// Declared, but not defined to prevent accidental use of copy constructor.
  cHost(const cHost&);

protected:
  cDiscoMeter m_disco;
  list<string> m_attNames;
  cLoadMeter m_load;

  map<unsigned long, cGate*> m_gate;

  cMessage * m_pTickEvent;
  cMessage * m_pCheckEvent;

  // netload measurement stuff
  cSlidingWindow m_packetsSent;
  cSlidingWindow m_bytesSent;
  cSlidingWindow m_packetsReceived;
  cSlidingWindow m_bytesReceived;

  double m_lossRate;    // packet loss rate (as specified from outside)
  double m_drops;       // number of packets really lost
  double m_total;       // total number of packets arrived

  // These are the components of the host
  scoped_ptr<cMsgFactory> m_factory;   ///< Creates messages.
  scoped_ptr<cRouter> m_router;     ///< Sending/routing/reception of messages.
  scoped_ptr<cConsistency> m_consistency; ///< Consistency algorithm.
  scoped_ptr<cWorld> m_world; ///< Currently visible world.
  scoped_ptr<cScenario> m_scenario;  ///< scenario code

public:
  /// The norms for each attribute.
  /// @sa cDiscoMeter::init()
  map<const ATTRIBUTE_ID, double> m_attributeNorm; ///< The norms for each attr.
  /// Time per frame.
  /// Once per frame, the world simulation is advanced and the AI is run.
  static const double TICK_TIME;
  /// Fastest time-grain.
  /// This is used to check for message timeouts and such.
  static const double CHECK_TIME;


public:
  cHost(void);
  virtual ~cHost(void);

  // Access the components
  virtual const cMsgFactory &getMsgFactory(void) const { return *m_factory; }
  virtual cMsgFactory& getMsgFactory(void) { return *m_factory; }

  virtual const cScenario &getScenario(void)const {return *m_scenario;}
  virtual cScenario& getScenario(void) { return *m_scenario; }

  virtual const cRouter &getRouter(void) const { return *m_router; }
  virtual cRouter& getRouter(void) { return *m_router; }

  virtual const cConsistency &getConsistency(void)const {return *m_consistency;}
  virtual cConsistency& getConsistency(void) { return *m_consistency; }

  virtual const cWorld &getWorld(void)const {return *m_world;}
  virtual cWorld& getWorld(void) { return *m_world; }


  // access the measurement code
  const cDiscoMeter &getDisco(void) const { return m_disco; }
  cDiscoMeter &getDisco(void) { return m_disco; }

  cLoadMeter &getLoad(void) { return m_load; }


  /// Reset everything.
  /// Sets up the tick events (both tick and check) and reads the parameters
  /// from the OMNeT ini files.
  /// This member is called by OMNeT when the simulation is about to start.
  virtual void initialize(void);

  /// System Done.
  /// This member is called by OMNeT when the simulation has ended.
  virtual void finish(void);

  /// Accept a message.
  /// This can either be a network message sent by one of the other simulated
  /// hosts or it could be some tick event.
  /// This is called by OMNeT (and ONLY by OMNeT, please).
  /// @param pMsg A pointer to the message.  OMNeT (i.e. the caller of this
  /// function) retains ownership of the message, do NOT delete it.
  /// If you need your own copy, you'll have to dup() it.
  virtual void handleMessage(cMessage* pMsg);

  /// A frame has passed.
  /// This is called at the end of every frame, which are TICK_TIME seconds
  /// apart.  Called by handleMessage(cMessage *).
  virtual void handleTick(void);

  /// A check interval has passed.
  /// This is called at the end of every check_interval, which are  CHECK_TIME
  /// seconds apart.  Called by handleMessage(cMessage *).
  virtual void handleCheck(void);

  /// Send out a message to another host.
  /// A unicast message is sent to a single other host.
  /// @param pMsg A pointer to the message to be sent away.
  ///   CAUTION:  The message now belongs to this function (or OMNeT, rather).
  ///   Do not free/change it, just forget the pointer to it.
  /// @param id The targeted host.
  virtual void sendOut(cMsg *pMsg, unsigned long id);

  /// Perform a network split.
  /// Split the specified host out of the network.
  virtual void split(unsigned long id);

  // A unique id in the network of hosts (like an IP-number).
//  int id(void) const;       // INHERITED!

  /// The current time.
  /// @return The current simtime, __ROUNDED__!
  double now(void) const;

  /// Round a time.
  /// This rounds a time to the builtin precision.
  static double roundTime(double time);
};


/// A base class for all host-components.

class cHostComponent
{
protected:
  cHost &m_host;

public:
  cHostComponent(cHost &host)
    : m_host(host)
    { }

  const cHost &host(void) const { return m_host; }
  cHost &host(void) { return m_host; }
};
  
};

#endif // __HOST_H
