/*
Copyright (C) 2007, 2008  Condel project, TU-Berlin.
Written by Sven Grottke, Jan Sablatnig, Jiehua Chen.

This file is part of Adam, the DVE testbed.

Adam is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3, or (at your option) any later
version.

Adam is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Adam; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.
*/
#ifndef __INSTANCE_H
#define __INSTANCE_H

#include <list>

#include "gen.hh"
#include "object_factory.hh"
#include "attribute.hh"

namespace Condel {
  
// forward references
class cInputStream;
class cOutputStream;
class cLoadMeter;
  
/// Base class for all instances.
/// An instance is one version of an object (residing on some world on
/// some host).
/// @sa cInstanceFactory, cObjectFactory

class cInstance {
private:
  const OBJECT_ID m_id;
  const string m_name;
  const unsigned m_size;  // How many attributes this keeps
  
  double m_simtime;	// the instance's state is at this simulation time
  double m_updateTime;	// Last time an update was SENT for this inst.
  
  /// Do not use this.
  /// Declared, but not defined to disable copy constructor.
  cInstance(const cInstance&);
public:

  /// Constructor.
  /// Note that this is not usually called directly for two reasons.
  ///     <ul><li>A pure cInstance class does not make sense.  It needs to be
  ///      inherited from.  Besides, it's an abstract class.
  ///     <li>Even the inherited classes are not simply instantiated but rather
  ///      created through the cInstanceFactory.
  ///     </ul>
  /// @param id The id of the corresponding object.
  /// @param name The name of the corresponding object.
  /// @param size The amount of attributes this object has.
  cInstance(const OBJECT_ID& id, const string& name, const unsigned size);
  virtual ~cInstance(void) {}

  /// The ID of the corresponding object.
  /// All instances of an object share the same numerical ID.
  OBJECT_ID getObjectID(void) const { return m_id; }

  /// The name of the corresponding object.
  /// All instances of an object share the same name.
  const string& getName(void) const { return m_name; }

  /// Query the instance's internal simulation time.
  /// @return The instance's state is at this simulation time.
  double simTime(void) const {  return m_simtime; }
  /// Set the instance's internal simulation time.
  /// @param newtime The instance's state has change, this is the corresponding
  /// time.
  void setSimTime(double newtime) {  m_simtime= newtime; }

  /// Query the instance's last update time.
  /// @return The last time an update was SENT for this instance.
  double updateTime(void) const { return m_updateTime; }
  /// Set the instance's last update time.
  /// @param newtime An update was sent at this time.
  /// @todo Currently, this mechanism only supports global updates (i.e.
  ///  updates are sent to all other hosts).
  void setUpdateTime(double newtime) {  m_updateTime= newtime; }

  /// Get all attributes.
  /// Packs the current state of this instance into a cAttributeSet.
  /// @return A copy of the current state.
  virtual cAttributeSet getAttributes(void) const = 0;

  /// Set all attributes.
  /// Overwrite the internal state.
  /// @param na A new attribute set.  Every attribute found in na is
  ///  overwritten locally.  It is legal to pass a sparse set here.
  virtual void setAttributes(const cAttributeSet& na) = 0;

  /// Create a deep copy of this instance.
  /// @return An exact copy of this instance, including name.  This must
  /// not be added to the same cWorld that this instance resides in.
  virtual cInstance* clone(void) const = 0;

  // The construction callback for cInstanceFactory.
  // The children of cInstance usually implement a STATIC function like this.
  // You usually do not wish to call this function, call
  // cInstanceFactory.createInstance() instead.
  //static cInstance *create(const OBJECT_ID& id, const uname &name);

  /// Compares two instances.
  /// This measures the similarity, in an appropriate way fitting for this
  /// object.  Please note that this means that this metric is not comparable
  /// between different instances of different objects, i.e. the distance
  /// between two balls cannot be compared to the distance between two scores.
  /// @param other Another instance of the same object.
  /// @return A metric of the similarity of the two instances.
  virtual double getDistance(const cInstance& other) const = 0;

  /// Simulate this instance.
  /// Simulates from m_simtime until newtime, then sets simtime to newtime.
  /// @param newtime A simulation time in the future of simTime().  Only
  /// the difference between the two times is advanced.
  /// @param load A tool to measure CPU and memory usage during the simulation.
  virtual void simulate(double newtime, cLoadMeter &load) = 0;

  /// Estimate the entire in-memory size of this instance.
  /// @return The # of bytes necessary to keep the attribute set of this
  /// instance.
  /// @sa cAttributeSet::getCodedLength()
  unsigned getCodedLength(void) { return m_size * 8; }
};

};

#endif // __INSTANCE_H
