/*
Copyright (C) 2007, 2008  Condel project, TU-Berlin.
Written by Jan Sablatnig, Sven Grottke, Jiehua Chen.

This file is part of Adam, the DVE testbed.

Adam is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3, or (at your option) any later
version.

Adam is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Adam; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.
*/
#ifndef __INSTANCEFACTORY_H
#define __INSTANCEFACTORY_H

#include <map>

#include "gen.hh"
#include "object_factory.hh"


namespace Condel {

// forward references
class cInstance;
class cReplication;


/// An instance creation callback.
typedef cInstance *(*instanceCreator)(const OBJECT_ID&, const string&);
  
/// Creates instances.
/// An instance is one version of an object (residing on some world on
/// some host).
///
/// It is then a frequent problem for a host to instantiate a new instance
/// given just the object ID.  This factory class here solves this problem.
///
/// This is a purely static class.  It can create any instance in the virtual
/// universe.  In fact, all instances should be created through this class.
/// @sa cInstance, cObjectFactory

class cInstanceFactory {
private:
    // Gets the list of registered functions.
    static map<const string, instanceCreator> &factoryList(void);

public:
    /// Register a creation callback.
    /// Each INSTANCE-class needs to call this function exactly ONCE (globally)
    /// to register itself.  The createinstance function should create an
    /// instance of the appropriate type.
    /// @param obj The C++ name of the corresponding instance class.
    /// @param cbf The callback function to create an instance of class obj.
    static void registerInstance(const string &obj, instanceCreator cbf);

    /// Create an instance of the given name.
    /// The instance creation callback must have been registered previously,
    /// or you'll get a condel_error() here (which in turn exits).
    /// @param inst The name of the object for which to create an instance.
    ///   The correct callback function is found by searching for the c++
    ///   class name of the instance, which is the first part of the instance
    ///   name (up to any #, if there is one).
    /// @param oid The object ID this corresponds to.
    /// @return A fresh instance of the correct type and with the given name.
    /// @todo The object ID is kind of redundant and could be supplied by
    /// cObjectFactory::nameToID(const string &).
    static cInstance *createInstance(const OBJECT_ID& oid, const string& inst);
};

};

#endif // __INSTANCEFACTORY_H
