/*
Copyright (C) 2007, 2008  Condel project, TU-Berlin.
Written by Jan Sablatnig, Sven Grottke, Jiehua Chen.

This file is part of Adam, the DVE testbed.

Adam is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3, or (at your option) any later
version.

Adam is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Adam; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.
*/

#ifndef __LOAD_H
#define __LOAD_H

#include "gen.hh"

namespace Condel {


/// A CPU/memory load measuring tool.
/// This class gathers data on cpu/memory load using instrumentation by hand.
/// @todo This should inherit from OMNeT++'s cStatistic some day.

class cLoadMeter
{
protected:
  unsigned long long m_CpuSum; // Sum of all Cpu load used
  unsigned long m_CpuSumCount; // The above sum contain this many ticks
  unsigned long long m_memSum; // Sum of all Memory used (one measurement per tick)
  unsigned long m_memSumCount; // The above sum contain this many ticks
  unsigned long long m_CpuMax; // Maximum Cpu load used in a tick, ever
  unsigned long long m_memMax; // Maximum Memory used, ever

  unsigned long long  m_CpuCurrent;  // Current Frame's CPU load

public:
  cLoadMeter(void);
  virtual ~cLoadMeter(void) {}

  /// Reset everything.
  virtual void initialize(void);

  /// Finish a frame.
  /// This also starts a new frame of CPU measurements.
  virtual void nextFrame(void);

  /// Record CPU usage.
  /// Call this whenever you wish to record some CPU load.
  /// @param ops Some estimate of the pseudo ops that were just used in
  /// some calculation.
  void addCPU(unsigned long long ops)
    { m_CpuCurrent += ops; }

  /// Record memory usage.
  /// Call this once per frame to specify how many bytes of memory were
  /// used this tick.
  /// @param bytes The total maximum amount of memory (in bytes) necessary to
  /// run all algorithms on this host (since there is only one load meter
  /// per host) for this frame.  In our case, this is often an estimate for
  /// a production optimized version (i.e. NOT a hardware solution).
  void addMem(unsigned long long bytes)
    {
    if (bytes > m_memMax)
      m_memMax= bytes;
    m_memSum += bytes;
    m_memSumCount++;
    }

  /// Average CPU load so far.
  /// @return The average CPU pseudo ops per frame up to now.
  double averageCPU(void)
    { 
    if (!m_CpuSumCount)
      return 0.0;
    return ((double) m_CpuSum) / m_CpuSumCount;
    }

  /// Maximum CPU load so far.
  /// @return The maximum CPU pseudo ops in any frame recorded up to now.
  unsigned long long maxCPU(void)
    { return m_CpuMax; }

  /// Average memory use so far.
  /// @return The average # of bytes of memory per frame up to now.
  double averageMem(void)
    {
    if (!m_memSumCount)
      return 0.0;
    return ((double) m_memSum) / m_memSumCount;
    }

  /// Maximum memory use so far.
  /// @return The maximum # of bytes of memory in any frame recorded up
  /// to now.
  unsigned long long maxMem(void)
    { return m_memMax; }
};
  
};

#endif // __SLIDING_WINDOW_H
