/*
Copyright (C) 2007, 2008  Condel project, TU-Berlin.
Written by Jan Sablatnig, Sven Grottke, Jiehua Chen.

This file is part of Adam, the DVE testbed.

Adam is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3, or (at your option) any later
version.

Adam is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Adam; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.
*/
#include <limits.h>

#include "msg.hh"




Register_Class(cMsg);

cMsg::cMsg(long host, unsigned long long order, double timeout, double batching,
           int type, const char *name)
: cMessage(name), m_type(type), m_timeout(timeout), m_batching(batching),
  m_hostID(host), m_order(order), m_objectID(0), m_objectName(),
  m_attributes()
{ setTimestamp(); }

cMsg::cMsg(const cMsg& other)
: cMessage(other), m_type(other.m_type), m_timeout(other.m_timeout),
  m_batching(other.m_batching),
  m_hostID(other.m_hostID), m_order(other.m_order),
  m_objectID(other.m_objectID), m_objectName(other.m_objectName),
  m_attributes(other.m_attributes)
{ }

cMsg& cMsg::operator=(const cMsg& other)
{
    if (this==&other) return *this;
    cMessage::operator=(other);
    setName(other.name());

    m_type = other.m_type;
    m_timeout = other.m_timeout;
    m_batching = other.m_batching;
    m_hostID = other.m_hostID;
    m_order = other.m_order;
    m_objectID = other.m_objectID;
    m_objectName = other.m_objectName;
    m_attributes = other.m_attributes;

    return *this;
}



#include "host.hh"


cMsgFactory::cMsgFactory(cHost &host)
: cHostComponent(host), m_current(0)
{ }

auto_ptr<cMsg> cMsgFactory::createMsg(const char *name, double timeout,
                double batching)
{
  auto_ptr<cMsg> msg(new cMsg(host().id(), m_current++, timeout, batching,
                        cMsg::DATA, name));
  return msg;
}

auto_ptr<cMsg> cMsgFactory::createStartMsg(void)
{
  // incorrect hostID, but shouldn't be a problem, this stays local.
  auto_ptr<cMsg> msg(new cMsg(0, 0, false, 0.0, cMsg::START, "Start"));

  msg->setTimestamp(-1.0);
  msg->setObjectID(UINT_MAX);
  msg->setObjectName("none");

  return msg;
}

void cMsgFactory::initialize(void)
{
  m_current= 0;
}



// Much of this is taken from the .cc file created from this through opp_msg:
//      message cMsg
//      {
//      fields:
//         int type;
//         double timeout;
//         double batching;
//         long hostID;
//         long long order;
//         unsigned int objectID;
//         string objectName;
//         int attributes;
//      };

// Actually, opp_msgc does not groc long long, had to do that one by hand.


// ---------------------------------- SNIP ------------------------------------

// Disable warnings about unused variables, empty switch stmts, etc:
#ifdef _MSC_VER
#  pragma warning(disable:4101)
#  pragma warning(disable:4065)
#endif

#include <iostream>
#include <sstream>

// Template rule which fires if a struct or class doesn't have operator<<
template<typename T>
std::ostream& operator<<(std::ostream& out,const T&) {return out;}

// Another default rule (prevents compiler from choosing base class' doPacking())
template<typename T>
void doPacking(cCommBuffer *, T& t) {
    throw new cException("Parsim error: no doPacking() function for type %s or its base class (check message.cc!)",opp_typename(typeid(t)));
}
template<typename T>
void doUnpacking(cCommBuffer *, T& t) {
    throw new cException("Parsim error: no doUnpacking() function for type %s or its base class (check message.cc!)",opp_typename(typeid(t)));
}

// Automatically supply array (un)packing functions
template<typename T>
void doPacking(cCommBuffer *b, T *t, int n) {
    for (int i=0; i<n; i++)
        doPacking(b,t[i]);
}
template<typename T>
void doUnpacking(cCommBuffer *b, T *t, int n) {
    for (int i=0; i<n; i++)
        doUnpacking(b,t[i]);
}
inline void doPacking(cCommBuffer *, cPolymorphic&) {}
inline void doUnpacking(cCommBuffer *, cPolymorphic&) {}

#define DOPACKING(T,R) \
    inline void doPacking(cCommBuffer *b, T R a) {b->pack(a);}  \
    inline void doPacking(cCommBuffer *b, T *a, int n) {b->pack(a,n);}  \
    inline void doUnpacking(cCommBuffer *b, T& a) {b->unpack(a);}  \
    inline void doUnpacking(cCommBuffer *b, T *a, int n) {b->unpack(a,n);}
#define _
DOPACKING(char,_)
DOPACKING(unsigned char,_)
DOPACKING(bool,_)
DOPACKING(short,_)
DOPACKING(unsigned short,_)
DOPACKING(int,_)
DOPACKING(unsigned int,_)
DOPACKING(long,_)
DOPACKING(unsigned long,_)
DOPACKING(float,_)
DOPACKING(double,_)
DOPACKING(long double,_)
DOPACKING(char *,_)
DOPACKING(const char *,_)
DOPACKING(opp_string,&)
//DOPACKING(std::string,&)
#undef _
#undef DOPACKING

// ---------------------------------- SNAP ------------------------------------

void doPacking(cCommBuffer *b, const cAttributeSet &atts)
{
  cAttributeSet::const_iterator at;
  for (at= atts.begin(); at != atts.end(); at++)
    {
    doPacking(b, (*at).second.getID());
    doPacking(b, (*at).second.getValue());
    }
  doPacking(b, UINT_MAX);
}

void doUnpacking(cCommBuffer *b, cAttributeSet &atts)
{
  unsigned id;
  double val;

  while (1)
    {
    doUnpacking(b, id);
    if (id == UINT_MAX)
      { break; }
    doUnpacking(b, val);
    cAttribute n(id, val);
    atts.add(n);
    }
}

void doPacking(cCommBuffer *b, const unsigned long long &order)
{
  doPacking(b, (unsigned long) (order>>32));
  doPacking(b, (unsigned long) order);
}

void doUnpacking(cCommBuffer *b, unsigned long long &order)
{
  unsigned long high, low;
  doUnpacking(b, high);
  doUnpacking(b, low);
  order= (((unsigned long long ) high) <<32 ) | ((unsigned long long) low);
}

// ---------------------------------- SNIP ------------------------------------
// slight editing (_var -> m_)


void cMsg::netPack(cCommBuffer *b)
{
    cMessage::netPack(b);
    doPacking(b,this->m_type);
    doPacking(b,this->m_timeout);
    doPacking(b,this->m_batching);
    doPacking(b,this->m_hostID);
    doPacking(b,this->m_order);
    doPacking(b,this->m_objectID);
    doPacking(b,this->m_objectName);
    doPacking(b,this->m_attributes);
}

void cMsg::netUnpack(cCommBuffer *b)
{
    cMessage::netUnpack(b);
    doUnpacking(b,this->m_type);
    doUnpacking(b,this->m_timeout);
    doUnpacking(b,this->m_batching);
    doUnpacking(b,this->m_hostID);
    doUnpacking(b,this->m_order);
    doUnpacking(b,this->m_objectID);
    doUnpacking(b,this->m_objectName);
    doUnpacking(b,this->m_attributes);
}

// ---------------------------------- SNAP ------------------------------------

// ---------------------------------- SNIP ------------------------------------
// slight editing (_var -> m_; handle attributes)

class cMsgDescriptor : public cStructDescriptor
{
  public:
    cMsgDescriptor();
    virtual ~cMsgDescriptor();
    cMsgDescriptor& operator=(const cMsgDescriptor& other);
    virtual cPolymorphic *dup() const {return new cMsgDescriptor(*this);}

    virtual int getFieldCount();
    virtual const char *getFieldName(int field);
    virtual int getFieldType(int field);
    virtual const char *getFieldTypeString(int field);
    virtual const char *getFieldEnumName(int field);
    virtual int getArraySize(int field);

    virtual bool getFieldAsString(int field, int i, char *resultbuf, int bufsize);
    virtual bool setFieldAsString(int field, int i, const char *value);

    virtual const char *getFieldStructName(int field);
    virtual void *getFieldStructPointer(int field, int i);
    virtual sFieldWrapper *getFieldWrapper(int field, int i);
};

Register_Class(cMsgDescriptor);

cMsgDescriptor::cMsgDescriptor() : cStructDescriptor("cMessage")
{
}

cMsgDescriptor::~cMsgDescriptor()
{
}

int cMsgDescriptor::getFieldCount()
{
    return baseclassdesc ? 8+baseclassdesc->getFieldCount() : 8;
}

int cMsgDescriptor::getFieldType(int field)
{
    if (baseclassdesc) {
        if (field < baseclassdesc->getFieldCount())
            return baseclassdesc->getFieldType(field);
        field -= baseclassdesc->getFieldCount();
    }
    switch (field) {
        case 0: return FT_BASIC;
        case 1: return FT_BASIC;
        case 2: return FT_BASIC;
        case 3: return FT_BASIC;
        case 4: return FT_BASIC_ARRAY;
        case 5: return FT_BASIC;
        case 6: return FT_BASIC;
        case 7: return FT_STRUCT_ARRAY;
        default: return FT_INVALID;
    }
}

const char *cMsgDescriptor::getFieldName(int field)
{
    if (baseclassdesc) {
        if (field < baseclassdesc->getFieldCount())
            return baseclassdesc->getFieldName(field);
        field -= baseclassdesc->getFieldCount();
    }
    switch (field) {
        case 0: return "type";
        case 1: return "timeout";
        case 2: return "batching";
        case 3: return "hostID";
        case 4: return "order";
        case 5: return "objectID";
        case 6: return "objectName";
        case 7: return "attributes";
        default: return NULL;
    }
}

const char *cMsgDescriptor::getFieldTypeString(int field)
{
    if (baseclassdesc) {
        if (field < baseclassdesc->getFieldCount())
            return baseclassdesc->getFieldTypeString(field);
        field -= baseclassdesc->getFieldCount();
    }
    switch (field) {
        case 0: return "int";
        case 1: return "double";
        case 2: return "double";
        case 3: return "long";
        case 4: return "long long";
        case 5: return "unsigned int";
        case 6: return "string";
        case 7: return "cAttributeSet";
        default: return NULL;
    }
}

const char *cMsgDescriptor::getFieldEnumName(int field)
{
    if (baseclassdesc) {
        if (field < baseclassdesc->getFieldCount())
            return baseclassdesc->getFieldEnumName(field);
        field -= baseclassdesc->getFieldCount();
    }
    switch (field) {
        default: return NULL;
    }
}

int cMsgDescriptor::getArraySize(int field)
{
    if (baseclassdesc) {
        if (field < baseclassdesc->getFieldCount())
            return baseclassdesc->getArraySize(field);
        field -= baseclassdesc->getFieldCount();
    }
    cMsg *pp = (cMsg *)p;
    switch (field) {
        case 4: return 2;
        case 7: return pp->getAttributes().size();
        default: return 0;
    }
}

bool cMsgDescriptor::getFieldAsString(int field, int i, char *resultbuf, int bufsize)
{
    if (baseclassdesc) {
        if (field < baseclassdesc->getFieldCount())
            return baseclassdesc->getFieldAsString(field,i,resultbuf,bufsize);
        field -= baseclassdesc->getFieldCount();
    }
    cMsg *pp = (cMsg *)p;
    switch (field) {
        case 0: long2string(pp->getType(),resultbuf,bufsize); return true;
        case 1: double2string(pp->getTimeOut(),resultbuf,bufsize); return true;
        case 2: double2string(pp->getBatching(),resultbuf,bufsize); return true;
        case 3: long2string(pp->getHostID(),resultbuf,bufsize); return true;
        case 4:
          {
          if (i==0)
            {
            long2string((unsigned long) pp->getOrder(),resultbuf,bufsize);
            return true;
            }
          else if (i==1)
            {
            long2string((unsigned long) (pp->getOrder()>>32),resultbuf,bufsize);
            return true;
            }
          else
            return false;
          }
        case 5: long2string(pp->getObjectID(),resultbuf,bufsize); return true;
        case 6: oppstring2string(pp->getObjectName(),resultbuf,bufsize); return true;
        case 7: return false;
        default: return false;
    }
}

bool cMsgDescriptor::setFieldAsString(int field, int i, const char *value)
{
    if (baseclassdesc) {
        if (field < baseclassdesc->getFieldCount())
            return baseclassdesc->setFieldAsString(field,i,value);
        field -= baseclassdesc->getFieldCount();
    }
    cMsg *pp = (cMsg *)p;
    switch (field) {
        case 0: pp->setType(string2long(value)); return true;
        case 1: pp->setTimeOut(string2double(value)); return true;
        case 2: pp->setBatching(string2double(value)); return true;
        case 3: pp->setHostID(string2long(value)); return true;
        case 4: 
          {
          if (i==0)
            {
            unsigned long long ord= pp->getOrder() | 0xffff0000ULL;
            pp->setOrder(ord | ((unsigned long long ) string2long(value)));
            return true;
            }
          else if (i==1)
            {
            unsigned long long ord= pp->getOrder() | 0x0000ffffULL;
            pp->setOrder(ord | (((unsigned long long ) string2long(value)) <<32));
            return true;
            }
          else
            return false;
          }
        case 5: pp->setObjectID(string2long(value)); return true;
        case 6: pp->setObjectName((value)); return true;
        case 7: return false;
        default: return false;
    }
}

const char *cMsgDescriptor::getFieldStructName(int field)
{
    if (baseclassdesc) {
        if (field < baseclassdesc->getFieldCount())
            return baseclassdesc->getFieldStructName(field);
        field -= baseclassdesc->getFieldCount();
    }
    switch (field) {
        case 7: return "cAttribute"; break;
        default: return NULL;
    }
}

void *cMsgDescriptor::getFieldStructPointer(int field, int i)
{
    if (baseclassdesc) {
        if (field < baseclassdesc->getFieldCount())
            return baseclassdesc->getFieldStructPointer(field, i);
        field -= baseclassdesc->getFieldCount();
    }
    cMsg *pp = (cMsg *)p;
    switch (field) {
        case 7:
                {
                cAttributeSet::iterator at;
                int j = 0;
                cAttributeSet &atts(pp->getAttributes());
                for (at= atts.begin(); at != atts.end(); at++, j++)
                  if (j==i)
                    return &((*at).second);
                return NULL;
                }
        default: return NULL;
    }
}

sFieldWrapper *cMsgDescriptor::getFieldWrapper(int field, int i)
{
    return NULL;
}

// ---------------------------------- SNAP ------------------------------------

