/*
Copyright (C) 2007, 2008  Condel project, TU-Berlin.
Written by Jan Sablatnig, Sven Grottke, Jiehua Chen.

This file is part of Adam, the DVE testbed.

Adam is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3, or (at your option) any later
version.

Adam is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Adam; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.
*/

#ifndef __MSG_H
#define __MSG_H

#include <omnetpp.h>

#include "gen.hh"
#include "attribute.hh"
#include "host.hh"

namespace Condel {

/// The base class for all messages.
/// All messages that are exchanged between hosts are based on this
/// class.
///
/// Note that messages are not only interchanged and then interpreted
/// immediately, they also often kept.  This may be to have a copy if
/// the transmission failed or to have a message history (useful for
/// optimistic consistency).
///
/// Note that getHostID(), timestamp(), and getOrder() together allow for
/// a universally unique identification of this message.  In other words,
/// by using these three numbers as a serial number, each original message
/// in the entire network has a different serial number.  Note, however, that
/// you may get the same message several times (because of the ARQ algorithm)
/// and that the same message may be sent to many different hosts.
///
/// This file is partly inspired by the result of running opp_msgc on
/// msg.opp.  It thus interfaces with OMNeT very closely.

class cMsg : public cMessage
{
public:

  /// Types of messages.
  /// More types could be added.
  enum MsgType
    {
    START = -1, ///< The first message in an optimistic consistency setting.
    DATA = 0,   ///< A normal message.
    ACK  = 1    ///< An acknowledgement message.
    };

protected:

  // Note:  if you change these parameters, you need to change their
  // size-estimates in setAttributes as well.

  int m_type;   // type, one of the above MsgType
  double m_timeout;  // if >0, this message will be sent reliably with that T/O.

  double m_batching; // May wait this many s before sending,
                     // aggregating with other messages if possible.

  // These three parameters both allow unique designation/identification
  // of a message AND a complete ordering of all messages.
  long m_hostID;      // id of the host who created this message
  // simtime_t timestamp()  // time when this message was created
  unsigned long long m_order;   // specifies a specific order in case hostid and
                      // timestamp are equal

  unsigned m_objectID;
  string m_objectName;


  cAttributeSet m_attributes;

  /// Do not use this.
  /// The comparison operator is protected, declared, but not implemented,
  /// to prevent accidental usage.
  bool operator==(const cMsg&);

public:
  /// The constructor.
  /// Do NOT use this one.  Use the cMsgFactory, instead.
  /// @param host The host this message was created on.
  /// @param order A unique serial number for messages created on this host.
  /// @param timeout If not 0.0, the message will be sent reliably and this
  /// is the timeout of the selective repeat algorithm.
  /// @param batching If negative, no batching is performed.  If 0.0, all
  /// messages created for the same target host during this CHECK_TIME will be
  /// combined.  If > 0.0, The sending algorithm will delay this message for
  /// a maximum of batching seconds to combine it with other messages going
  /// to the same host.
  /// @param type One of the types of messages, usually cMsg::DATA.
  /// @param name A name for this message, this is purely for debugging.
  cMsg(long host=0, unsigned long long order=0, double timeout=0.0,
        double batching=0.0, int type=cMsg::DATA, const char *name=NULL);
  cMsg(const cMsg& other);
  virtual ~cMsg(void) {}
  cMsg& operator=(const cMsg& other);
  virtual cPolymorphic *dup(void) const {return new cMsg(*this);}

  virtual void netPack(cCommBuffer *b);
  virtual void netUnpack(cCommBuffer *b);

  // Get the message type.
  // @return One of MsgType.
  virtual int getType(void) const { return m_type; }
  // Set the message type.
  // @param type One of MsgType.
  virtual void setType(int type) { m_type= type; }

  /// Get the timeout.
  /// @return If not 0.0, the message is sent reliably and this
  /// is the timeout of the selective repeat algorithm.
  virtual double getTimeOut(void) const { return m_timeout; }
  /// Set the timeout.
  /// Make sure to set this before the message is actually sent out,
  /// it makes little sense to change this on a received message.
  /// @param timeout If not 0.0, the message will be sent reliably and this
  /// is the timeout of the selective repeat algorithm.
  virtual void setTimeOut(double timeout) { m_timeout= timeout; }

  /// Get the batching parameter.
  /// @return If negative, no batching is performed.  If 0.0, all
  /// messages created for the same target host during this CHECK_TIME will be
  /// combined.  If > 0.0, The sending algorithm will delay this message for
  /// a maximum of batching seconds to combine it with other messages going
  /// to the same host.
  virtual double getBatching(void) const { return m_batching; }
  /// Set the batching parameter.
  /// Make sure to set this before the message is actually sent out,
  /// @param batching If negative, no batching is performed.  If 0.0, all
  /// messages created for the same target host during this CHECK_TIME will be
  /// combined.  If > 0.0, The sending algorithm will delay this message for
  /// a maximum of batching seconds to combine it with other messages going
  /// to the same host.
  virtual void setBatching(bool batching) { m_batching= batching; }

  /// Get the originating host ID.
  /// @return The host this message was created on.
  virtual long getHostID(void) const { return m_hostID; }
  /// Set the originating host ID.
  /// Usually, you do not want to change this.
  /// @param hostID The host this message was created on.
  virtual void setHostID(long hostID) { m_hostID= hostID; }

  /// Get the order number.
  /// @return A unique serial number for messages created on this host.
  virtual unsigned long long getOrder(void) const { return m_order; }
  /// Set the order number.
  /// Usually, you do not want to change this.
  /// @param order A unique serial number for messages created on this host.
  virtual void setOrder(unsigned long long order) { m_order= order; }

  /// Get the object ID.
  /// Most messages pertain to exactly one object.  In those cases, this
  /// is the ID of that object.  The attributes encoded in this message will
  /// overwrite the state of the corresponding instance on receiving hosts.
  /// @return ID of the object whose state is encoded in this message.
  virtual unsigned getObjectID(void) const { return m_objectID; }
  /// Set the object ID.
  /// @param objectID The object ID this message is about.  The attributes
  /// encoded in this message will overwrite the states of the corresponding
  /// instance on receiving hosts.
  virtual void setObjectID(unsigned objectID) { m_objectID= objectID; }
  /// Get the object name.
  /// Most messages pertain to exactly one object.  In those cases, this
  /// is the name of that object.  The attributes encoded in this message will
  /// overwrite the state of the corresponding instance on receiving hosts.
  /// @return Name of the object whose state is encoded in this message.
  virtual const char * getObjectName(void) const { return m_objectName.c_str();}
  /// Set the object name.
  /// @param objectName The object name this message is about.  The attributes
  /// encoded in this message will overwrite the state of the corresponding
  /// instance on receiving hosts.
  virtual void setObjectName(const char * objectName)
        { m_objectName= objectName; }

  /// Get the encoded attributes.
  /// @return A cAttributeSet generated by some cInstance's
  /// cInstance::getAttributes().  This is the payload of the message.
  /// The instance in question should have this message's getObjectID().
  virtual cAttributeSet& getAttributes(void) { return m_attributes; }
  /// Get the encoded attributes.
  /// @return A cAttributeSet generated by some cInstance's
  /// cInstance::getAttributes().  This is the payload of the message.
  /// The instance in question should have this message's getObjectID().
  virtual const cAttributeSet& getAttributes(void) const
        { return const_cast<cMsg *>(this)->getAttributes(); }
  /// Encode attributes.
  /// @param attributes A cAttributeSet generated by some cInstance's
  /// cInstance::getAttributes().  This is the payload of the message.
  /// The instance in question should have this message's getObjectID().
  virtual void setAttributes(const cAttributeSet& attributes)
        {
        m_attributes=attributes;
        setByteLength(attributes.getCodedLength()
        // This is the ONE place where the size of the hidden members is
        // guesstimated.
        // Currently assuming (in bytes):
        //      4 - objectID
        //      4 - hostID
        //      4 - timestamp
        //      4 - order       (really using 8)
        // Obviously this might be compressed a bit better, but this is an
        // upper bound.
        + 16);
        }

  /// An order operator.
  /// Using this operation, a total order of all events can be attained by
  /// simple sorting.
  bool operator<(const cMsg& other) const
        {
        return timestamp() < other.timestamp() ||
              ( timestamp() == other.timestamp() &&
                getHostID() < other.getHostID() ) ||
              ( timestamp() == other.timestamp() &&
                getHostID() == other.getHostID() &&
                getOrder() < other.getOrder() );
        }
  /// A comparison operator.
  /// if this operator evaluates to true, then the two messages carry the
  /// same extended ID and should therefore be equal (unless something went
  /// seriously wrong).
  bool operator==(const cMsg &other) const
        {
        return timestamp() == other.timestamp() &&
               getHostID() == other.getHostID() &&
               getOrder() == other.getOrder();
        }
};

inline void doPacking(cCommBuffer *b, cMsg& obj) {obj.netPack(b);}
inline void doUnpacking(cCommBuffer *b, cMsg& obj) {obj.netUnpack(b);}



class cHost;

/// A message factory.
/// Use this helper to construct cMsg'es for you.
/// Basically, it's just a counter to keep order numbers sequential and unique.
class cMsgFactory : public cHostComponent
{
private:
  unsigned long long m_current;

public:
  cMsgFactory(cHost& host);
  virtual ~cMsgFactory(void) { }

  /// Create a cMsg, with a running order number.
  /// The host and order number will automatically be set by this function,
  /// timestamp() is the current cHost::simtime().  Type of the message
  /// is cMsg::DATA.
  /// @param name A name for this message, this is purely for debugging.
  /// @param timeout If not 0.0, the message will be sent reliably and this
  /// is the timeout of the selective repeat algorithm.
  /// @param batching If negative, no batching is performed.  If 0.0, all
  /// messages created for the same target host during this CHECK_TIME will be
  /// combined.  If > 0.0, The sending algorithm will delay this message for
  /// a maximum of batching seconds to combine it with other messages going
  /// to the same host.
  /// @return A pointer to a correctly instantiated message.  Assign this
  /// to an auto_ptr<cMsg> to avoid any possibility of leakage.
  /// @sa cMsg::cMsg(long, unsigned long long, double, double, int,const char *)
  auto_ptr<cMsg> createMsg(const char *name=NULL, double timeout=0.0,
                double batching=0.0);

  /// Create a special start event cMsg.
  /// You should only do this once per host.
  static auto_ptr<cMsg> createStartMsg(void);

  /// Reset all.
  virtual void initialize(void);
};
  
}

#endif // __MSG_H
