/*
Copyright (C) 2007, 2008  Condel project, TU-Berlin.
Written by Sven Grottke, Jan Sablatnig, Jiehua Chen.

This file is part of Adam, the DVE testbed.

Adam is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3, or (at your option) any later
version.

Adam is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Adam; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.
*/

#ifndef __OBJECT_FACTORY_HH
#define __OBJECT_FACTORY_HH

#include <map>

#include "gen.hh"

namespace Condel {

/// Type for object IDs.
/// This typedef may change in the future, so use this defined type.
typedef unsigned OBJECT_ID;

/// Object ID manager.
/// In our model, objects may have instances of themselves on some or all
/// of the hosts.  To be quite precise, each cWorld can house an instance
/// of each object.  Objects are differentiated both by their unique name
/// and their corresponding unique object ID.
///
/// The name of each object is of the form
///     \<c++classname\>[#\<host_name\>]
/// , e.g. cRacket#2.  Instances do not have names (apart from their object
/// name) to further distinguish them.
///
/// There is no actual base class for objects, since objects are never
/// created.  Only their instances are created.  All instances are based
/// on cInstance.
/// 
/// cObjectFactory is a purely static class used to globabally assign the
/// IDs for objects.  It also handles the translation of object name->object ID.
///
/// Note that this scheme is a cludge, it would not be possible to do this
/// on a truly distributed system.  There are schemes to assign such numbers
/// in a distributed fashion though, and we decided to ignore this sub-
/// problem for greater clarity of the rest of the code.
/// @sa cInstance, cInstanceFactory, cAttributeFactory

 
class cObjectFactory
{
private:  
  //gets the next free id and counts it up.
  static OBJECT_ID nextFreeID(void);

  // map names to object IDs
  static map<const string, OBJECT_ID> &factoryList(void);
  
public:
  /// Create a unique ID for a new object.
  /// @param name The name of the new object.
  /// @return The newly assigned object ID.
  static OBJECT_ID registerObject(const string &name);

  /// Returns the ID of a formerly registered object.
  /// If the object was NOT registered before, the application will
  /// exit because of a builtin assert here.
  /// @param name The name of the registered object.
  /// @return The corresponding object ID.
  /// @sa nameExists
  static OBJECT_ID nameToID(const string &name); // get ID from name

  /// Returns a list of currently registered objects.
  /// Try not to use this if you don't really have to.
  static const map<const string, OBJECT_ID> &objectMap(void)
	{ return factoryList(); }

  /// Check if an attribute was already registered.
  /// @param name The name of the attribute in question.
  /// @return True if this name exists already in the DB.
  static bool nameExists(const string &name);

  /// Clears ALL previous entries.
  static void clear(void);
};
  
};


#endif // __OBJECT_FACTORY_HH
