/*
Copyright (C) 2007, 2008  Condel project, TU-Berlin.
Written by Jan Sablatnig, Sven Grottke, Jiehua Chen.

This file is part of Adam, the DVE testbed.

Adam is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3, or (at your option) any later
version.

Adam is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Adam; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.
*/

#ifndef __POLAR_H
#define __POLAR_H

#include <iostream>

#include "vector.hh"

namespace Condel {

/// A two-dimensional vector with polar coordinates.
/// This is just a 2-d vector with some added support to read/write
/// polar values (phi, abs).  The True data is always the data in the
/// vector.
///
/// Definition of the angle phi:
/// <ul><li> phi=0 -> vector shows right (in direction of x-axis).
///     <li> phi increasing -> vector turns counterclockwise (mathematically
///          positive).
///     <li> Given in radians (one full turn is @f$ 2 \Pi @f$).
/// </ul>
/// So, phi((0,1)) := 0, phi((1,0)) := @f$ \frac{\Pi}{2} @f$
/// Values of phi as returned by this class are always normalized (i.e.
/// @f$ \varphi \in [0, 2\Pi[ @f$).

class cPolar : public cVector<2, double>
{
protected:
  // These two datums here are cashed.  The TRUE data is still the one
  // in cVector.
  double m_phi;
  double m_abs;           

private:
  // unimplemented, do NOT use these.
  const cVector<2, double>& assign(const cVector<2, double>& src);
  const cVector<2, double>& assign(const double& value);
  double& operator[](unsigned pos);
  const cVector<2, double>& add(const cVector<2, double>& src);
  const cVector<2, double>& subtract(const cVector<2, double>& src);
  const cVector<2, double>& scale(const double& value);

public:
  ///
  cPolar(void ) : cVector<2, double>(), m_phi(0.0), m_abs(0.0) { }
  cPolar(const cVector<2, double> &src) : cVector<2, double>(src),
        m_phi(0.0), m_abs(0.0) { update(); }
  ~cPolar(void) {}

  /// Update the local cache.
  /// This will update the cached phi and abs values to match the underlying
  /// vector data.
  ///
  /// Call this whenever the underlying values in the vector may have changed
  /// without going through the members of this class.
  void update(void);

  // assignments
  const cPolar &operator=(const cVector<2, double> &src) 
    { cVector<2,double>::operator=(src); update(); return *this; }

  // accessors
  // inherited: const double& operator[](unsigned pos) const
  const double& operator[](unsigned pos) const
    { return this->cVector<2,double>::operator[](pos); }
  const double &x(void) const { return (*this)[0]; }
  const double &y(void) const { return (*this)[1]; }
  const double &phi(void) const { return m_phi; }
  const double &abs(void) const { return m_abs; }
  /// Set a phi and an abs.
  /// Sets the vector accordingly.
  void setPhiAbs(double phi, double abs);
  /// Change just phi.
  /// This keeps abs as it was, changes the vector to adjust phi.
  void setPhi(double nphi) { setPhiAbs(nphi, abs()); }
  /// change just abs.
  /// This keeps phi as it was, changes the vector to adjust abs.
  void setAbs(double nabs) { setPhiAbs(phi(), nabs); }
  void setX(double x) { cVector<2,double>::operator[](0)= x; update(); }
  void setY(double y) { cVector<2,double>::operator[](1)= y; update(); }
  void setXY(double x, double y)
    {
    cVector<2,double>::operator[](0)= x; cVector<2,double>::operator[](1)= y;
    update();
    }

  /// Normalize an angle.
  /// The result will be between 0 and 2 M_PI.
  static double normalize_phi(double phi);

  
  // basic algebraic operations (addition, subtraction, scalar
  // multiplication...)
  const cPolar &operator+=(const cVector<2, double>& src)
    { this->cVector<2,double>::operator+= (src); update(); return *this; }
  const cPolar &operator-=(const cVector<2, double>& src)
    { this->cVector<2,double>::operator-= (src); update(); return *this; }
  const cPolar &operator*=(const double& value)
    { this->cVector<2,double>::operator*= (value); update(); return *this; }
  cPolar operator+(const cVector<2, double>& src) const
    { cPolar result(*this); result += src; return result; }
  cPolar operator-(const cVector<2, double>& src) const
    { cPolar result(*this); result -= src; return result; }
  cPolar operator*(const double& value) const
    { cPolar result(*this); result *= value; return result; }
  
  double operator*(const cVector<2, double>& src) const
    { return this->cVector<2,double>::operator*(src); }

};

};

cPolar operator*(double v, const cPolar &p); // { return p * v; }

std::ostream& operator<<(std::ostream& os, const cPolar &value);

                         
#endif // __POLAR_H
