/*
Copyright (C) 2007, 2008  Condel project, TU-Berlin.
Written by Jan Sablatnig, Sven Grottke, Jiehua Chen.

This file is part of Adam, the DVE testbed.

Adam is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3, or (at your option) any later
version.

Adam is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Adam; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.
*/

#ifndef __SCENARIO_H
#define __SCENARIO_H

#include "gen.hh"
#include "host.hh"

// forward references

namespace Condel {

// forward references
class cWorld;
class cConsistency;
class cLoadMeter;

/// The basic scenario interface.
/// This is the common base class of all scenario codes.
///
/// The scenario contains all the code that represent the particular game
/// etc. to be simulated.  This includes in-game rules, initialization code,
/// and the AI.
 
class cScenario : public cHostComponent
{
protected:
  double m_delayHint;           // how much local actions will be delayed

public:
  cScenario(cHost& host);
  virtual ~cScenario(void);

  /// Tell the AI about a local lag.
  /// Local lag is the delay between when an action is issued to when it
  /// is performed.  It is essential for the AI to preplan this.

  /// This is a cludge, the AI @em could find this out on its own.  Human
  /// users, for example, adjust to lag within seconds.
  /// But the information is known, so if it is also constant, it is
  /// easiest to pass it through here.
  void setDelayHint(double newhint)
        { m_delayHint= newhint; }

  /// Reset everything.
  /// This member is called by the host when the simulation is about to start.
  /// This does NOT initialize setDelayHint(), it's set before this call.
  virtual void initialize(void) { }
  
  /// Perform deterministic actions.
  /// This function performs all deterministic actions in the given world.
  /// In effect, it calls simulate() on all instances in world, then does
  /// all the things that happen in the scenario that are NOT done in the
  /// single instance's simulate function, e.g. collision detection.
  /// This is const, so side-effects can only happen in the instances.
  /// @param world The world that is advanced.
  /// @param newtime Time to advance the system to.  This should be in the
  /// future of the world's cWorld::simTime().  Given in seconds.
  /// @param load A tool to measure CPU and memory usage during the simulation.
  virtual void simulate(cWorld& world, double newtime, cLoadMeter &load)
        const= 0;

  /// Perform the AI.
  /// This function aims to supplant the user and therefore issues the
  /// nondeterministic actions a human user would cause.
  ///
  /// This will look at the main world (the visible one) available through
  /// the host only.
  /// @param newtime The current time.
  virtual void ai(double newtime) = 0;

  /// System Done.
  /// This member is called by the host when the simulation has ended.
  virtual void finish(void) =0;
};

};

#endif // __SCENARIO_H
