/*
Copyright (C) 2007, 2008  Condel project, TU-Berlin.
Written by Jan Sablatnig, Sven Grottke, Jiehua Chen.

This file is part of Adam, the DVE testbed.

Adam is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3, or (at your option) any later
version.

Adam is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Adam; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.
*/

#ifndef __SLIDING_WINDOW_H
#define __SLIDING_WINDOW_H
#include <list>

#include "gen.hh"
#include "vector.hh"

namespace Condel {

/// Gather statistical data in a sliding window.
/// This window will give you moving averages over a specified length.
/// Note the window-size will be smaller as long as the measurement-time
/// is below windowSizeTime.
/// @todo This should inherit from OMNeT++'s cStatistic some day

class cSlidingWindow {
private:
  const double m_windowSize;    // size of the sliding window, in s

  list<R_2> m_buffer;           // used for storing data that's still inside
                                // the window

  double m_lastTime;            // time of last entry
  double m_lastVal;             // value of last entry
  double m_start;               // time of start
  double m_sum;                 // for calculating the mean
  double m_total;               // total sum
  double m_max;                 // for calculating the maximum
  bool m_exclusive;             // if true, use all data before, excluding
                                // the current, otherwise include the new
                                // data.

public:
  /// Construct a new sliding window.
  /// This window will supply you with moving averages over the specified
  /// length.  The applicable data is the cost of any ongoing process, such
  /// as bytes sent out over a network or energy units used.
  /// Start is the time when the window is created.
  /// @param windowSizeTime The length of the measuring window in seconds.
  /// Note the window-size will be smaller as long as the measurement-time
  /// is below windowSizeTime.
  /// @param currentTime The start time of this measuring window.
  /// @param exclusive If true, the data that will be given via collect()
  /// applies to AFTER the time given in collect().  If false, the data applies
  /// to BEFORE the time given in collect().  Use the true for data you're
  /// about to send off, false for data already received.
  cSlidingWindow(double windowSizeTime, double currentTime=0.0,
        bool exclusive=true);

  virtual ~cSlidingWindow(void) {}

  /// Reset everything.
  /// @param start The new starting time, in seconds.
  void initialize(double start);

  /// Gather data.
  /// @param time The time this data applies to.  The time @em must be
  /// non-decreasing with previous calls (and also with calls to getAverage()).
  /// Also, time must be >= 0.0.  @sa exclusive in cSlidingWindow()
  /// @param value The data to be gathered.  No interpretation here, you
  /// may use this object for anything you like.
  void collect(double time, double value);

  /// Get results gathered so far.
  /// @param time The time for which to calculate the moving average.
  /// The time @em must be non-decreasing with previous calls (and also with
  /// calls to collect()).  Also, time must be >= 0.  If time is not given,
  /// the time of the latest call to either collect or getMean() is used.
  /// @return A unit per second average of all data from time -windowSize to
  /// time.
  double getAverage(double time=-1.0);

  /// Get the highest measured result so far.
  /// @return The maximum calculated average.  An average is calculated for
  /// every call to collect() and to getAverage().  Note this is not the
  /// maximum of the collected data.  Note that because the window-size is
  /// smaller in the beginning, the first measurements would be more
  /// erratic and therefore have a tendency to dominate this value here.
  /// For this reason, this value is -infinity until time>=windowSize.
  double getMax(double time=-1.0);

  /// Get the average over all time.
  /// @param time The current time.
  /// @return A unit per second measurement running over all data ever given.
  double getTotalAverage(double time=-1.0);

private:
  void update_to(double time);
};
  
};

#endif // __SLIDING_WINDOW_H
