/*
Copyright (C) 2007, 2008  Condel project, TU-Berlin.
Written by Jan Sablatnig, Sven Grottke, Jiehua Chen.

This file is part of Adam, the DVE testbed.

Adam is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3, or (at your option) any later
version.

Adam is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Adam; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.
*/
#include "variance_meter.hh"
#include "consistency.hh"
#include "instance.hh"
#include "scenario.hh"
#include "host.hh"

cVarianceMeter::cVarianceMeter(void)
: m_hosts(), m_attributeWeight(), m_attributeSumL2(), m_attributeSumL1(),
  m_pAttributeNorm(NULL), m_count(0), m_sumL2(0.0), m_sumL1(0.0)
{ }

void cVarianceMeter::addHost(const cHost* pHost) {
  assert(pHost);
  m_hosts.push_back(pHost);
}

void cVarianceMeter::addAttribute(const ATTRIBUTE_ID& attribute, double weight) {
#if DEBUG >= 1 
  // Make sure the attribute wasn't added before
  if (m_attributeWeight.find(attribute) != m_attributeWeight.end())
    condel_error("attribute weight registered multiple times");
#endif

  m_attributeWeight[attribute]= weight;
}

void cVarianceMeter::initWeights(void)
{
  map<const ATTRIBUTE_ID, double>::iterator w;
  double sum= 0.0;

  for (w= m_attributeWeight.begin(); w != m_attributeWeight.end(); w++)
    sum += (*w).second;
  for (w= m_attributeWeight.begin(); w != m_attributeWeight.end(); w++)
    (*w).second /= sum;
}

struct varstats
{
  double moment_1;
  double moment_2;
  double norm;
  ATTRIBUTE_ID aid ;

varstats(void)
  : moment_1(0.0), moment_2(0.0), aid(0)
  { }
};


double cVarianceMeter::computeConsistency(void)
{
  double l2_norm= 0.0;
  double l1_norm= 0.0;

  bool initialized;
  scoped_array<struct varstats> stats;
  unsigned long size= 0;	// unnecessary init, for compiler
  unsigned long i=0, j;		// unnecessary init, for compiler

  const map <const string, OBJECT_ID>& om(cObjectFactory::objectMap());
  map<const string, OBJECT_ID>::const_iterator object_it;
  for (object_it = om.begin(); object_it != om.end(); object_it++) {
    OBJECT_ID object = (*object_it).second;
    // init statistics
    initialized = false;
 
    list<const cHost*>::const_iterator hostIter;
    for (hostIter = m_hosts.begin(); hostIter != m_hosts.end(); hostIter++) {
      const cConsistency& cons =
        (*hostIter)->getConsistency();
      if (cons.isReplicated(object)) {
	const cAttributeSet atts(cons.getInstance(object).getAttributes());

	if (!initialized) {
          // first time object is found
	  size = atts.size();
	  stats.reset(new struct varstats[atts.size()]);

	  cAttributeSet::const_iterator at;
	  j = 0;
	  for (at = atts.begin(); at != atts.end(); at++, j++)
            {
	    stats[j].aid= (*at).second.getID();

#           if DEBUG >=2
              assert (m_pAttributeNorm->find(stats[j].aid) != m_pAttributeNorm->end());
#           endif
            stats[j].norm= (*m_pAttributeNorm->find(stats[j].aid)).second;
            }
	  i = 0;
	  initialized = true;
        }

	cAttributeSet::const_iterator at;
	int j = 0;
	for (at = atts.begin(); at != atts.end(); at++, j++)
          {
          const double v= (*at).second.getValue() * stats[j].norm;
	  stats[j].moment_1 += v;
	  stats[j].moment_2 += v * v;
          }
	i++;
      }
      
    }

    // evaluate statistics
    if (i > 1) {
      for (j = 0; j < size; j++)
	{
        if (stats[j].moment_1 == 0.0)
          continue;
        double l1_snip = stats[j].moment_1 /i;
        double c_j_t = stats[j].moment_2 /i - l1_snip * l1_snip;
        double sq_j_t = 0.0;
        if (c_j_t > 0.0)
          sq_j_t = sqrt(c_j_t);
	double r= (*m_attributeWeight.find(stats[j].aid)).second;

#if VERBOSE >= 3
	if (c_j_t > 0.001)
	    cout << "EYE: " << stats[j].aid << " -> " << c_j_t << endl;
#endif

	l2_norm += r * c_j_t;
	l1_norm += r * sq_j_t;

        if (m_attributeSumL2.find(stats[j].aid) == m_attributeSumL2.end())
            m_attributeSumL2[stats[j].aid] = 0.0;
        if (m_attributeSumL1.find(stats[j].aid) == m_attributeSumL1.end())
            m_attributeSumL1[stats[j].aid] = 0.0;

        m_attributeSumL2[stats[j].aid] += c_j_t;
        m_attributeSumL1[stats[j].aid] += sq_j_t;
        }
    }
    
  }
  

  m_count ++;
  m_sumL2 += l2_norm;
  m_sumL1 += l1_norm;
  return l2_norm;
}

void cVarianceMeter::init(map<const ATTRIBUTE_ID, double> &attributeNorm)
{
  m_pAttributeNorm= &attributeNorm;

  m_count=0;
  m_sumL2=0.0;
  m_sumL1=0.0;

  m_hosts.clear();

  m_attributeWeight.clear();

  m_attributeSumL2.clear();

  m_attributeSumL1.clear();
}

double cVarianceMeter::averageConsistency2(ATTRIBUTE_ID aid) const
{
  map<const ATTRIBUTE_ID, double>::const_iterator it(m_attributeSumL2.find(aid));

  if (it == m_attributeSumL2.end() || !m_count)
    return 0.0;

  return (*it).second /m_count;
}

double cVarianceMeter::averageConsistency2(void) const
{
  if (!m_count)
    return 0.0;

  return m_sumL2 /m_count;
}


double cVarianceMeter::averageConsistency1(ATTRIBUTE_ID aid) const
{
  map<const ATTRIBUTE_ID, double>::const_iterator it(m_attributeSumL1.find(aid));

  if (it == m_attributeSumL1.end() || !m_count)
    return 0.0;

  return (*it).second /m_count;
}

double cVarianceMeter::averageConsistency1(void) const
{
  if (!m_count)
    return 0.0;

  return m_sumL1 /m_count;

}




