/*
Copyright (C) 2007, 2008  Condel project, TU-Berlin.
Written by Jan Sablatnig, Sven Grottke, Jiehua Chen.

This file is part of Adam, the DVE testbed.

Adam is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3, or (at your option) any later
version.

Adam is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Adam; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.
*/

#ifndef __VARIANCE_METER_H
#define __VARIANCE_METER_H

#include <list>

#include "gen.hh"
#include "object_factory.hh"
#include "attribute_factory.hh"

namespace Condel {

class cHost;

/// Computes the divergence measure D.
/// This tool is activated by cEye, which ticks just after every normal tick.
/// It traverses all hosts and measures the std deviation of the attributes
/// of corresponding instances.
///
/// Differences can affect position as well as speeds, even angles
/// and accelerations.  Since these are perceived differently by human
/// observers, each attribute needs to be weighted.
/// @sa cDiscoMeter


class cVarianceMeter
{
private:
  list<const cHost*> m_hosts;          // the list of hosts monitored
  map<const ATTRIBUTE_ID, double> m_attributeWeight;  // list of attributes monitored
  map<const ATTRIBUTE_ID, double> m_attributeSumL2;
  map<const ATTRIBUTE_ID, double> m_attributeSumL1;
  map<const ATTRIBUTE_ID, double> *m_pAttributeNorm;

  unsigned long m_count;
  double m_sumL2;
  double m_sumL1;

public:
  cVarianceMeter(void);
  virtual ~cVarianceMeter(void) {}

  /// Add a new host to the watchlist.
  /// @param pHost The new host.
  void addHost(const cHost* pHost);

  /// Add a new attribute to the watchlist.
  /// @param attribute The ID of the new attribute.
  /// @param weight The relative importance of the attribute to the overall
  ///   perception of disontinuity.  In definitionen.pdf, this would be
  ///   @f$ w_a @f$.
  void addAttribute(const ATTRIBUTE_ID& attribute, double weight);

  /// Run this once after all attributes have been added.
  /// Only call this ONCE.  This fixes the weights to have
  /// @f$ \sum w_a == 1 @f$.
  void initWeights(void);

  
  /// Tally the system divergence.
  /// Compute the entire system's current consistency as defined by the
  /// divergence metric.   @f$ v := \sum (x_i - \bar x)^2 / N @f$,
  ///    @f$\sigma  := \sqrt(v) @f$
  double computeConsistency(void);

  /// Reinitialize this tool.
  /// After this call, the tool looks like right after its creation.
  /// @param attributeNorm A list of attribute NORMs to use for the attributes.
  ///   This list is NOT copied, the original version specified here is used.
  ///   The norms are independant from the weights, they are set by the
  ///   scenario so that by multiplying it with its attribute, the attribute
  ///   will have a natural domain of +/- 1.  This does not mean the product
  ///   cannot exceed this range, btw.  An example would be the x position,
  ///   where the weight is 1/width.  Another example is score, where weight
  ///   is 1.
  void init(map<const ATTRIBUTE_ID, double> &attributeNorm);


  /// Get the arithmetic average so far.
  /// @return The average of the total variances calculated each frame.
  double averageConsistency2(void) const;
  /// Get the arithmetic average so far.
  /// @param aid An attribute.  Note this attribute may exist in several
  /// different objects.
  /// @return The average of the variance calculated for the specified
  ///   attribute each frame.
  double averageConsistency2(ATTRIBUTE_ID aid) const;

  /// Get the arithmetic average so far.
  /// @return The average of the total standard deviation calculated each frame.
  double averageConsistency1(void) const;
  /// Get the arithmetic average so far.
  /// @param aid An attribute.  Note this attribute may exist in several
  /// different objects.
  /// @return The average of the standard deviation calculated for the specified
  ///   attribute each frame.
  double averageConsistency1(ATTRIBUTE_ID aid) const;
};
  
};

#endif // __VARIANCE_METER_H
