/*
Copyright (C) 2007, 2008  Condel project, TU-Berlin.
Written by Sven Grottke, Jan Sablatnig, Jiehua Chen.

This file is part of Adam, the DVE testbed.

Adam is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3, or (at your option) any later
version.

Adam is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Adam; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.
*/

#ifndef __WORLD_H
#define __WORLD_H

#include <map>

#include "gen.hh"
#include "object_factory.hh"
#include "host.hh"

namespace Condel {

// forward references
class cInstance;
class cLoadMeter;

/// A set of instances.
/// A world holds a collection of instances, each with a unique object ID.
/// Note this is A world, not THE world.  IOW, it's one of the views
/// of the world a single host can have.  The collection of all worlds
/// is called the universe.
/// The universe is also entirely different from the single globally correct
/// world (if there is such a thing, not true in some consistencies).

class cWorld : public map<const OBJECT_ID, shared_ptr<cInstance> > ,
                     public cHostComponent
{
protected:
  string m_name;    // used for debugging purposes
  double m_simtime;

public:
  cWorld(cHost &host);
  cWorld(cHost &host, const string& name);
  virtual ~cWorld(void) {}

  /// Create a deep clone of this world.
  /// The world will have all the instances of the original, with the
  /// same state, etc.  In fact, aside from the address, the new world
  /// will be indistinguishable from the old one.
  ///
  /// Note this is a very slow process.
  /// @return A pointer to a newly created copy of this world.  Assign this
  /// to an auto_ptr<cWorld> to avoid any possibility of leakage.
  virtual auto_ptr<cWorld> clone(void);

  /// A DEEP assignment.
  /// This copies everything over to this world except for the name and
  /// the host reference.  Because of this limitation, do not use this
  /// for full initialization, only to copy state.
  /// @param ow Another world on the same host.
  const cWorld &operator=(const cWorld& ow);

  /// Get the name of this world.
  /// @return An arbitrary name, used only for debugging.
  const string& getName(void) const { return m_name; }

  /// Get the world's simulation time.
  /// Note the only time there is a difference between this and any
  /// instance in this world is during a simulate() run.
  /// @return The simulation time that all instances in the world are at,
  /// or were at if simulate() is currently running.
  double simTime(void) const {  return m_simtime; }

  /// Set the world's simulation time.
  /// After all instances have been updated to a new time, set the world's
  /// time to the same value via this call
  /// @param newtime The new simulation time, in seconds.
  void setSimTime(double newtime) {  m_simtime= newtime; }
  
  /// Find out if an object has an instance in this world.
  bool isReplicated(const OBJECT_ID& objectID) const;

  /// Get an instance.
  /// @param objectID The object ID of an object that @em does have an
  /// instance in this world.  To ensure this, use isReplicated() before
  /// this call.
  /// @return A reference to the actual instance, not a copy.
  cInstance& getInstance(const OBJECT_ID& objectID);
  /// Get an instance.
  /// @param objectID The object ID of an object that @em does have an
  /// instance in this world.  To ensure this, use isReplicated() before
  /// this call.
  /// @return A reference to the actual instance, not a copy.
  const cInstance& getInstance(const OBJECT_ID& objectID) const;

  /// Insert a new instance.
  /// @param pInstance A pointer to an instance to be added.  An instance
  /// of the same object must not yet exist in this world.  The instance
  /// is cloned during the call; the caller maintains ownership of the
  /// passed instance.
  /// @todo should use a reference, not a pointer.
  void addInstance(const cInstance* pInstance);

  /// Perform deterministic actions.
  /// This function performs all deterministic actions in this world.
  /// This usually calls cScenario::simulate().
  /// @param newtime Time to advance the system to.  This should be in the
  /// future of simTime().  Given in seconds.
  /// @param load A tool to measure CPU and memory usage during the simulation.
  virtual void simulate(double newtime, cLoadMeter &load);

  /// Estimate the entire in-memory size of this world.
  /// @return The # of bytes necessary to keep this entire world in memory.
  /// @sa cInstance::getCodedLength()
  unsigned long long getTotalLength(void);
  
  /// An iterator.
  typedef map<const OBJECT_ID, shared_ptr<cInstance> >::iterator iterator;
  /// An iterator.
  typedef map<const OBJECT_ID, shared_ptr<cInstance> >::const_iterator
  const_iterator;  


  /// Reset everything.
  /// This member is called by the host when the simulation is about to start.
  virtual void initialize(void);

  /// System Done.
  /// This member is called by the host when the simulation has ended.
  virtual void finish(void) {}
};
  

};

#endif // __WORLD_H
