/*
Copyright (C) 2007, 2008  Condel project, TU-Berlin.
Written by Jan Sablatnig, Sven Grottke, Jiehua Chen.

This file is part of Adam, the DVE testbed.

Adam is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3, or (at your option) any later
version.

Adam is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Adam; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.
*/
#include <omnetpp.h>
#include "message_router.hh"
#include "consistency.hh"
#include "msg.hh"
#include "host.hh"

#include <iomanip>
#include <map>


// When Message batching is used, only aggregate until this max length
// is reached (in bytes).
#define BATCHING_MTU 1400



namespace Condel {

class cSendout
{
protected:
  cMsg m_msg;           // The message
  simtime_t m_deadline; // Latest time this message should be sent out

public:
  cSendout(void) : m_msg(), m_deadline(0.0) { }
  cSendout(const cSendout &other)
    : m_msg(other.m_msg), m_deadline(other.m_deadline) { }
  cSendout(cMsg &msg, simtime_t latest) : m_msg(msg), m_deadline(latest) { }
  ~cSendout(void) { }
  
  const cMsg &getMsg(void) const { return m_msg; }
  simtime_t getDeadline(void) const { return m_deadline; }
  void setDeadline(simtime_t latest) { m_deadline= latest; }
};

class cBMonitor: public map<unsigned long, cSendout>
{
public:
  cBMonitor(void): map<unsigned long, cSendout>() { }
  ~cBMonitor(void) { }
};

class cTimeout
{
protected:
  simtime_t m_next;             // next timeout approaching
  bool m_sent;                  // Was this one sent out already?

public:
  cTimeout(void) : m_next(0.0), m_sent(false) { }

  cTimeout(simtime_t nex, bool sen= false)
    : m_next(nex), m_sent(sen)
    { }
  ~cTimeout(void) { }

  simtime_t getNext(void) const { return m_next; }
  void setNext(simtime_t nex) { m_next= nex; }
  simtime_t getSent(void) const { return m_sent; }
  void setSent(bool par= true) { m_sent= par; }
};

class cMsgRoute
{
public:
  cMsg m_msg;                   // the message
  unsigned long m_target;       // ID of target host

public:
  cMsgRoute(void) : m_msg(), m_target(0) { }
  cMsgRoute(const cMsgRoute &other)
    : m_msg(other.m_msg), m_target(other.m_target) { }
  cMsgRoute(cMsg &msg, unsigned long target) : m_msg(msg), m_target(target) { }
  ~cMsgRoute(void) { }

  const cMsg &getMsg(void) const { return m_msg; }
  unsigned long getTarget(void) const { return m_target; }

  bool operator<(const cMsgRoute &other) const
        { return getMsg() < other.getMsg() ||
          (getMsg() == other.getMsg() && getTarget() < other.getTarget()); }
};


class cSRMonitor: public map<cMsgRoute, cTimeout>
{
public:
  cSRMonitor(void): map<cMsgRoute, cTimeout>() { }
  ~cSRMonitor(void) { }
};


}


// On Batching & ARQ:
// Two possibilities exist:
// Option I:  Batch, THEN ARQ
//   + Only a single ACC for a batched package
//   - Must send batched package, but then filter out non-reliables (like ACKS)
//   - Can run into trouble if the head-package is a non-reliable
//   - Can run into trouble if the head-package is an ACK (weird id)
// Option II:  ARQ, THEN Batch
//   + Can resend just the packages with aggressive timeout
//   - timeout difficult to estimate (unless hooking into the batch-sendoff)
//   - More bandwith (many ACKS, all batched)

// For now, we're installing option II, mostly because it's easier/more logical
// to do.

void cMessageRouter::send_out(unsigned long id)
{
  {
  map<unsigned long, cSendout>::iterator it(m_pBatching->find(id));
  if (it != m_pBatching->end())
    {
    host().sendOut((cMsg *) (*it).second.getMsg().dup(), id);
    m_pBatching->erase(it);
    }
  }

  {
  // Find all unsent messages and fix their timeouts to the correct time!
  map<cMsgRoute, cTimeout>::iterator it;
  for (it= m_pARQ->begin(); it != m_pARQ->end(); it++)
    {
    if ((*it).first.getTarget() == id && !(*it).second.getSent())
      {
      double batching= (*it).first.getMsg().getBatching();
      if (batching< 0.0)
        batching= 0.0;
      (*it).second.setNext(host().now() + (*it).first.getMsg().getTimeOut()
                        + batching /2);
      (*it).second.setSent();
      }
    }
  }
}


void cMessageRouter::send_out(cMsg *pMsg, unsigned long id)
{
  scoped_ptr<cMsg> lm(pMsg);
  m_msgSent++;

  // is it to be sent reliably?  If so, keep a copy.
  if (lm->getTimeOut() >0.0)
    {
    m_msgReliable++;
    cMsgRoute mr(*lm, id);

    // Batching itself causes a delay:
    // If batching is sparse, the extended T/O is + batching.
    // If batching is crowded, the extended T/O is + batching/2.
    // We'll be optimistic here.

    // Of course the TOTAL T/O increases by twice the delay because the
    // ACK will be batched as well!
    double batching= lm->getBatching();
    if (batching< 0.0)
      batching= 0.0;
    cTimeout to(host().now() +lm->getTimeOut() +batching);
    (*m_pARQ)[mr] = to;
    }


  // Batching:
  // Any other messages waiting for that same host-id?
  map<unsigned long, cSendout>::iterator it(m_pBatching->find(id));
  if (it != m_pBatching->end())
    {
    // Possible to add current one to the batch?
    if ((*it).second.getMsg().byteLength() + lm->byteLength() > BATCHING_MTU)
      {
      // Otherwise, send out the old batch, start a new batch
      send_out(id);
      it = m_pBatching->end();
      }
    }

  double batching= lm->getBatching();
  double mydead= host().now() + batching;
  if (it != m_pBatching->end())
    {
    // OK, add the batch to current
    if (mydead > (*it).second.getDeadline())
      mydead= (*it).second.getDeadline();

    lm->encapsulate((cMsg *) (*it).second.getMsg().dup());
    m_pBatching->erase(it);
    }

  // Put current into the batch-buffer
  cSendout so(*lm, mydead);
//  m_pBatching->insert(pair<unsigned long, cSendout>(id,so));
  (*m_pBatching)[id] = so;

  // Is the current message urgent?
  if (batching <0.0)
    send_out(id);
}





cMessageRouter::cMessageRouter(cHost &Host)
: cRouter(Host), m_pARQ(new cSRMonitor()), m_pBatching(new cBMonitor()),
    m_handling(0), m_receptionDelay(0.0),
    m_msgSent(0), m_msgReliable(0), m_msgResent(0), m_ACKs(0)
{
}

cMessageRouter::~cMessageRouter(void)
{ }

void cMessageRouter::initialize(void)
{
  if (host().hasPar("reception_delay"))
    m_receptionDelay= host().par("reception_delay");
  m_handling= 0;
  m_pARQ->clear();
  m_pBatching->clear();
  m_msgSent= m_msgReliable= m_msgResent= m_ACKs= 0;
  cRouter::initialize();
}

void cMessageRouter::send(cMsg* pMsg)
{
  host().scheduleAt(host().now(), (cMessage *) pMsg->dup());

  list<unsigned long>::const_iterator hit;
  for(hit= host().getRouter().beginHID(); hit!= host().getRouter().endHID();
      hit++)
    if (*hit != (unsigned long) host().id())
      send_out((cMsg *) pMsg->dup(), *hit);
}

void cMessageRouter::handleMessage(cMsg* pMsg)
{
  m_handling++;

  // Handle Aggregated messages
  scoped_ptr<cMessage> em(pMsg->decapsulate());
  if (em)
    {
    cMsg *msg= check_and_cast<cMsg *>(em.get());
    // Make it look like it was sent, not embedded
    msg->setSentFrom(pMsg->senderModule(), pMsg->senderGateId(),
                                                pMsg->sendingTime());
    handleMessage(msg);
    }

  // Don't use selfMessage, it's confused by my hack with setSentFrom!
  if (pMsg->senderModuleId() != host().id())
    {
    // Handle ACK-messages.
    if (pMsg->getType() == cMsg::ACK)
      {
      cMsgRoute mr(*pMsg, pMsg->senderModuleId());
      m_pARQ->erase(mr);
      if (!--m_handling)
        handleCheck();
      return;
      }

    // Send ACK-messages.
    if (pMsg->getTimeOut() >0.0)
      {
      cMsg *ack= new cMsg(pMsg->getHostID(), pMsg->getOrder(), 0.0,
                        pMsg->getBatching(), cMsg::ACK, "ack");
      ack->setTimestamp(pMsg->timestamp());

      send_out(ack, pMsg->senderModuleId());
      m_ACKs++;
      }

    // message reception delay
    double minhorizon= pMsg->sendingTime() +m_receptionDelay;
    if (minhorizon > host().simTime())
      {
      host().scheduleAt(minhorizon, (cMessage *) pMsg->dup());
      return;
      }
    }

  host().getConsistency().handleMessage(pMsg);

  if (!--m_handling)
    handleCheck();
}

void cMessageRouter::handleCheck(void)
{
  simtime_t now= host().now();

  {
  // check if the sendout is run out on any as yet unsent messages, if so, send.
  map<unsigned long, cSendout>::iterator it;
  bool rerun;
  do { rerun= false;
  for (it= m_pBatching->begin(); it != m_pBatching->end(); it++)
    {
    if (now >= (*it).second.getDeadline())
      {
      send_out((*it).first);
      rerun= true; break;                    // REALLY stupid
      }
    }
  } while (rerun); 
  }

  {
  // check if the timeout is run out on any sent messages, if so, send again.
  map<cMsgRoute, cTimeout>::iterator it;
  bool rerun;
  do { rerun= false;
  for (it= m_pARQ->begin(); it != m_pARQ->end(); it++)
    {
    if (now >= (*it).second.getNext() && (*it).second.getSent())
      {
#     if VERBOSE >= 3
      cout << setiosflags(ios::fixed) << setprecision(2)
           << "Retransmitting timed out message from "
           <<(*it).first.getMsg().timestamp() << "@"
           <<(*it).first.getMsg().getHostID()
           << "[" << (*it).first.getMsg().getOrder() << "], now " << now <<endl;
#     endif
      cMsg *resend = (cMsg *) (*it).first.getMsg().dup();
      // Turn OFF batching now, it's an emergency
      resend->setBatching(-1.0);
      unsigned long t= (*it).first.getTarget();

      m_pARQ->erase(it);

      send_out(resend, t);
      m_msgResent++;
      m_msgReliable--;       // Avoid counting this message twice for reliable%
      rerun= true; break;               // REALLY stupid
      }
    }
  } while (rerun);
  }

  cRouter::handleCheck();
}

void cMessageRouter::finish(void)
{
  host().recordScalar("% of messages sent reliably",
                             100.0 * m_msgReliable / (m_msgSent-m_ACKs-m_msgResent));
  // Note this % can exceed 100%...
  host().recordScalar("% of reliable messages resent",
                             100.0 * m_msgResent / m_msgReliable);
  host().recordScalar("% of total messages were ACKs",
                             100.0 * m_ACKs / m_msgSent);
}

