/*
Copyright (C) 2007, 2008  Condel project, TU-Berlin.
Written by Jan Sablatnig, Sven Grottke, Jiehua Chen.

This file is part of Adam, the DVE testbed.

Adam is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3, or (at your option) any later
version.

Adam is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Adam; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.
*/

#ifndef __MESSAGE_ROUTER_H
#define __MESSAGE_ROUTER_H

#include <cstat.h>

#include "gen.hh"
#include "router.hh"


namespace Condel {

// forward references
class cSRMonitor;        // in .cc file, holds messages for selective repeat
class cBMonitor;         // in .cc file, holds messages for batching

/// The network layer for a flat network model.
/// This works for total replication.
///
/// The class supports batching, which is a message aggregation based on
/// time.  If a message needs to be sent to a host, it may be delayed for
/// a short time (the batching parameter).  If other messages occur during
/// that time that are destined for the same host, they are combined with
/// the first message.
///
/// The class supports a selective repeat ARQ.  If the timeout parameter
/// is set in a message, that message is copied before sending it off.
/// If such a message arrives at any host, it will be acknowledged (by
/// sending an ACK message back to the original host).  If the ACK message
/// does not arrive within the timeout, that message will be resent.
///
/// On Batching & ARQ:
/// Two possibilities exist:
/// <ol><li>Option I:  Batch, THEN ARQ
///   <ul><li> + Only a single ACC for a batched package
///       <li> - Must send batched package, but then filter out non-reliables
///              (like ACKS)
///       <li> - Can run into trouble if the head-package is a non-reliable
///       <li> - Can run into trouble if the head-package is an ACK (weird id)
///   </ul>
///     <li>Option II:  ARQ, THEN Batch
///   <ul><li> + Can resend just the packages with aggressive timeout
///       <li> - timeout difficult to estimate (unless hooking into the
///              batch-sendoff)
///       <li> - More bandwith (many ACKS, all batched)
///    </ul></ol>
///
/// For now, we're installing option II, mostly because it's easier/more logical
/// to do.
///
/// Note on the batching scheme:
/// <ul><li> First message is the LAST message (timewise) to be batched.
///     <li> The encapsulated message is the one before (timewise).
///     <li> Last encapsulated message is the first one (timewise).
/// </ul>



class cMessageRouter: public cRouter
{
private:
  scoped_ptr<cSRMonitor> m_pARQ;     // messages for which I'm still waiting for
                                     // an ACK.  Implements selective repeat.
  scoped_ptr<cBMonitor> m_pBatching; // message-batch for which I'm still
                                     // waiting to send off.



  int m_handling;       // When != 0, the message router is in handleMessage()
  
  // Minimum time a message is aged before processing it.
  double m_receptionDelay;

  // These are statistics, ALL taken before batching.
  // Look at host for bytes actually sent.
  unsigned long m_msgSent;      // messages sent by application
  unsigned long m_msgReliable;  // messages sent by application reliably
  unsigned long m_msgResent;    // messages resent because of T/O, TOTAL
  unsigned long m_ACKs;         // ACKs sent
  
  // REALLY low level send.  Sends off the queue waiting to go to that host.
  // id is the host this goes to.
  void send_out(unsigned long id);

  // Lowlevel send, possibly performing arq&batching, then passing it over to
  // the host.  Message belongs to this function, forget the pointer.
  // id is the host this goes to.
  void send_out(cMsg *pMsg, unsigned long id);
  
public:
  cMessageRouter(cHost &host);
  virtual ~cMessageRouter(void);

  virtual void initialize(void);

  /// Send a message to the network.
  /// The message is sent to each known host.
  /// @param pMsg A pointer to the message.  This message will be duped during
  /// this call, the original caller retains ownership of the supplied
  /// msg, so you may delete it after the call.
  virtual void send(cMsg* pMsg);

  /// Receive a message from the network.
  /// This deaggregates messages, handles ACK messages, and implements a
  /// reception delay (by rescheduling early messages). Finally hands off
  /// the message to the consistency.
  /// This is called by the host (i.e., a subclass of cHost).
  /// @param pMsg A pointer to the message.  The host (i.e. the caller of this
  /// function) retains ownership of the message, do NOT delete it.
  /// If you need your own copy, you'll have to dup() it.
  virtual void handleMessage(cMsg* pMsg);

  // Not overridden
  // virtual void handleTick(void);

  /// Called each check-interval.
  /// This handles timeouts both on messages waiting to be batched and
  /// on messages waiting to be ACKed in the ARQ.
  /// Called regularly (each cHost::CHECK_TIME) by the host.
  virtual void handleCheck(void);

  /// System Done.
  /// Prints out some statistics on % of reliable messages, etc.
  /// This member is called by the host when the simulation has ended.
  virtual void finish(void);
};

  
};

#endif // __MESSAGE_ROUTER_H
