/*
Copyright (C) 2007, 2008  Condel project, TU-Berlin.
Written by Jan Sablatnig, Sven Grottke, Jiehua Chen.

This file is part of Adam, the DVE testbed.

Adam is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3, or (at your option) any later
version.

Adam is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Adam; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.
*/
#include <math.h>

#include "ball.hh"
#include "instance_factory.hh"
#include "pong_scenario.hh"



ATTRIBUTE_ID cBall::s_positionXID = cAttributeFactory::registerAttribute("cBall::positionX");
ATTRIBUTE_ID cBall::s_positionYID = cAttributeFactory::registerAttribute("cBall::positionY");
ATTRIBUTE_ID cBall::s_velocityXID = cAttributeFactory::registerAttribute("cBall::velocityX");
ATTRIBUTE_ID cBall::s_velocityYID = cAttributeFactory::registerAttribute("cBall::velocityY");
ATTRIBUTE_ID cBall::s_seedID = cAttributeFactory::registerAttribute("cBall::seed");




const double BALL_SPEED_Y_START_RANGE_MIN = -200;
const double BALL_SPEED_Y_START_RANGE_MAX =  200;


// absolutely, v_ball is forced into these ranges
// This is enforced whenever SetVelocity() (or randomYSpeed()) is called.
const double BALL_SPEED_Y_MIN = 10;    // arbitrary value > 0
const double BALL_SPEED_Y_MAX = 4000;  // (BOTTOM_BORDER-TOP_BORDER) /2 /TICK
const double BALL_SPEED_X_MIN = 10;    // not used
const double BALL_SPEED_X_MAX = 8000;  // (RIGHT_BORDER-LEFT_BORDER) /2 /TICK


// TODO:  I would like to use some sort of randomness to prevent these
//        limits from causing a static case.
//        Since we cannot at this time use probability deterministically,though,
//        this is not easy at all.
// UPDATE:  This would be doable with this current new design...
static double check_ball_y(double in)
{
  if (fabs(in) < BALL_SPEED_Y_MIN)
    in= copysign(BALL_SPEED_Y_MIN, in);
  if (fabs(in) > BALL_SPEED_Y_MAX)
    in= copysign(BALL_SPEED_Y_MAX, in);

  return in;
}

static double check_ball_x(double in)
{
  if (fabs(in) > BALL_SPEED_X_MAX)
    in= copysign(BALL_SPEED_X_MAX, in);

  return in;
}


cBall::cBall(const OBJECT_ID& id, const string& name)
: cInstance(id, name, 5), m_seed(intuniform(INT_MIN, INT_MAX) -INT_MIN)
{ }


void cBall::setVelocity(const R_2& v)
{
  m_velocity[0]= check_ball_x(v[0]);
  m_velocity[1]= check_ball_y(v[1]);
}


void cBall::randomYSpeed(void)
{
   double random= ((double) rand_r(&m_seed)) / RAND_MAX;
   
   m_velocity[1]= check_ball_y(random *
        (BALL_SPEED_Y_START_RANGE_MAX -BALL_SPEED_Y_START_RANGE_MIN)
        +BALL_SPEED_Y_START_RANGE_MIN);
}

cAttributeSet cBall::getAttributes() const
{
  cAttributeSet ret;

  ret.add(cAttribute(s_positionXID, getPosition()[0]));
  ret.add(cAttribute(s_positionYID, getPosition()[1]));
  ret.add(cAttribute(s_velocityXID, getVelocity()[0]));
  ret.add(cAttribute(s_velocityYID, getVelocity()[1]));
  ret.add(cAttribute(s_seedID, (double) (long) m_seed));

  return ret;
}


void cBall::setAttributes(const cAttributeSet& na)
{
  cAttributeSet::const_iterator iter;

  R_2 oldpos = getPosition();
  R_2 oldvel = getVelocity();

  for (iter = na.begin(); iter != na.end(); iter++)
    {
    const cAttribute& attr = (*iter).second;
    
    if (attr.getID() == s_positionXID) oldpos[0]= attr.getValue();
    else if (attr.getID() == s_positionYID) oldpos[1]= attr.getValue();
    else if (attr.getID() == s_velocityXID) oldvel[0]= attr.getValue();
    else if (attr.getID() == s_velocityYID) oldvel[1]= attr.getValue();
    else if (attr.getID() == s_seedID) m_seed= (unsigned long)
                                                  lrint(attr.getValue());
    else assert(false);		// unknown attribute found
    }

  setPosition(oldpos);
  setVelocity(oldvel);
}


cBall* cBall::clone() const {
  cBall* pNew = new cBall(getObjectID(), getName());

  pNew->setSimTime(simTime());
  pNew->setUpdateTime(updateTime());

  pNew->setPosition(getPosition());
  pNew->setVelocity(getVelocity());
  pNew->m_seed= m_seed;

  return pNew;
}

double cBall::getDistance(const cInstance& other) const {
  assert(other.getObjectID() == getObjectID());
  const R_2 diff = getPosition() - ((const cBall&) other).getPosition();
  double prod= diff * diff;
#ifdef PERFECT_RUN
  const R_2 vdif = getVelocity() - ((const cBall&) other).getVelocity();
  prod += vdif * vdif;
#endif
  return sqrt(prod);
}

void cBall::simulate(double newtime, cLoadMeter &load)
{
  double delta = newtime - simTime();

  if (delta <= 0.0) return;

  // move the ball
  m_position += getVelocity() * delta;

  // check for collision with upper or lower wall; do this only when the ball
  // is moving in the right direction (up or down)
  if (   (m_position[1] < cPongScenario::TOP_BORDER && m_velocity[1] < 0)
      || (m_position[1] > cPongScenario::BOTTOM_BORDER &&
          m_velocity[1] > 0)) {
    // TODO: accurately calculate reflection
    // invert vertical velocity component
    m_velocity[1] *= -1;
    load.addCPU(3);
  }

  load.addCPU(4);
  setSimTime(newtime);
}

cInstance* cBall::create(const OBJECT_ID& id, const string &name)
{ return new cBall(id, name); }

static const class cBallSlave
{
public: cBallSlave()
    { cInstanceFactory::registerInstance(string("cBall"), cBall::create); }
} ballSlave;
