/*
Copyright (C) 2007, 2008  Condel project, TU-Berlin.
Written by Jan Sablatnig, Sven Grottke, Jiehua Chen.

This file is part of Adam, the DVE testbed.

Adam is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3, or (at your option) any later
version.

Adam is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Adam; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.
*/
#include "pong_scenario.hh"
#include "message_router.hh"
#include "instance_factory.hh"
#include "ball.hh"
#include "racket.hh"
#include "score.hh"
#include "world.hh"
#include "consistency.hh"



// if defined, will save rendering information.  Quite slow.
//#define RENDER 1

// init constants
const double cPongScenario::LEFT_BORDER = 0;
const double cPongScenario::RIGHT_BORDER = 1000;
const double cPongScenario::TOP_BORDER = 0;
const double cPongScenario::BOTTOM_BORDER = 500;
const double cPongScenario::INITIAL_X_SPEED = 200;
const double cPongScenario::X_ACCELERATION = 1.05;
const double cPongScenario::RACKET_SIZE = 50;
const double cPongScenario::RACKET_SPEED = 500;


// initialize statics
OBJECT_ID cPongScenario::s_racketLeftID = cObjectFactory::registerObject("cRacket#left");
OBJECT_ID cPongScenario::s_racketRightID = cObjectFactory::registerObject("cRacket#right");
OBJECT_ID cPongScenario::s_ballID = cObjectFactory::registerObject("cBall#");
OBJECT_ID cPongScenario::s_scoreLeftID = cObjectFactory::registerObject("cScore#left");
OBJECT_ID cPongScenario::s_scoreRightID = cObjectFactory::registerObject("cScore#right");





cPongScenario::cPongScenario(cHost& host)
: cScenario(host), m_doRender(false), m_ballVec(),
  m_racketLeftVec(), m_racketRightVec(), m_scoreLeftVec(), m_scoreRightVec(),
  m_lastScore(0), m_scoreCounter(0)
{ }


void cPongScenario::initialize(void)
{
  m_host.m_attributeNorm[cAttributeFactory::nameToID("cBall::positionX")]=
                1.0 / (RIGHT_BORDER - LEFT_BORDER);
  m_host.m_attributeNorm[cAttributeFactory::nameToID("cBall::positionY")]=
                1.0 / (BOTTOM_BORDER - TOP_BORDER);
  m_host.m_attributeNorm[cAttributeFactory::nameToID("cBall::velocityX")]=
                1.0 / INITIAL_X_SPEED;
  m_host.m_attributeNorm[cAttributeFactory::nameToID("cBall::velocityY")]=
                1.0 / INITIAL_X_SPEED;          // should use Y..

  m_host.m_attributeNorm[cAttributeFactory::nameToID("cRacket::positionX")]=
                1.0 / (RIGHT_BORDER - LEFT_BORDER);
  m_host.m_attributeNorm[cAttributeFactory::nameToID("cRacket::positionY")]=
                1.0 / (BOTTOM_BORDER - TOP_BORDER);
  m_host.m_attributeNorm[cAttributeFactory::nameToID("cRacket::velocityX")]=
                1.0 / RACKET_SPEED;             // nonsense
  m_host.m_attributeNorm[cAttributeFactory::nameToID("cRacket::velocityY")]=
                1.0 / RACKET_SPEED;

  m_host.m_attributeNorm[cAttributeFactory::nameToID("cScore::score")]=
                1.0;

  m_host.m_attributeNorm[cAttributeFactory::nameToID("cBall::seed")]=
                0.0;

  m_lastScore= 0;
  m_scoreCounter= 0;

  m_isPlayerLeft= *(m_host.getRouter().beginHID())==(unsigned long)m_host.id();

  // depending on which player we are...
  if (m_isPlayerLeft)
    {
    // create, initialize and add the racket for player_left
    scoped_ptr<cRacket> pRacket((cRacket*) cInstanceFactory::
                         createInstance(s_racketLeftID, "cRacket#left"));
    R_2 racketPos;
    racketPos[0] = LEFT_BORDER;
    racketPos[1] = (TOP_BORDER + BOTTOM_BORDER) / 2.;
    pRacket->setPosition(racketPos);
    R_2 racketVel;
    racketVel[0] = 0; racketVel[1] = 0;
    pRacket->setVelocity(racketVel);
    host().getConsistency().addInstance(*pRacket.get());

    // create, initialize and add the score for player_right (yes,
    // that's the OTHER player)
    scoped_ptr<cScore> pScore((cScore*) cInstanceFactory::
                         createInstance(s_scoreRightID, "cScore#right"));
    pScore->setScore(0);
    host().getConsistency().addInstance(*pScore.get());

    // create, initialize and add the ball; it will move towards
    // player_right
    scoped_ptr<cBall> pBall((cBall*) cInstanceFactory::
                            createInstance(s_ballID, "cBall#"));
    R_2 ballPos;
    ballPos[0] = LEFT_BORDER;
    ballPos[1] = (TOP_BORDER + BOTTOM_BORDER) / 2.;
    pBall->setPosition(ballPos);
    R_2 ballVel;
    ballVel[0] = INITIAL_X_SPEED;
    ballVel[1] = 0;
    pBall->setVelocity(ballVel);
    pBall->randomYSpeed();
    host().getConsistency().addInstance(*pBall.get());
    
#if VERBOSITY >= 3
    ev << "initial ball position on " << m_host.name() << " is ("
       << pBall->getPosition()[0] << ", " << pBall->getPosition()[1] << ")\n";
#endif
    }
  else
    {
    // create, init and add the racket for player_right
    scoped_ptr<cRacket> pRacket((cRacket*) cInstanceFactory::
                          createInstance(s_racketRightID, "cRacket#right"));
    R_2 racketPos;
    racketPos[0] = RIGHT_BORDER;
    racketPos[1] = (TOP_BORDER + BOTTOM_BORDER) / 2.;
    pRacket->setPosition(racketPos);
    R_2 racketVel;
    racketVel[0] = 0; racketVel[1] = 0;
    pRacket->setVelocity(racketVel);
    host().getConsistency().addInstance(*pRacket.get());

    // create, initialize and add the score for player_left (yes, that's the
    // OTHER player)
    scoped_ptr<cScore> pScore((cScore*) cInstanceFactory::
                              createInstance(s_scoreLeftID, "cScore#left"));
    pScore->setScore(0);
    host().getConsistency().addInstance(*pScore.get());

    }

  if (host().hasPar("render") && host().par("render"))
    initRendering();
  else
    m_doRender= false;
}


void cPongScenario::simulate(cWorld& world, double newtime,
        cLoadMeter &load) const
{
#if VERBOSITY >= 2
  ev << "handling world \"" << world.getName() << "\" on host "
     << m_host.name() << "\n";
#endif
  
  // simulate the instances
  cWorld::iterator instIter;

  for (instIter = world.begin(); instIter != world.end(); instIter++)
    (*instIter).second->simulate(newtime, load);

  // global checks

  // neither player can act w/o knowing the ball
  if (!world.isReplicated(s_ballID))
    {
    world.setSimTime(newtime);
    return;
    }
  cBall& ball = (cBall&) world.getInstance(s_ballID);

#if VERBOSITY >= 2
  ev << "ball position on " << m_host.name() << " is " << ball.getPosition()
     << "\n";
#endif

  // have we crossed the left or right baseline?  Hit the racket?
  if (ball.getPosition()[0] <= LEFT_BORDER && ball.getVelocity()[0] < 0)
    {                                           // crossed left
    const cRacket *racket= NULL;
    if (world.isReplicated(s_racketLeftID))
      racket= (cRacket *) &world.getInstance(s_racketLeftID);
    checkCollision(world, ball, racket, s_scoreRightID, load, LEFT_BORDER);
    }

  // check for collision with right racket if necessary and possible
  if (ball.getPosition()[0] >= RIGHT_BORDER && ball.getVelocity()[0] > 0)
    {                                           // crossed right
    const cRacket *racket= NULL;
    if (world.isReplicated(s_racketRightID))
      racket= (cRacket *) &world.getInstance(s_racketRightID);
    checkCollision(world, ball, racket, s_scoreLeftID, load, RIGHT_BORDER);
    }
  
#if VERBOSITY >= 2
  ev << "done with world \"" << world.getName() << "\" on host "
     << m_host.name() << "\n\n";
#endif
  load.addCPU(5);

  world.setSimTime(newtime);
}

void cPongScenario::checkCollision(cWorld& world, cBall& ball,
        const cRacket *racket, const OBJECT_ID& scoreID, cLoadMeter &load,
        double newX) const
{
  // check for collision with racket; this is done by a rough estimate, should
  // calculate the exact Y coordinate where the ball crossed the base line
  // instead
  const double ballY = ball.getPosition()[1];

#if VERBOSITY >= 2
  if (racket)
    ev << "checking collision, ballY = " << ballY << ", racketY = "
       << racket->getPosition()[1] << ", host " << m_host.name()
       << ", ballX = " << ball.getPosition()[0] << ", racketX = "
       << racket->getPosition()[0] << "\n";
  else
    ev << "checking collision, ballY = " << ballY << ", no racket, host "
       << m_host.name() << ", ballX = " << ball.getPosition()[0] << endl;
#endif
  
  if (racket && ballY > racket->upperEnd() && ballY <= racket->lowerEnd())
    {                                                    // hit racket!
    R_2 ballV= ball.getVelocity();
    if (ballY <= racket->upperEnd() + RACKET_SIZE / 3)
      {                                                  // upper third, deflect
      if (ballV[1] < 0)
        ballV[1] *= 2;
      else
        ballV[1] /= 2;
      load.addCPU(4);
      }
    else if (ballY > racket->upperEnd() + 2 * RACKET_SIZE / 3)
      {                                                  // lower third, deflect
      if (ballV[1] < 0)
        ballV[1] /= 2;
      else
        ballV[1] *= 2;
      load.addCPU(4);
      }

    // accelerate ball on backswing
    ballV[0] *= -X_ACCELERATION;
    ball.setVelocity(ballV);
    load.addCPU(5);
    }
  else
    {
    // missed the racket --> score for the opponent!!!
    if (world.isReplicated(scoreID))
      ((cScore&) world.getInstance(scoreID)).getScore() = 
      	round(((cScore&) world.getInstance(scoreID)).getScore() + 1.0);

    // reset ball status
    R_2 ballP;
    ballP[0] = newX;
    ballP[1] = (TOP_BORDER + BOTTOM_BORDER) / 2.;
    ball.setPosition(ballP);
    R_2 ballV= ball.getVelocity();
    ballV[0] = - copysign(INITIAL_X_SPEED, ballV[0]);
    ball.setVelocity(ballV);
    ball.randomYSpeed();        // 8 ops
    load.addCPU(19);
    }
}

void cPongScenario::ai(double newtime)
{
  // call the pseudo-renderer
  if (m_doRender)
    render();

  unsigned long score= 0;

  OBJECT_ID scoreID= m_isPlayerLeft ? s_scoreLeftID : s_scoreRightID;
  if (host().getConsistency().isReplicated(scoreID))
    score= lrint(((cScore&) host().getConsistency().getInstance(scoreID))
                .getScore());
  if (score > m_lastScore)
    m_scoreCounter++;
  m_lastScore= score;
  
  // neither player can act w/o knowing the ball
  if (!host().getConsistency().isReplicated(s_ballID)) return;
  cBall& ball = (cBall&) host().getConsistency().getInstance(s_ballID);

  // is the ball moving away from this player?
  const bool movingAway = m_isPlayerLeft && ball.getVelocity()[0] >= 0
        || !m_isPlayerLeft && ball.getVelocity()[0] <0;

  // Done if ball is moving away.
  if (movingAway) return;

  cWorld &world= host().getWorld();
  cWorld *future= &world;
  auto_ptr<cWorld> future_changed;
  if (m_delayHint > 0.0)
    {
    future_changed= future->clone();
    future= future_changed.get();
    cLoadMeter fake;
    future->simulate(host().now() + m_delayHint, fake);
    }

  // this player's racket
  const OBJECT_ID& racketID = m_isPlayerLeft ? s_racketLeftID : s_racketRightID;
  cRacket& racket = (cRacket&) future->getInstance(racketID);

  // See where the ball will impact.  Note I'm using the CURRENT ball
  // (not future).
  // determine time till impact
  const double time = (racket.getPosition()[0] - ball.getPosition()[0]) /
    ball.getVelocity()[0];
  // determine the point of impact (vertical position)
  const double impactPos =
    ball.getPosition()[1] + ball.getVelocity()[1] * time;

  double racket_speed= 0.0;

  // move the racket towards the point of impact if necessary
  if (racket.lowerEnd() < impactPos)
    racket_speed = RACKET_SPEED;
  else if (racket.upperEnd() > impactPos)
    racket_speed = -RACKET_SPEED;

  // Send off the new racket.velocity
  if (racket.getVelocity()[1] != racket_speed)
    host().getConsistency().changeAttribute(racketID, cRacket::s_velocityYID,
                                  racket_speed);
}

void cPongScenario::initRendering(void) {
  // set the names for all attributes
  m_doRender= true;
  m_ballVec.posX.setName("ballPosX");
  m_ballVec.posY.setName("ballPosY");
  m_ballVec.velX.setName("ballVelX");
  m_ballVec.velY.setName("ballVelY");
  m_racketLeftVec.posX.setName("racketLeftPosX");
  m_racketLeftVec.posY.setName("racketLeftPosY");
  m_racketLeftVec.velX.setName("racketLeftVelX");
  m_racketLeftVec.velY.setName("racketLeftVelY");
  m_racketRightVec.posX.setName("racketRightPosX");
  m_racketRightVec.posY.setName("racketRightPosY");
  m_racketRightVec.velX.setName("racketRightVelX");
  m_racketRightVec.velY.setName("racketRightVelY");
  m_scoreLeftVec.setName("score_left");
  m_scoreRightVec.setName("score_right");
}

void cPongScenario::render(void)
{
  // for each object, render it if it's replicated
  
  if (host().getConsistency().isReplicated(s_ballID)) {
    const R_2& ballPos = ((cBall&) host().getConsistency().getInstance(s_ballID)).getPosition();
    const R_2& ballVel = ((cBall&) host().getConsistency().getInstance(s_ballID)).getVelocity();
    m_ballVec.posX.record(ballPos[0]);
    m_ballVec.posY.record(ballPos[1]);
    m_ballVec.velX.record(ballVel[0]);
    m_ballVec.velY.record(ballVel[1]);
  }

  if (host().getConsistency().isReplicated(s_racketLeftID)) {
    const R_2& racketPos =
      ((cRacket&) host().getConsistency().getInstance(s_racketLeftID)).getPosition();
    const R_2& racketVel =
      ((cRacket&) host().getConsistency().getInstance(s_racketLeftID)).getVelocity();
    m_racketLeftVec.posX.record(racketPos[0]);
    m_racketLeftVec.posY.record(racketPos[1]);
    m_racketLeftVec.velX.record(racketVel[0]);
    m_racketLeftVec.velY.record(racketVel[1]);
  }

  if (host().getConsistency().isReplicated(s_racketRightID)) {
    const R_2& racketPos =
      ((cRacket&) host().getConsistency().getInstance(s_racketRightID)).getPosition();
    const R_2& racketVel =
      ((cRacket&) host().getConsistency().getInstance(s_racketRightID)).getVelocity();
    m_racketRightVec.posX.record(racketPos[0]);
    m_racketRightVec.posY.record(racketPos[1]);
    m_racketRightVec.velX.record(racketVel[0]);
    m_racketRightVec.velY.record(racketVel[1]);
  }

  if (host().getConsistency().isReplicated(s_scoreLeftID)) {
    const double score =
      ((cScore&) host().getConsistency().getInstance(s_scoreLeftID)).getScore();
    m_scoreLeftVec.record(score);
  }

  if (host().getConsistency().isReplicated(s_scoreRightID)) {
    const double score =
      ((cScore&) host().getConsistency().getInstance(s_scoreRightID)).getScore();
    m_scoreRightVec.record(score);
  }
}

void cPongScenario::finish(void)
{
  m_host.recordScalar("score achieved /s", m_scoreCounter/host().now());

  host().getConsistency().finish();
}




