/*
Copyright (C) 2007, 2008  Condel project, TU-Berlin.
Written by Jan Sablatnig, Sven Grottke, Jiehua Chen.

This file is part of Adam, the DVE testbed.

Adam is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3, or (at your option) any later
version.

Adam is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Adam; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.
*/

#ifndef __PONG_SCENARIO_H
#define __PONG_SCENARIO_H

#include "scenario.hh"
#include "object_factory.hh"


namespace Condel {

class cRacket;
class cBall;
class cMessageRouter;

/// Helper object for visualization.
/// This is used in dumping all data for later visualization.
/// Only used within the scenario-file.

struct cRenderVec
{
  cOutVector posX;
  cOutVector posY;
  cOutVector velX;
  cOutVector velY;
};

/// The Pong Scenario.
/// This is a simple game, fashioned after the classic video game Pong.
/// Two players play a simplified tennis vs. each other.  Each player
/// controls a racket (can only move this up and down).  A ball is thrown
/// in and the player has to move the racket so that it reflects the ball
/// and does not cross the player's rear line.
///
/// In comparison to the original Pong, there are two complications:
/// <ul><li>The rackets have three zones (each one third of the racket).  The
///         middle zone reflects normally, but the lower zone deflects
///         downwards, while the upper zone reflects upwards.
///     <li>The ball is accellerated (by 5%) in the X-direction whenever it
///         is reflected. </ul>
/// 
/// The AI is very simple, it will only move the racket towards the estimated
/// collision point.  When the ball is reflected off the side-lines, this
/// estimation changes, causing the AI to issue a new move command.
///
/// If the render parameter is true for this host, positions of ball, rackets
/// and scores will be logged to the omnetpp.vec file each frame.  This
/// can be used to create a visualization later.


class cPongScenario : public cScenario
{
public:
  // public constants
  static const double LEFT_BORDER;     ///< X-coordinate of the left border.
  static const double RIGHT_BORDER;    ///< X-coordinate of the right border.
  static const double TOP_BORDER;      ///< Y-coordinate of the upper border
  static const double BOTTOM_BORDER;   ///< Y-coordinate of the lower border
  static const double INITIAL_X_SPEED; ///< Initial horizontal velocity of the
                                       ///  ball, in pels/second.
  static const double X_ACCELERATION;  ///< Horizontal acceleration factor.
  static const double RACKET_SIZE;     ///< Height of the racket, in pels.
  static const double RACKET_SPEED;    ///< Vertical racket velocity.


protected:
  // this is ugly, but fast
  static OBJECT_ID s_racketLeftID;
  static OBJECT_ID s_racketRightID;
  static OBJECT_ID s_ballID;
  static OBJECT_ID s_scoreLeftID;
  static OBJECT_ID s_scoreRightID;

  // private helper for handling ball/racket collisions
  void checkCollision(cWorld& world, cBall& ball, const cRacket *racket,
                 const OBJECT_ID& scoreID, cLoadMeter &load, double newX) const;

  // for output to the visualisation interface
  bool m_doRender;
  cRenderVec m_ballVec;
  cRenderVec m_racketLeftVec;
  cRenderVec m_racketRightVec;
  cOutVector m_scoreLeftVec;
  cOutVector m_scoreRightVec;
  void initRendering(void);
  void render(void);

  unsigned long m_lastScore;            // Last seen score for myself in ai()
  unsigned long m_scoreCounter;         // Count scores I see for MYSELF

  bool m_isPlayerLeft;                  // true for the left player

public:
  cPongScenario(cHost& host);
  virtual ~cPongScenario(void) {}
 
  /// Reset everything.
  /// Create player's racket, opponents score and (if left player) the ball.
  void initialize(void);

  /// Perform deterministic actions.
  /// Let each instance in the world simulate to newtime (moves ball/rackets).
  /// Check if ball moved over a baseline.  If so, increment score and
  /// rethrow the ball.  If reflected off racket, adjust ballspeed according
  /// to racket-zone.
  /// @param world The world that is advanced.
  /// @param newtime Time to advance the system to.  This should be in the
  /// future of the world's cWorld::simTime().  Given in seconds.
  /// @param load A tool to measure CPU and memory usage during the simulation.
  virtual void simulate(cWorld& world, double newtime, cLoadMeter &load)
        const;

  /// Perform the AI.
  /// Simply estimate the point of intersection between the ball and the
  /// baseline and move towards that.
  /// @param newtime The current time.
  virtual void ai(double newtime);

  /// System Done.
  /// Write the score to the log file (it's the application dependant yield
  /// measure).
  virtual void finish(void);
};
  
};

#endif // __PONG_SCENARIO_H
