/*
Copyright (C) 2007, 2008  Condel project, TU-Berlin.
Written by Sven Grottke, Jan Sablatnig, Jiehua Chen.

This file is part of Adam, the DVE testbed.

Adam is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3, or (at your option) any later
version.

Adam is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Adam; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.
*/
#include <math.h>

#include "racket.hh"
#include "instance_factory.hh"
#include "pong_scenario.hh"

ATTRIBUTE_ID cRacket::s_positionXID = cAttributeFactory::registerAttribute("cRacket::positionX");
ATTRIBUTE_ID cRacket::s_positionYID = cAttributeFactory::registerAttribute("cRacket::positionY");
ATTRIBUTE_ID cRacket::s_velocityXID = cAttributeFactory::registerAttribute("cRacket::velocityX");
ATTRIBUTE_ID cRacket::s_velocityYID = cAttributeFactory::registerAttribute("cRacket::velocityY");

cRacket::cRacket(const OBJECT_ID& id, const string& name)
: cInstance(id, name, 4)
{ }

double cRacket::upperEnd() const {
  return m_position[1] - cPongScenario::RACKET_SIZE / 2.;
}

double cRacket::lowerEnd() const {
  return m_position[1] + cPongScenario::RACKET_SIZE / 2.;
}

cAttributeSet cRacket::getAttributes() const {
  cAttributeSet ret;

  ret.add(cAttribute(s_positionXID, getPosition()[0]));
  ret.add(cAttribute(s_positionYID, getPosition()[1]));
  ret.add(cAttribute(s_velocityXID, getVelocity()[0]));
  ret.add(cAttribute(s_velocityYID, getVelocity()[1]));

  return ret;
}

void cRacket::setAttributes(const cAttributeSet& na) {
  cAttributeSet::const_iterator iter;

  R_2 oldpos = getPosition();
  R_2 oldvel = getVelocity();

  for (iter = na.begin(); iter != na.end(); iter++) {
    const cAttribute& attr = (*iter).second;
    
    if (attr.getID() == s_positionXID) oldpos[0] = attr.getValue();
    else if (attr.getID() == s_positionYID) oldpos[1] = attr.getValue();
    else if (attr.getID() == s_velocityXID) oldvel[0] = attr.getValue();
    else if (attr.getID() == s_velocityYID) oldvel[1] = attr.getValue();
    else assert(false);                // unkown attribute found
  }
  
  setPosition(oldpos);
  setVelocity(oldvel);
}


cRacket* cRacket::clone() const {
  cRacket* pNew = new cRacket(getObjectID(), getName());

  pNew->setSimTime(simTime());
  pNew->setUpdateTime(updateTime());

  pNew->setPosition(getPosition());
  pNew->setVelocity(getVelocity());

  return pNew;
}

double cRacket::getDistance(const cInstance& other) const {
  assert(other.getObjectID() == getObjectID());
  const R_2 diff = getPosition() - ((const cRacket&) other).getPosition();
  double prod= diff * diff;
#ifdef PERFECT_RUN
  const R_2 vdif = getVelocity() - ((const cRacket&) other).getVelocity();
  prod += vdif * vdif;
#endif
  return sqrt(prod);
}

void cRacket::simulate(double newtime, cLoadMeter &load)
{
  double delta= newtime- simTime();
  if (delta <=0.0) return;

  // move the racket
  m_position += m_velocity * delta;

  // boundary checks
  const double minPos = cPongScenario::TOP_BORDER +
    cPongScenario::RACKET_SIZE / 2.;
  if (m_position[1] < minPos)
    m_position[1] = minPos;

  const double maxPos = cPongScenario::BOTTOM_BORDER -
    cPongScenario::RACKET_SIZE / 2.;
  if (m_position[1] > maxPos)
    m_position[1] = maxPos;

  load.addCPU(4);
  setSimTime(newtime);
}

cInstance* cRacket::create(const OBJECT_ID& id, const string& name) {
  return new cRacket(id, name);
}

static const class cRacketSlave { public: cRacketSlave() {
    cInstanceFactory::registerInstance(string("cRacket"), cRacket::create);
} } racketSlave;
