/*
Copyright (C) 2007, 2008  Condel project, TU-Berlin.
Written by Jan Sablatnig, Sven Grottke, Jiehua Chen.

This file is part of Adam, the DVE testbed.

Adam is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3, or (at your option) any later
version.

Adam is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Adam; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.
*/
#include <math.h>

#include "bee.hh"
#include "instance_factory.hh"
#include "swarm_scenario.hh"


const double cBee::RADIUS = 10;
const double cBee::MAX_V = 200;        // so, 10pel/TICK
const double cBee::ACCELERATION = 10;    // so, need 2 secs to reach max speed



ATTRIBUTE_ID cBee::s_positionXID = cAttributeFactory::registerAttribute("cBee::positionX");
ATTRIBUTE_ID cBee::s_positionYID = cAttributeFactory::registerAttribute("cBee::positionY");
ATTRIBUTE_ID cBee::s_velocityAID = cAttributeFactory::registerAttribute("cBee::velocityAbs");
ATTRIBUTE_ID cBee::s_velocityPID = cAttributeFactory::registerAttribute("cBee::velocityPhi");
ATTRIBUTE_ID cBee::s_accelerationAID = cAttributeFactory::registerAttribute("cBee::accelerationAbs");
ATTRIBUTE_ID cBee::s_accelerationPID = cAttributeFactory::registerAttribute("cBee::accelerationPhi");
ATTRIBUTE_ID cBee::s_deadID = cAttributeFactory::registerAttribute("cBee::dead");

cBee::cBee(const OBJECT_ID& id, const string& name, unsigned long size)
: cInstance(id, name, size), m_position(), m_velocity(), m_acceleration(), m_dead(0.0)
{
  m_position[0]= m_position[1]= 0.0;
  m_velocity.setXY(0.0, 0.0);
  m_acceleration.setXY(0.0, 0.0);
}

cAttributeSet cBee::getAttributes() const
{
  cAttributeSet ret;

  ret.add(cAttribute(s_positionXID, getPosition()[0]));
  ret.add(cAttribute(s_positionYID, getPosition()[1]));
  ret.add(cAttribute(s_velocityAID, getVelocity().abs()));
  ret.add(cAttribute(s_velocityPID, getVelocity().phi()));
  ret.add(cAttribute(s_accelerationAID, getAcceleration().abs()));
  ret.add(cAttribute(s_accelerationPID, getAcceleration().phi()));
  ret.add(cAttribute(s_deadID, getDead()));

  return ret;
}

void cBee::setAttributes(const cAttributeSet& na)
{
  cAttributeSet::const_iterator iter;

  R_2 oldpos = getPosition();
  cPolar oldvel = getVelocity();
  cPolar oldacc = getAcceleration();

  for (iter = na.begin(); iter != na.end(); iter++)
    {
    const cAttribute& attr = (*iter).second;
    
    if (attr.getID() == s_positionXID) oldpos[0] = attr.getValue();
    else if (attr.getID() == s_positionYID) oldpos[1] = attr.getValue();
    else if (attr.getID() == s_velocityAID) oldvel.setAbs(attr.getValue());
    else if (attr.getID() == s_velocityPID) oldvel.setPhi(attr.getValue());
    else if (attr.getID() == s_accelerationAID) oldacc.setAbs(attr.getValue());
    else if (attr.getID() == s_accelerationPID) oldacc.setPhi(attr.getValue());
    else if (attr.getID() == s_deadID) setDead(attr.getValue());
    else assert(false);                // unkown attribute found
    }
  
  setPosition(oldpos);
  setVelocity(oldvel);
  setAcceleration(oldacc);
}


cBee* cBee::clone(void) const
{
  cBee* pNew = new cBee(getObjectID(), getName());

  pNew->setSimTime(simTime());
  pNew->setUpdateTime(updateTime());

  pNew->setPosition(getPosition());
  pNew->setVelocity(getVelocity());
  pNew->setAcceleration(getAcceleration());
  pNew->setDead(getDead());

  return pNew;
}

void cBee::simulate(double newtime, cLoadMeter &load)
{
  double delta= newtime- simTime();
  if (delta <=0.0) return;

  if (!isAlive())
    addDead(-delta);
  else
    m_velocity += m_acceleration * delta;

  if (m_velocity.abs() > MAX_V)
    m_velocity.setAbs(MAX_V);

  m_position += m_velocity * delta;

  // boundary checks
  double minPos = cSwarmScenario::TOP_BORDER + RADIUS;
  double maxPos = cSwarmScenario::BOTTOM_BORDER - RADIUS;
  if (m_position[1] < minPos)
    {
    m_position[1] += 2 * (minPos - m_position[1]);
    m_velocity.setY(m_velocity.y() * -1);
    }
  else if (m_position[1] > maxPos)
    {
    m_position[1] -= 2 * (m_position[1] - maxPos);
    m_velocity.setY(m_velocity.y() * -1);
    }

  minPos = cSwarmScenario::LEFT_BORDER + RADIUS;
  maxPos = cSwarmScenario::RIGHT_BORDER - RADIUS;
  if (m_position[0] < minPos)
    {
    m_position[0] += 2 * (minPos - m_position[0]); 
    m_velocity.setX(m_velocity.x() * -1);
    }
  else if (m_position[0] > maxPos)
    {
    m_position[0] -= 2 * (m_position[0] - maxPos);
    m_velocity.setX(m_velocity.x() * -1);
    }

  load.addCPU(8);
  setSimTime(newtime);
}

cInstance* cBee::create(const OBJECT_ID& id, const string& name)
{ return new cBee(id, name); }

static const class cBeeSlave
{ public: cBeeSlave()
  { cInstanceFactory::registerInstance(string("cBee"), cBee::create); }
} beeSlave;
