/*
Copyright (C) 2007, 2008  Condel project, TU-Berlin.
Written by Jan Sablatnig, Sven Grottke, Jiehua Chen.

This file is part of Adam, the DVE testbed.

Adam is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3, or (at your option) any later
version.

Adam is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Adam; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.
*/

#ifndef __BEE_H
#define __BEE_H

#include "gen.hh"
#include "instance.hh"
#include "polar.hh"
#include "attribute.hh"

namespace Condel {

/// Swarm's bee object.
/// This represents the players.
///
/// Note there are @em many bee objects, one for each player.


class cBee : public cInstance
{
public:
  static const double RADIUS;           ///< Radius of a bee.  In pels.
  static const double ACCELERATION;     ///< Typical acceleration in pels/s /s
  static const double MAX_V;            ///< Maximum velocity.
                                        ///  @f$ ||v|| \le @f$ MAX_V


  // the bee attribute IDs
  static ATTRIBUTE_ID s_positionXID;        ///< X-position of bee.
  static ATTRIBUTE_ID s_positionYID;        ///< Y-position of bee.
  static ATTRIBUTE_ID s_velocityAID;        ///< X-velocity of bee.
  static ATTRIBUTE_ID s_velocityPID;        ///< Y-velocity of bee.
  static ATTRIBUTE_ID s_accelerationAID;    ///< X-acceleration of bee.
  static ATTRIBUTE_ID s_accelerationPID;    ///< Y-acceleration of bee.
  static ATTRIBUTE_ID s_deadID;             ///< Dead state of bee.  If not 0,
                                            /// indicates how long dead yet.
  
  
protected:
  R_2 m_position;    // This is the position of the centre of the bee
  cPolar m_velocity;    // Its movement vector
  cPolar m_acceleration;    // Its acceleration vector
  double m_dead;     // if >0.0, the bee is currently dead, awaiting revival
                     // in m_dead seconds.
public:
  cBee(const OBJECT_ID& id, const string& name, unsigned long size=4);
  virtual ~cBee(void) {}

  // accessors
  void setPosition(const R_2& pos) { m_position = pos; }
  void setVelocity(const cPolar& v)
    {
    m_velocity = v;
    if (m_velocity.abs() > MAX_V)
      m_velocity.setAbs(MAX_V);
    }
  void setAcceleration(const cPolar& acc) { m_acceleration = acc; }
  const R_2& getPosition(void) const { return m_position; }
  const cPolar& getVelocity(void) const { return m_velocity; }
  const cPolar& getAcceleration(void) const { return m_acceleration; }

  bool isAlive(void) const { return getDead() <= 0.0; }
  double getDead(void) const { return m_dead; }
  void setDead(double time) { m_dead = time; }
  void addDead(double time) { m_dead += time; if (m_dead <0.0) m_dead=0.0; }

  static cInstance* create(const OBJECT_ID& id, const string& name);
  
  virtual cAttributeSet getAttributes(void) const;
  virtual void setAttributes(const cAttributeSet& na);
  
  virtual cBee* clone(void) const;

  /// Compares two instances.
  /// Normally, this is based on the cartesian distance.
  /// For the ideal case, velocity/acceleration is also used.
  /// @todo Need to fix ideal case here.
  /// @param other Another instance of the same object.
  /// @return A metric of the similarity of the two instances.
  virtual double getDistance(const cInstance& other) const
    {
    const cPolar pol(((const cBee&) other).getPosition() - getPosition());
    return pol.abs();
    }

  /// Get the direction to the other bee.
  /// @return The angle I'd have to choose to shoot at the other bee.
  /// @sa cPolar
  virtual double getDirection(const cInstance& other) const
    {
    const cPolar pol(((const cBee&) other).getPosition() - getPosition());
    return pol.phi();
    }

  /// Simulate this instance.
  /// Accelerate the bee according to its acceleration parameter.
  /// Move the bee according to its new speed.
  /// Bounce off any walls.
  /// Move the b according to its speed.  Stop at upper and lower
  /// boundaries.
  /// @param newtime A simulation time in the future of simTime().  Only
  /// the difference between the two times is advanced.
  /// @param load A tool to measure CPU and memory usage during the simulation.
  virtual void simulate(double newtime, cLoadMeter &load);
};
  
};

#endif // __BEE_H
