/*
Copyright (C) 2007, 2008  Condel project, TU-Berlin.
Written by Jan Sablatnig, Sven Grottke, Jiehua Chen.

This file is part of Adam, the DVE testbed.

Adam is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3, or (at your option) any later
version.

Adam is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Adam; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.
*/
#include <math.h>

#include "queen.hh"
#include "instance_factory.hh"
#include "swarm_scenario.hh"



// Find a new big random speed every 5 seconds
#define RANDOM_RESET_A  5
// Find a new small random speed on every TICK
#define RANDOM_RESET_B  0.05

#define A_OVER_B_WEIGHT 10

// This many seconds until a full circle is done if random always yielded MAX.
#define ACC_PHI_SEC 10

#define TICK_PRECISION  0.00001


// The queen should accelerate slower than the bee's otherwise they just cannot
// follow.  0.8 is 80%.
#define QUEEN_ACCELERATION_FACTOR 0.8

#define QUEEN_ACCELERATION (cBee::ACCELERATION * QUEEN_ACCELERATION_FACTOR)


// Keeping the queen inbounds:
// Try to stay away from the border at least this many pixels:
#define QUEEN_INBOUND_SAFETY 50
// Start to avoid crashing into a wall on average this many seconds before.
#define QUEEN_INBOUND_KAPPA 5

ATTRIBUTE_ID cQueen::s_positionXID = cAttributeFactory::registerAttribute("cQueen::positionX");
ATTRIBUTE_ID cQueen::s_positionYID = cAttributeFactory::registerAttribute("cQueen::positionY");
ATTRIBUTE_ID cQueen::s_velocityAID = cAttributeFactory::registerAttribute("cQueen::velocityAbs");
ATTRIBUTE_ID cQueen::s_velocityPID = cAttributeFactory::registerAttribute("cQueen::velocityPhi");
ATTRIBUTE_ID cQueen::s_accelerationAID = cAttributeFactory::registerAttribute("cQueen::accelerationAbs");
ATTRIBUTE_ID cQueen::s_accelerationPID = cAttributeFactory::registerAttribute("cQueen::accelerationPhi");
ATTRIBUTE_ID cQueen::s_seedID = cAttributeFactory::registerAttribute("cQueen::seed");

double cQueen::check_critical(cPolar &warrow)
{
  // The velocity in the critical direction
  double vel= (warrow.x() * getVelocity().x() + warrow.y() * getVelocity().y())
                / warrow.abs();

  // The distance to the wall
  double dist= warrow.abs() -cBee::RADIUS -QUEEN_INBOUND_SAFETY;

  double min_stop_dist= vel * vel / QUEEN_ACCELERATION / 2.0;
  double remaining_time= (dist - min_stop_dist) / vel;
  

  return QUEEN_INBOUND_KAPPA - remaining_time;
}

void cQueen::avoid_wall(cPolar &warrow)
{
  double angle= cPolar::normalize_phi(getVelocity().phi() - warrow.phi());

  double dir= 1;
  if (angle < 0)
    dir= -1;

  // turn 90 deg in the natural direction
  m_acceleration.setPhi(getVelocity().phi() + dir*M_PI_2);
}


void cQueen::simulate(double newtime, cLoadMeter &load)
{
  double delta= newtime- simTime();
  if (delta <=0.0) return;

  m_acceleration.setAbs(QUEEN_ACCELERATION);

  if (fabs(remainder(newtime, RANDOM_RESET_B)) < TICK_PRECISION)
    {
    // a random number between -1..+1.
    double random= ((double) rand_r(&m_seed)) / RAND_MAX * 2.0 - 1.0;
    random *= 2 * M_PI / ACC_PHI_SEC * RANDOM_RESET_B;

    m_acceleration.setPhi(m_acceleration.phi() + random);
    }
  if (fabs(remainder(newtime, RANDOM_RESET_A)) < TICK_PRECISION)
    {
    // a random number between -1..+1.
    double random= ((double) rand_r(&m_seed)) / RAND_MAX * 2.0 - 1.0;
    random *= 2 * M_PI / ACC_PHI_SEC * RANDOM_RESET_A * A_OVER_B_WEIGHT;

    m_acceleration.setPhi(m_acceleration.phi() + random);
    }


  {
  // Avoid running the queen into the wall

  // Step 1:  The queen is moving in some direction.  There are (at most)
  // two sides of the field that it is moving towards to.
  // Find those, expressed as arrows toward the closest point on those walls

  cPolar warrow1, warrow2;
  if (m_velocity.y() >= 0)      // moving downwards
    warrow1.setPhiAbs(-M_PI_2, cSwarmScenario::BOTTOM_BORDER -m_position[1]);
  else
    warrow1.setPhiAbs(M_PI_2, m_position[1] -cSwarmScenario::TOP_BORDER);

  if (m_velocity.x() >= 0)
    warrow2.setPhiAbs(0, cSwarmScenario::RIGHT_BORDER -m_position[0]);
  else
    warrow2.setPhiAbs(M_PI, m_position[0] -cSwarmScenario::LEFT_BORDER);


  // Step 2:  Find out if movement into any of those directions is critical,
  // and, if both, which one is MORE critical.

  double crit1 = check_critical(warrow1);
  double crit2 = check_critical(warrow2);

  cPolar *av1= &warrow1;
  cPolar *av2= &warrow2;

  if (crit2 > crit1)
    {
    av1= &warrow2;
    av2= &warrow1;
    double t= crit2;
    crit2= crit1;
    crit1= t;
    }

  // Step 3:  Actually avoid the wall, by turning away from it
  // We need to differentiate between two cases:

  // a) The queen is going into a corner
  if (crit1 > 0 && crit2 > 0)
    {
    // A vector pointing OUT of the corner, 45deg
    cPolar carrow = warrow1 *(-1/warrow1.abs()) + warrow2*(-1/warrow2.abs());
    carrow.setAbs(getAcceleration().abs());
    m_acceleration= carrow;
    }

  // b) Only need to avoid a single wall
  else if (crit1 > 0)
    avoid_wall(*av1);
  else if (crit2 > 0)
    avoid_wall(*av2);
  }

  load.addCPU(12);

  cBee::simulate(newtime, load);
}





cQueen::cQueen(const OBJECT_ID& id, const string& name)
: cBee(id, name, 5), m_seed(intuniform(INT_MIN, INT_MAX) -INT_MIN)
{ }

cAttributeSet cQueen::getAttributes() const
{
  cAttributeSet ret;

  ret.add(cAttribute(s_positionXID, getPosition()[0]));
  ret.add(cAttribute(s_positionYID, getPosition()[1]));
  ret.add(cAttribute(s_velocityAID, getVelocity().abs()));
  ret.add(cAttribute(s_velocityPID, getVelocity().phi()));
  ret.add(cAttribute(s_accelerationAID, getAcceleration().abs()));
  ret.add(cAttribute(s_accelerationPID, getAcceleration().phi()));
  ret.add(cAttribute(s_seedID, (double) (long) m_seed));

  return ret;
}




void cQueen::setAttributes(const cAttributeSet& na)
{
  cAttributeSet::const_iterator iter;

  R_2 oldpos = getPosition();
  cPolar oldvel = getVelocity();
  cPolar oldacc = getAcceleration();

  for (iter = na.begin(); iter != na.end(); iter++)
    {
    const cAttribute& attr = (*iter).second;
    
    if (attr.getID() == s_positionXID) oldpos[0]= attr.getValue();
    else if (attr.getID() == s_positionYID) oldpos[1]= attr.getValue();
    else if (attr.getID() == s_velocityAID) oldvel.setAbs(attr.getValue());
    else if (attr.getID() == s_velocityPID) oldvel.setPhi(attr.getValue());
    else if (attr.getID() == s_accelerationAID) oldacc.setAbs(attr.getValue());
    else if (attr.getID() == s_accelerationPID) oldacc.setPhi(attr.getValue());
    else if (attr.getID() == s_seedID) m_seed= (unsigned long)
                                                  lrint(attr.getValue());
    else assert(false);		// unknown attribute found
    }

  setPosition(oldpos);
  setVelocity(oldvel);
  setAcceleration(oldacc);
}


cQueen* cQueen::clone() const {
  cQueen* pNew = new cQueen(getObjectID(), getName());

  pNew->setSimTime(simTime());
  pNew->setUpdateTime(updateTime());

  pNew->setPosition(getPosition());
  pNew->setVelocity(getVelocity());
  pNew->setAcceleration(getAcceleration());
  pNew->m_seed= m_seed;

  return pNew;
}

cInstance* cQueen::create(const OBJECT_ID& id, const string &name)
{ return new cQueen(id, name); }

static const class cQueenSlave
{ public: cQueenSlave()
  { cInstanceFactory::registerInstance(string("cQueen"), cQueen::create); }
} queenSlave;
