/*
Copyright (C) 2007, 2008  Condel project, TU-Berlin.
Written by Jan Sablatnig, Sven Grottke, Jiehua Chen.

This file is part of Adam, the DVE testbed.

Adam is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3, or (at your option) any later
version.

Adam is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Adam; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.
*/

#ifndef __QUEEN_H
#define __QUEEN_H

#include "bee.hh"

namespace Condel {

/// Swarm's queen object.
/// The queen is computer-controlled and does a random walk.
/// This random walk avoids running into walls.
/// The randomness of its walk is controlled by a builtin random seed.
///
/// Just like any other bee, the queen cannot simply set its position
/// or its velocity, it can only control its acceleration.

class cQueen : public cBee
{
protected: // this could be private
  static ATTRIBUTE_ID s_positionXID;        ///< X-position of queen.
  static ATTRIBUTE_ID s_positionYID;        ///< Y-position of queen.
  static ATTRIBUTE_ID s_velocityAID;        ///< X-velocity of queen.
  static ATTRIBUTE_ID s_velocityPID;        ///< Y-velocity of queen.
  static ATTRIBUTE_ID s_accelerationAID;    ///< X-acceleration of queen.
  static ATTRIBUTE_ID s_accelerationPID;    ///< Y-acceleration of queen.
  static ATTRIBUTE_ID s_seedID;             ///< Queen's random seed, used
                                            /// for random-walk

private:
  unsigned int m_seed;

protected:
  cQueen(const OBJECT_ID& id, const string& name);

  /// Find out if current move will collide with a wall.
  /// @param warrow Points to a wall.
  /// @return A critical number.  If this is >0, then an action is required.
  double check_critical(cPolar &warrow);

  /// Turn away from the wall.
  /// @param warrow Points to a wall.
  void avoid_wall(cPolar &warrow);

public:
  virtual ~cQueen(void) {}

  static cInstance* create(const OBJECT_ID& id, const string &name);
  
  virtual cAttributeSet getAttributes(void) const;

  virtual void setAttributes(const cAttributeSet& na);
  
  virtual cQueen* clone(void) const;

  /// Simulate this instance.
  /// Picks a random-walk direction.  Checks if this would run into
  /// a wall and avoids that if necessary.
  /// Finally, call cBee::simulate() for everything else.
  /// @param newtime A simulation time in the future of simTime().  Only
  /// the difference between the two times is advanced.
  /// @param load A tool to measure CPU and memory usage during the simulation.
  virtual void simulate(double newtime, cLoadMeter &load);
};

};

#endif // __QUEEN_H
