/*
Copyright (C) 2007, 2008  Condel project, TU-Berlin.
Written by Jan Sablatnig, Sven Grottke, Jiehua Chen.

This file is part of Adam, the DVE testbed.

Adam is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3, or (at your option) any later
version.

Adam is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Adam; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.
*/
#include "swarm_scenario.hh"
#include "message_router.hh"
#include "instance_factory.hh"
#include "queen.hh"
#include "score.hh"
#include "world.hh"
#include "consistency.hh"



// if defined, will render more parameters (BIG!)
//#define RENDER_COMPLETE


// Time dead after colliding with the queen
#define DEAD_TIME_QUEEN 6.0
#define DEAD_TIME_BEE 3.0


// Actions per minute, pro-gamers can go to 200, even 300, normal people
// are usually at <50.
#define AI_APM 120

// If the next decision is similar within this many % of the old decision's
// abs value, AND the condition below applies, the new decision is skipped.
#define AI_IGNORE_ABS_REL 0.25

// If the next decision is similar within this much of the old decision's
// abs value, AND the condition above applies, the new decision is skipped.
#define AI_IGNORE_ABS_DIF 0.5

// If the next decision is similar within this many degrees of the old
// abs value, the new decision is skipped.
// About 11.5 degrees.
#define AI_IGNORE_PHI 0.15

// Ever AI_RELEARN seconds, adjust the safety parameters (lowering values)
#define AI_RELEARN 20

#define AI_RELEARN_PRECISION 0.000001

// When I hit another bee/queen, increase the safety for that object by
// AI_PENALTY * DEAD_TIME_XXX.
#define AI_PENALTY 3

// Every AI_RELEARN time, reduce safety by AI_AUDACITY.
// Therefore, in the stable case, there will on average be one collision
// every AI_RELEARN * AI_PENALTY *DEAD_TIME_XXX / AI_AUDACITY s.
#define AI_AUDACITY 1

// Weight on how many seconds before crashing into the queen I should react.
#define QUEEN_KAPPA 3
// Try to stop this far before crashing into the queen
#define QUEEN_SAFETY 15
// How important collision avoidance with the queen is
#define QUEEN_MAXFAC 20

// Weight on how many seconds before crashing into other bees I should react.
#define BEE_KAPPA 3
// Try to stop this far before crashing into other bees
#define BEE_SAFETY 15
// How important collision avoidance with other bees is
#define BEE_MAXFAC 10


// The speed with which a bee is pushed out of another bee in case of a sticky
// collision.  This is a bug-workaround.
#define COLLISION_PUSHOUT 3


// Scoring is a ramp, starting at 1 (per second) close to queen, going down
// linearly to this point, where it becomes 0.
#define SCORE_MAXDIST 200

// init constants
const double cSwarmScenario::LEFT_BORDER = 0;
const double cSwarmScenario::RIGHT_BORDER = 1000;
const double cSwarmScenario::TOP_BORDER = 0;
const double cSwarmScenario::BOTTOM_BORDER = 500;

// initialize statics
OBJECT_ID cSwarmScenario::s_queenID = cObjectFactory::registerObject("cQueen#");
map<unsigned long, string> cSwarmScenario::s_beeName;
map<unsigned long, string> cSwarmScenario::s_scoreName;


cSwarmScenario::cSwarmScenario(cHost& host)
: cScenario(host), m_doRender(false), m_queenVec(), m_beeVec(), m_deadVec(),
  m_scoreVec(), m_lastScore(0.0), m_scoreCounter(0.0), m_actionCount(-1),
  m_safetyQueen(QUEEN_SAFETY), m_safetyBee(BEE_SAFETY), m_lastDead(0.0),
  m_lastDecisionAbs(0.0), m_lastDecisionPhi(0.0)
{ }


void cSwarmScenario::registerHost(unsigned long id)
{
  assert(s_beeName.find(id) ==s_beeName.end());
  assert(s_scoreName.find(id) ==s_scoreName.end());

  char bee_name[100];
  char score_name[100];
  snprintf(bee_name, 100, "cBee#%lx", id);
  snprintf(score_name, 100, "cScore#%lx", id);
  
  cObjectFactory::registerObject(bee_name);
  s_beeName[id]= bee_name;
  cObjectFactory::registerObject(score_name);
  s_scoreName[id] = score_name;
}


void cSwarmScenario::initialize(void)
{
  m_host.m_attributeNorm[cAttributeFactory::nameToID("cBee::positionX")]=
                1.0 / (RIGHT_BORDER - LEFT_BORDER);
  m_host.m_attributeNorm[cAttributeFactory::nameToID("cBee::positionY")]=
                1.0 / (BOTTOM_BORDER - TOP_BORDER);
  m_host.m_attributeNorm[cAttributeFactory::nameToID("cBee::velocityAbs")]=
                1.0 / cBee::MAX_V;
  m_host.m_attributeNorm[cAttributeFactory::nameToID("cBee::velocityPhi")]=
                1.0 / 2 / M_PI;
  m_host.m_attributeNorm[cAttributeFactory::nameToID("cBee::accelerationAbs")]=
                1.0 / cBee::ACCELERATION;
  m_host.m_attributeNorm[cAttributeFactory::nameToID("cBee::accelerationPhi")]=
                1.0 / 2 / M_PI;
  m_host.m_attributeNorm[cAttributeFactory::nameToID("cQueen::positionX")]=
                1.0 / (RIGHT_BORDER - LEFT_BORDER);
  m_host.m_attributeNorm[cAttributeFactory::nameToID("cQueen::positionY")]=
                1.0 / (BOTTOM_BORDER - TOP_BORDER);
  m_host.m_attributeNorm[cAttributeFactory::nameToID("cQueen::velocityAbs")]=
                1.0 / cBee::MAX_V;
  m_host.m_attributeNorm[cAttributeFactory::nameToID("cQueen::velocityPhi")]=
                1.0 / 2 / M_PI;
  m_host.m_attributeNorm[cAttributeFactory::nameToID("cQueen::accelerationAbs")]=
                1.0 / cBee::ACCELERATION;
  m_host.m_attributeNorm[cAttributeFactory::nameToID("cQueen::accelerationPhi")]=
                1.0 / 2 / M_PI;

  m_host.m_attributeNorm[cAttributeFactory::nameToID("cScore::score")]=
                1.0;

  m_host.m_attributeNorm[cAttributeFactory::nameToID("cBee::dead")]=
                1.0;
  m_host.m_attributeNorm[cAttributeFactory::nameToID("cQueen::seed")]=
                0.0;

  m_lastScore= 0;
  m_scoreCounter= 0;
  m_actionCount= -1;
  m_safetyBee= BEE_SAFETY;
  m_safetyQueen= QUEEN_SAFETY;
  m_lastDead= 0.0;

  // The first player starts off the Queen
  if (host().hasPar("start_queen") && host().par("start_queen"))
    {
    // create, initialize and add the queen
    scoped_ptr<cQueen> pQueen((cQueen*) cInstanceFactory::
                            createInstance(s_queenID, "cQueen#"));
    R_2 queenPos;
    queenPos[0] = uniform(LEFT_BORDER +10*cBee::RADIUS, RIGHT_BORDER -10*cBee::RADIUS);
    queenPos[1] = uniform(TOP_BORDER +10*cBee::RADIUS, BOTTOM_BORDER -10*cBee::RADIUS);
    pQueen->setPosition(queenPos);
    R_2 queenVel;
    queenVel[0] = 0;
    queenVel[1] = 0;
    pQueen->setVelocity(queenVel);
    R_2 queenAcc;
    queenAcc[0] = 0;
    queenAcc[1] = 0;
    pQueen->setAcceleration(queenAcc);
    host().getConsistency().addInstance(*pQueen.get());
#if VERBOSITY >= 3
    ev << "initial queen position on " << m_host.name() << " is ("
       << pQueen->getPosition()[0] << ", " << pQueen->getPosition()[1] << ")\n";
#endif
    }

  // create, initialize and add the bee for this player
  unsigned long hid= host().id();
  assert(s_beeName.find(hid) != s_beeName.end());
  const string &bee_n= s_beeName[hid];
  scoped_ptr<cBee> pBee((cBee *) cInstanceFactory::
               createInstance(cObjectFactory::nameToID(bee_n), bee_n));
  R_2 beePos;
  beePos[0] = uniform(LEFT_BORDER +10*cBee::RADIUS, RIGHT_BORDER -10*cBee::RADIUS);
  beePos[1] = uniform(TOP_BORDER +10*cBee::RADIUS, BOTTOM_BORDER -10*cBee::RADIUS);
  pBee->setPosition(beePos);
  cPolar beeVel;
  beeVel.setXY(0.0, 0.0);
  pBee->setVelocity(beeVel);
  cPolar beeAcc;
  beeAcc.setXY(0.0, 0.0);
  pBee->setAcceleration(beeAcc);
  host().getConsistency().addInstance(*pBee.get());

  // create, initialize and add the score
  assert(s_scoreName.find(hid) != s_scoreName.end());
  const string &score_n= s_scoreName[hid];
  scoped_ptr<cScore> pScore((cScore*) cInstanceFactory::
               createInstance(cObjectFactory::nameToID(score_n), score_n));
  pScore->setScore(0);
  host().getConsistency().addInstance(*pScore.get());

  if (host().hasPar("render") && host().par("render"))
    initRendering();
  else
    m_doRender= false;
}

static double BD2= 4* cBee::RADIUS * cBee::RADIUS;

void cSwarmScenario::collision_analyze(const cBee &bee, const cBee &other,
        double &time, cPolar &v) const
{
  // 0:  Calulate approaching bee as seen by this bee
  R_2 opos= other.getPosition() - bee.getPosition();
  cPolar ovel= other.getVelocity() - bee.getVelocity();

  if (ovel.abs() < 0.0000001)
    { ovel.setAbs(0.0000001); }                 // prevent division by zero

  // 1:  Calulate closest approximation distance.  Uses the Hessian Normal form.
  // Note this INCORRECTLY assumes that the motion was linear before this.
  // In reality, both bee or other could have bounced off a wall, then this
  // calculation may yield a point where the bee may have never been at.
  // This is very hard to prevent, so we'll leave it as it is now.
  cPolar n;
  n.setXY(-ovel.y(), ovel.x());
  n.setAbs(1);
  double b= n * opos;           // b may be <0, but fabs(b) < cBee::Radius*2

  // 2:  Calculate point and time of impact
  double d= sqrt(BD2 - b*b);
  R_2 pimp= b * n - d / ovel.abs() * ovel;   // b*n is point B
  cPolar spdiff(opos-pimp);
  double tpast= 0.0;
  double dirdiff= spdiff.phi() - ovel.phi();        // should be 0
  if (fabs(dirdiff) < M_PI)
    tpast= spdiff.abs() / ovel.abs();
  if (tpast < time)
    {   // normal case
    time= tpast;                                        // RESULT

    // 3:  Calculate effect of impact on the current bee
    // The angle in the collision triangle (of center_bee, impact, B)
    // at center_bee is alpha, alpha= arccos(b / (2 * cBee::RADIUS));
    // v_perp = ovel * sin alpha
    double v_perp= ovel.abs() * sqrt(1 - b*b / BD2);
    v= pimp * (-1);
    v.setAbs(v_perp);                                  // RESULT
    }
  else
    {   // a ``stuck inside'' case
    time= 0;                                           // RESULT
    v = opos * (-1);
    v.setAbs(COLLISION_PUSHOUT);
    }
}

void cSwarmScenario::simulate(cWorld& world, double newtime,
        cLoadMeter &load) const
{
#if VERBOSITY >= 2
  ev << "handling world \"" << world.getName() << "\" on host "
     << m_host.name() << "\n";
#endif

  double delta= newtime- world.simTime();
  if (delta <=0.0) return;

  // simulate the instances
  cWorld::iterator instIter;

  for (instIter = world.begin(); instIter != world.end(); instIter++)
    (*instIter).second->simulate(newtime, load);

  cQueen *pQueen= NULL;
  if (world.isReplicated(s_queenID))
    pQueen= (cQueen *)&world.getInstance(s_queenID);
        
  // Go through all players, checking score and collisions
  list<unsigned long>::const_iterator hit;
  for(hit= host().getRouter().beginHID(); hit!= host().getRouter().endHID();
      hit++)
    {
    if (s_beeName.find(*hit) != s_beeName.end()
        && world.isReplicated(cObjectFactory::nameToID(s_beeName[*hit])))
      {                                                 // The bee exists
      cBee &bee= *(cBee *)&world.getInstance(
                cObjectFactory::nameToID(s_beeName[*hit]));
        

      // Collision with other bees?
      list<unsigned long>::const_iterator cit;
      for(cit= host().getRouter().beginHID(); cit!= host().getRouter().endHID();
          cit++)
        {
        if (*cit == *hit) continue;
        
        if (s_beeName.find(*cit) != s_beeName.end()
              && world.isReplicated(
              cObjectFactory::nameToID(s_beeName[*cit])))
          {
          cBee &other= *(cBee *)&world.getInstance(
                cObjectFactory::nameToID(s_beeName[*cit]));
          R_2 opos= other.getPosition() - bee.getPosition();

          if (opos[0]*opos[0] + opos[1]*opos[1] < BD2)
            {
            double timp= delta;
            cPolar v_imp;
            collision_analyze(bee, other, timp, v_imp);
            
            // Set the bee's state back to the time of impact, adjust speed,
            // resimulate it.
            if (bee.isAlive())
              bee.setDead(DEAD_TIME_BEE);
            bee.setPosition(bee.getPosition() - bee.getVelocity()* timp);
            bee.setSimTime(newtime -timp);
            
            // Both bees have the same mass, so we can just add v_imp to each
            // speed (but with a different sign, of course).
            bee.setVelocity(bee.getVelocity() + v_imp);
            bee.simulate(newtime, load);
             
            // Same for the OTHER bee
            if (other.isAlive())
              other.setDead(DEAD_TIME_BEE);
            other.setPosition(other.getPosition() - other.getVelocity()* timp);
            other.setSimTime(newtime -timp);
            
            other.setVelocity(other.getVelocity() - v_imp);
            other.simulate(newtime, load);

            load.addCPU(26);
            }
          load.addCPU(5);
          }
        }

      // Collision with the queen?
      R_2 opos;
      if (pQueen)
        opos= pQueen->getPosition() - bee.getPosition();
      if (pQueen && opos[0]*opos[0] + opos[1]*opos[1] < BD2)
        {
        double timp= delta;
        cPolar v_imp;
        collision_analyze(bee, *pQueen, timp, v_imp);

        // Set the bee's state back to the time of impact, adjust speed,
        // resimulate it.
        if (bee.isAlive())
          bee.setDead(DEAD_TIME_QUEEN);
        bee.setPosition(bee.getPosition() - bee.getVelocity()* timp);
        bee.setSimTime(newtime -timp);

        // The queen is assumed to have m -> inf, so it's just like a
        // collision with a wall:  v_new = - v_old.
        bee.setVelocity(bee.getVelocity() + 2 * v_imp);
        bee.simulate(newtime, load);

        load.addCPU(20);
        }

      // Scoring
      if (bee.isAlive()&& pQueen && s_scoreName.find(*hit) != s_scoreName.end()
        && world.isReplicated(
                cObjectFactory::nameToID(s_scoreName[*hit])))
        {                                               // score too!
        cScore &score=  *(cScore *)&world.getInstance(
                cObjectFactory::nameToID(s_scoreName[*hit]));
        
        double honey=(SCORE_MAXDIST -pQueen->getDistance(bee))
                        /(SCORE_MAXDIST -2*cBee::RADIUS) *delta;
        if (honey <0.0)
          honey= 0.0;
        score.setScore(score.getScore() +honey);
        }

      load.addCPU(7);
      }
    }

  world.setSimTime(newtime);
}

void cSwarmScenario::ai(double newtime)
{
  // call the pseudo-renderer
  if (m_doRender)
    render();

  unsigned long hid= host().id();

  if (host().getWorld().isReplicated(
                cObjectFactory::nameToID(s_scoreName[hid])))
    {
    double score= ((cScore *) &host().getWorld().getInstance(
        cObjectFactory::nameToID(s_scoreName[hid])))->getScore();
    if (score > m_lastScore)
      m_scoreCounter+= score-m_lastScore;
    m_lastScore= score;
    }

  // Can't do anything if I don't see the queen
  if (!host().getWorld().isReplicated(s_queenID))
    return;

  // Or the bee itself
  if (!host().getWorld().isReplicated(
                cObjectFactory::nameToID(s_beeName[hid])))
    return;

  OBJECT_ID bee_id= cObjectFactory::nameToID(s_beeName[hid]);
  cBee &bee_c= *(cBee *)&host().getWorld().getInstance(bee_id);


  if (bee_c.getDead() > m_lastDead +1.0) // +1 to avoid firing from minor incont. 
    {
    if (bee_c.getDead() +1.0 > DEAD_TIME_QUEEN)
      m_safetyQueen += AI_PENALTY * DEAD_TIME_QUEEN;
    else
      m_safetyBee += AI_PENALTY * DEAD_TIME_BEE;
    }
  m_lastDead = bee_c.getDead();

  if (fmod(newtime+AI_RELEARN_PRECISION,AI_RELEARN) < 2*AI_RELEARN_PRECISION)
    {
    m_safetyQueen -= AI_AUDACITY;
    m_safetyBee -= AI_AUDACITY;
    }

  // Don't try to do anything if I believe myself to be dead
  // Note this is still based on CURRENT time
  if (!bee_c.isAlive())
    return;

  long new_action_count= lrint(floor(newtime * AI_APM / 60.0 ));

  // Ai works only every action time.
  if (new_action_count <= m_actionCount)
    return;

  m_actionCount= new_action_count;

  // FUTURE, DELAY
  // if the prog is being run with a local delay, the ai has to look into the
  // future for decisions.

  cWorld &world= host().getWorld();
  cWorld *future= &world;
  auto_ptr<cWorld> future_changed;
  if (m_delayHint > 0.0)
    {
    future_changed= future->clone();
    future= future_changed.get();
    cLoadMeter fake;
    future->simulate(host().now() + m_delayHint, fake);
    }

  cQueen &queen= *(cQueen *)&future->getInstance(s_queenID);
  cBee &bee= *(cBee *)&future->getInstance(bee_id);


  // First, some basic info
  // The relative position of the queen from this position
  cPolar queen_pos(queen.getPosition() - bee.getPosition());
  // My speed in the view where the queen is static
  cPolar rel_vel(bee.getVelocity() - queen.getVelocity());

  // Speed will be changed this much during one action
  double dv= cBee::ACCELERATION * AI_APM / 60.0;

  cPolar final;
  final.setAbs(cBee::ACCELERATION);

  // Step 1:  Accelerate towards the queen
  // -------------------------------------
  bool turn_left= true;
  double phi= cPolar::normalize_phi(queen_pos.phi() - rel_vel.phi());
  if (phi < 0.0)
    {
    turn_left= false;
    phi= -phi;
    }
  double a= rel_vel.abs() * tan(phi);

  // There are three cases:
  if (a > 0 && a >= dv)
    // a) We just need to turn.
    //    The most effective turning method is going 90 deg to the side
    final.setPhi(rel_vel.phi() + (turn_left ? M_PI/2 : -M_PI/2 ));
  else if (a > 0 && a < dv)
    {
    // b) Turning like that would overturn
    //    Calculate the hardest turn that will just NOT overturn
    double beta= M_PI / 2.0 - phi - asin(a/dv *cos(phi));
    final.setPhi(rel_vel.phi() + (turn_left ? M_PI/2 -beta : -M_PI/2 +beta ));
    }
  else          // a <  0
    {
    // c) We're currently moving AWAY from our target.  We'll just accelerate
    // towards it, regardless of our current movement.
    final.setPhi(queen_pos.phi());
    }

  // Step 2:  Avoid collision with the queen
  // ---------------------------------------
  double move_rel= rel_vel.abs() * cos(phi);

  if (move_rel > 0.0)
    {
    double min_stop_dist= move_rel * move_rel / cBee::ACCELERATION / 2.0;

    double remaining_time= (queen_pos.abs() -cBee::RADIUS *2 -m_safetyQueen
                      -min_stop_dist) / move_rel;

    double factor= QUEEN_MAXFAC;
    if (remaining_time > 0.0)
      {
      factor= QUEEN_KAPPA/remaining_time;
      if (factor > QUEEN_MAXFAC)
        factor= QUEEN_MAXFAC;
      }

    cPolar breaking;
    breaking.setPhiAbs(queen_pos.phi()+ M_PI, cBee::ACCELERATION);
    final += breaking * factor;
    }

  // Step 3:  Avoid collision with other bees
  // ----------------------------------------
  list<unsigned long>::const_iterator cit;
  for(cit= host().getRouter().beginHID(); cit!= host().getRouter().endHID();
           cit++)
    {
    if (*cit == hid) continue;
        
    if (s_beeName.find(*cit) != s_beeName.end()
              && future->isReplicated(
              cObjectFactory::nameToID(s_beeName[*cit])))
      {
      cBee &other= *(cBee *)&future->getInstance(
                   cObjectFactory::nameToID(s_beeName[*cit]));

      cPolar other_pos= other.getPosition() - bee.getPosition();

      // TODO:  This next line should read other.getVelocity().
      //        It currently reads queen.getVelocity(), which is wrong!
      // However, this current version works BETTER (higher yield) than
      // the correct version.  Possible reasons:
      //   * queen-velocity is the group (swarm) velocity.
      //   * using a group-velocity makes bees less likely to let other
      //     bees force their way through them.
      // The only time this is bad is when a bee is dead and moves
      // significantly different from the others.  Other bees ignore this
      // and collide unneccessarily.
      rel_vel= bee.getVelocity() - queen.getVelocity();
          
      phi= cPolar::normalize_phi(other_pos.phi() - rel_vel.phi());

      move_rel= rel_vel.abs() * cos(phi);

      if (move_rel > 0.0)
        {
        double min_stop_dist= move_rel * move_rel / cBee::ACCELERATION / 2.0;
        double remaining_time= (other_pos.abs() - cBee::RADIUS *2 - m_safetyBee
                        - min_stop_dist) / move_rel;

        // TODO:  Might use abs-distance here, as well
        double factor= BEE_MAXFAC;
        if (remaining_time > 0.0)
          {
          factor= BEE_KAPPA/remaining_time;
          if (factor > BEE_MAXFAC)
            factor= BEE_MAXFAC;
          }

        cPolar breaking;
        breaking.setPhiAbs(other_pos.phi()+ M_PI, cBee::ACCELERATION);
        final += breaking * factor;
        }
      }
    }


  if (final.abs() > cBee::ACCELERATION)
    final.setAbs(cBee::ACCELERATION);

  // Check if this is the same as before
  if ((((m_lastDecisionAbs != 0.0 ||
        (m_lastDecisionAbs == 0.0 && final.abs() == 0.0))
      && final.abs()/m_lastDecisionAbs > (1 - AI_IGNORE_ABS_REL)
      && final.abs()/m_lastDecisionAbs < (1 + AI_IGNORE_ABS_REL))
      || fabs(final.abs() - m_lastDecisionAbs) < AI_IGNORE_ABS_DIF)
      && fabs(final.phi() - m_lastDecisionPhi) < AI_IGNORE_PHI)

        return;

  m_lastDecisionAbs= final.abs();
  m_lastDecisionPhi= final.phi();

  host().getConsistency().changeAttribute(bee_id, cBee::s_accelerationAID,
        final.abs());
  host().getConsistency().changeAttribute(bee_id, cBee::s_accelerationPID,
        final.phi());
}

void cSwarmScenario::initRendering(void)
{
  m_doRender= true;

  m_queenVec.posX.setName("queenPosX");
  m_queenVec.posY.setName("queenPosY");
#ifdef RENDER_COMPLETE
  m_queenVec.velX.setName("queenVelX");
  m_queenVec.velY.setName("queenVelY");
  m_queenVec.accX.setName("queenAccX");
  m_queenVec.accY.setName("queenAccY");
#endif

  list<unsigned long>::const_iterator hit;
  for(hit= host().getRouter().beginHID(); hit!= host().getRouter().endHID();
      hit++)
    {
    m_beeVec[*hit].reset(new cRenderVec);
    char bee_name[100];

    snprintf(bee_name, 100, "cBee@%lx.PosX", *hit);
    m_beeVec[*hit]->posX.setName(bee_name);
    snprintf(bee_name, 100, "cBee@%lx.PosY", *hit);
    m_beeVec[*hit]->posY.setName(bee_name);
#ifdef RENDER_COMPLETE
    snprintf(bee_name, 100, "cBee@%lx.VelX", *hit);
    m_beeVec[*hit]->velX.setName(bee_name);
    snprintf(bee_name, 100, "cBee@%lx.VelY", *hit);
    m_beeVec[*hit]->velY.setName(bee_name);
    snprintf(bee_name, 100, "cBee@%lx.AccX", *hit);
    m_beeVec[*hit]->accX.setName(bee_name);
    snprintf(bee_name, 100, "cBee@%lx.AccY", *hit);
    m_beeVec[*hit]->accY.setName(bee_name);
#endif
    
    m_deadVec[*hit].reset(new cOutVector);
    snprintf(bee_name, 100, "cBee@%lx.Dead", *hit);
    m_deadVec[*hit]->setName(bee_name);

    m_scoreVec[*hit].reset(new cOutVector);
    char score_name[100];
    snprintf(score_name, 100, "cScore@%lx", *hit);
    m_scoreVec[*hit]->setName(score_name);
    }
}

void cSwarmScenario::render(void)
{
  if (host().getWorld().isReplicated(s_queenID))
    {
    cQueen &queen= *(cQueen *)&host().getWorld().getInstance(s_queenID);
    const R_2& queenPos = queen.getPosition();
    m_queenVec.posX.record(queenPos[0]);
    m_queenVec.posY.record(queenPos[1]);
#ifdef RENDER_COMPLETE
    const R_2& queenVel = queen.getVelocity();
    const R_2& queenAcc = queen.getAcceleration();
    m_queenVec.velX.record(queenVel[0]);
    m_queenVec.velY.record(queenVel[1]);
    m_queenVec.accX.record(queenAcc[0]);
    m_queenVec.accY.record(queenAcc[1]);
#endif
    }

  list<unsigned long>::const_iterator hit;
  for(hit= host().getRouter().beginHID(); hit!= host().getRouter().endHID();
      hit++)
    {
    if (s_beeName.find(*hit) != s_beeName.end()
              && host().getWorld().isReplicated(
              cObjectFactory::nameToID(s_beeName[*hit])))
      {
      cBee &bee= *(cBee *)&host().getWorld().getInstance(
                cObjectFactory::nameToID(s_beeName[*hit]));
      const R_2& beePos = bee.getPosition();
      m_beeVec[*hit]->posX.record(beePos[0]);
      m_beeVec[*hit]->posY.record(beePos[1]);
#ifdef RENDER_COMPLETE
      const R_2& beeVel = bee.getVelocity();
      const R_2& beeAcc = bee.getAcceleration();
      m_beeVec[*hit]->velX.record(beeVel[0]);
      m_beeVec[*hit]->velY.record(beeVel[1]);
      m_beeVec[*hit]->accX.record(beeAcc[0]);
      m_beeVec[*hit]->accY.record(beeAcc[1]);
#endif

      m_deadVec[*hit]->record(bee.getDead());
      }

    if (s_scoreName.find(*hit) != s_scoreName.end()
        && host().getWorld().isReplicated(
                cObjectFactory::nameToID(s_scoreName[*hit])))
      {
      cScore &score=  *(cScore *)&host().getWorld().getInstance(
                cObjectFactory::nameToID(s_scoreName[*hit]));
      m_scoreVec[*hit]->record(score.getScore());
      }

    }
}

void cSwarmScenario::finish(void)
{
  m_host.recordScalar("score achieved /s", m_scoreCounter/host().now());

  host().getConsistency().finish();
}

