/*
Copyright (C) 2007, 2008  Condel project, TU-Berlin.
Written by Jan Sablatnig, Sven Grottke, Jiehua Chen.

This file is part of Adam, the DVE testbed.

Adam is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3, or (at your option) any later
version.

Adam is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Adam; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.
*/

#ifndef __SWARM_SCENARIO_H
#define __SWARM_SCENARIO_H

#include "scenario.hh"
#include "object_factory.hh"


namespace Condel {

class cRacket;
class cBall;
class cMessageRouter;
class cBee;
class cPolar;

/// Helper object for visualization.
/// This is used in dumping all data for later visualization.
/// Only used within the scenario-file.

struct cRenderVec
{
public:
  cOutVector posX;
  cOutVector posY;
#ifdef RENDER_COMPLETE
  cOutVector velX;
  cOutVector velY;
  cOutVector accX;
  cOutVector accY;
#endif
};
  
/// The Swarm Scenario.
/// This is a simple non-cooperative multi-player game.  A computer-controlled
/// queen performs a random-walk and each player attempts to stay as close
/// to the queen as possible.  Every tick, honey is awarded to each bee
/// according to their distance from the queen (shorter distance -> more honey).
/// If a bee collides with any other (or the queen), it is considered dead
/// for a short time, not collecting honey and not being able to control its
/// flight.
/// 
/// There is a physical model installed, each bee represents a fully elastic
/// solid ball.  If a bee collides with the queen, the queen is considered
/// a wall (i.e. it has infinite weight).
///
/// If the render parameter is true for this host, positions of all bees and
/// all scores will be logged to the omnetpp.vec file each frame.  This
/// can be used to create a visualization later.

class cSwarmScenario : public cScenario
{
public:
  // public constants
  static const double LEFT_BORDER;     ///< X-coordinate of the left border
  static const double RIGHT_BORDER;    ///< X-coordinate of the right border
  static const double TOP_BORDER;      ///< Y-coordinate of the upper border
  static const double BOTTOM_BORDER;   ///< Y-coordinate of the lower border


public:         // The following three public so that SwarmLooseHost can use em
  // this is ugly, but fast
  static OBJECT_ID s_queenID;

  static map<unsigned long, string> s_beeName;
  static map<unsigned long, string> s_scoreName;

protected:
  // for output to the visualisation interface
  bool m_doRender;
  cRenderVec m_queenVec;
  map<unsigned long, shared_ptr<cRenderVec> > m_beeVec;
  map<unsigned long, shared_ptr<cOutVector> > m_deadVec;
  map<unsigned long, shared_ptr<cOutVector> > m_scoreVec;
  void initRendering(void);
  void render(void);

  // Finds time, place of impact and the impact speed vector (of the other
  // bee onto this).
  // The value GIVEN to the function in time is the max time looked at.
  void collision_analyze(const cBee &me, const cBee &other,
        double &time, cPolar &v) const;

  double m_lastScore;            // Last seen score for myself in ai()
  double m_scoreCounter;         // Count scores I see for MYSELF

  long m_actionCount;         // The numeral of the last action taken

  double m_safetyQueen;         // How close to get to the queen in pels
  double m_safetyBee;           // How close to get to other bees, in pels
  double m_lastDead;            // last seen dead state

  double m_lastDecisionAbs;     // Last decision made by AI.  Abs part.
  double m_lastDecisionPhi;     // Last decision made by AI.  Phi Part.

public:
  cSwarmScenario(cHost& host);
  virtual ~cSwarmScenario(void) {}

  /// Register a new host/player/bee.
  /// This is a cludge to allow SwarmLooseHost to find the relevant other
  /// hosts fast&easy.  Should soon be replaced by a true join/leave
  /// function.
  /// @param id The id of the newly joined host.
  virtual void registerHost(unsigned long id);

  /// Reset everything.
  /// Create your own bee and score.
  /// One player (controlled by start_queen parameter) creates the queen.
  void initialize(void);

  /// Perform deterministic actions.
  /// Let each instance in the world simulate to newtime (moves ball/rackets).
  /// Then, check for collisions (a-posteriori).  Resolve them if necessary.
  /// Award honey to each bee.
  /// @param world The world that is advanced.
  /// @param newtime Time to advance the system to.  This should be in the
  /// future of the world's cWorld::simTime().  Given in seconds.
  /// @param load A tool to measure CPU and memory usage during the simulation.
  virtual void simulate(cWorld& world, double newtime, cLoadMeter &load)
        const;

  /// Perform the AI.
  /// A form of potential field/gradient steering is performed to move towards
  /// the queen but stay away from all obstacles (including the queen itself).
  /// 
  /// A simple learning algorithm is installed to adjust to the level of
  /// collisions that occur.  The algorithm adjusts the safety boarders
  /// in the steering algorithm.
  /// @param newtime The current time.
  virtual void ai(double newtime);

  /// System Done.
  /// Write the score (total honey) to the log file (it's the application
  /// dependant yield measure).
  virtual void finish(void);
};
  
};

#endif // __SWARM_SCENARIO_H
