#!/usr/bin/perl
# Copyright (C) 2007, 2008  Condel project, TU-Berlin.
# Written by Jiehua Chen, Sven Grottke, Jan Sablatnig.
# 
# This file is part of Adam, the DVE testbed.
# 
# Adam is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3, or (at your option) any later
# version.
# 
# Adam is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
# 
# You should have received a copy of the GNU General Public License
# along with Adam; see the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.
# 
# Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.


# usage: plot_timeout.pl filename min_timeout max_timeout [-lx] [-ly] HOST1 measurement1 HOST2 measurement2 ...
#############################################################
#                       HOSTs are:
#
# "timeoutNetwork.S"                  "timeoutNetwork.R"
#
#############################################################
#                measurements are:
#
#"N" (number of data packets sent)
#"A" (number of acknowledgements sent)

use strict;

sub usage{
     print STDERR "usage: plot_timeout.pl filename min_timeout max_timeout [-lx] [-ly] HOST1 measurement1 HOST2 measurement2 ... .\n";
}

sub failure{
    my $failedparm = shift;
        
    if ( defined $failedparm ){
        print STDERR "$failedparm format is wrong\n";
    }
    else{
        print STDERR "parameters failed!\n";
    }
    &usage;
    print STDERR "Please see the usage for more details.\n";
    die "aborting.";
}

my @QUOTED_ARGV = @ARGV;
map { s/(\W)/\\$1/g; } @QUOTED_ARGV;

# parse parameters
my $argc = 0;

# start with mandatory paramaters: filename min_timeout max_timeout parms
if ( @ARGV - $argc < 3 ) { &failure; }
my $scafile = <$QUOTED_ARGV[$argc++]>; #.sca file with various runs data.
if ( $ARGV[$argc] !~ /^[\.\d]+/ ) { &failure("min_timeout"); }
my $min_timeout = $ARGV[$argc++];   # min_timeout
if ( $ARGV[$argc] !~ /^[\.\d]+/ ) { &failure("max_timeout"); }
my $max_timeout = $ARGV[$argc++];     # max_timeout

#note if plot with x logscale
my $with_xlogscale = 0;
if ( $ARGV[$argc] eq "-lx" ){
    $with_xlogscale = 1;
    $argc++;
}

#note if plot with y logscale
my $with_ylogscale = 0;
if ( $ARGV[$argc] eq "-ly" ){
    $with_ylogscale = 1;
    $argc++;
}

# the number of hosts and measurements should be even 
if ( (@ARGV - $argc) % 2 != 0 ){ &failure; }
my $numberOfHosts = ( @ARGV - $argc ) / 2;
my (@measurements,@hosts);
my ($network,$host,$scenario);

my (@plotfiles,@sca_lines);
my @parms;
my ($x_value,$y_value);

# store the weights of variance and discontinuity
my $varianceWeights;
my $discoWeights;
my $countWeights=0;

# grep parmtitles
my $parm_defined=0;

for( my $i = 0 ; $i < $numberOfHosts ; $i++ ){
    my $run_count = 0;
    open(SCAFILE,$scafile);
    push(@plotfiles,"./plot$i.data"); #file for the output
    open(PLOTFILE,">$plotfiles[$i]");

    #split the name into network and host
    $QUOTED_ARGV[$argc+2*$i] =~ /^(\w+)\\\.(.+)$/ || die "wrong format of hostname\n";
    $network = $1;
    $hosts[$i] = $2;

    $measurements[$i] = $QUOTED_ARGV[$argc+2*$i+1];

    @sca_lines = <SCAFILE>;
    #find the accordance x- & y-value in scafile 
    #then copy them into the plotfiles 
    foreach( @sca_lines ){
	if ( /^run/ ){
            $parm_defined++;
            if( $run_count >= 1 ){
                print PLOTFILE "$x_value  $y_value\n";
            }
	    $run_count++;
	    next;
	}
	
	#find the y-value for a consistency case
	if ( /\"${network}\.${hosts[$i]}\"\s+\"${measurements[$i]}\"\s+([\.\d\-e]+)/ ){
	    $y_value = $1;
            next;
	}
	
	#find the x-value for a consistency case
	if ( /\"timeout\"\s+([\.|\d]+)/ ){
	    $x_value = $1;
#            print STDERR "${x_value}\n";
	    next;
	}
        
	#filter the parameter values 
	if ( $parm_defined == 1 ){
            if (/\"parm\"\s+\"(.+)\"\s+([\.\d\-e]+)/){
                if ( "$1" ne "timeout" ){
                    push(@parms,($1, $2, "\\n"));
                }
            }
	}
    }

    print PLOTFILE "$x_value  $y_value\n";

    close(PLOTFILE);
    close(SCAFILE);
}

#write the script for the plot
my $plotscript = <./plotscript.gnuplot>;
open(PLOTSCRIPT,">$plotscript");
my $time = localtime;
print PLOTSCRIPT "set term postscript colour\n";
print PLOTSCRIPT "set xlabel \"timeout\"\n";
print PLOTSCRIPT "set ylabel \"$measurements[0]\"\n";
print PLOTSCRIPT "set title \" $time @parms\n";
print PLOTSCRIPT "set yrange [0:]\n";
#print PLOTSCRIPT "set xtics rotate\n";
if ( $with_xlogscale ){
    print PLOTSCRIPT "set logscale x 2\n";
    print PLOTSCRIPT "set xrange [1:$max_timeout]\n";
}else{
    print PLOTSCRIPT "set xrange [$min_timeout:$max_timeout]\n";
}
if ( $with_ylogscale ){ print PLOTSCRIPT "set logscale y 2\n";}
print PLOTSCRIPT "set key below\n";
print PLOTSCRIPT "set grid\n";
#sort the plotdatas
system("sort -n $plotfiles[0] > $plotfiles[0].sorted");
print PLOTSCRIPT "plot \"$plotfiles[0].sorted\" title \"$measurements[0]\" with lines";
for( my $i = 1 ; $i < $numberOfHosts ; $i++ ){
    system("sort -n $plotfiles[$i] > $plotfiles[$i].sorted");
    print PLOTSCRIPT ", \"$plotfiles[$i].sorted\" title \"$measurements[$i]\" with lines";
}

print PLOTSCRIPT "\n";
close(PLOTSCRIPT);

#execute
system("gnuplot < $plotscript ");

#remove
system("rm $plotscript");
for( my $i = 0 ; $i < $numberOfHosts ; $i++ ){
   system("rm $plotfiles[$i]");
   system("rm $plotfiles[$i].sorted");
}

