// plot E(N) over T_O for a given distribution function
// usage: E_N --dist step|exp|pareto <expectation> [<arg>] --min <min>
//            --max <max> --stepsize <step> --mode prec|approx|P_RTT
//            --loss-data <P_L_D> --loss-ack <P_L_A> 
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include "distrib.h"

#define L (1. - (1. - P_L_D) * (1. - P_L_A))

double P_L_D;
double P_L_A;

DistribFunction* pDistRTT;

void usage() {
  fprintf(stderr,
          "usage: E_N --dist step|exp|pareto <expectation> [<arg>]\n"
          "       --min <min> --max <max> --stepsize <step>\n"
          "       --mode prec|approx|P_RTT\n"
          "-loss-data <P_L_D> --loss-ack <P_L_A>\n");
}

double P_RTT(double x) {
  return pDistRTT->P(x);
}

double P_succ(int i, double T_O, int n) {
  return (1 - L) * P_RTT((n - i) * T_O);
}

double E_N(double T_O) {
  double sum = 0;
  double add;
  
  int n = 1;
  do {
    double prod = 1;
    
    for (int i = 1; i <= n - 1; i++)
      prod *= 1 - P_succ(i, T_O, n);
    add = n * P_succ(0, T_O, n) * prod;
    sum += add;
    n++;
  } while (n < 100);
  
  return sum;
}

double E_N_approx(double T_O) {
  double sum = 0;
  double add;
  
  int n = 1;
  do {
    double prod = 1;
    
    for (int i = 1; i <= ((n - 1 < 3) ? n - 1 : 3); i++)
      prod *= 1 - P_succ(i, T_O, n);
    add = n * P_succ(0, T_O, n) * prod;
    sum += add;
    n++;
  } while (n < 100);
  
  return sum;
//   const double P_S = P_succ(0, T_O, 1);
//   if (fabs(P_S) == 0)
//     return 0;
//   else
//     return 1. / P_succ(0, T_O, 1);
}

int main(int argc, const char* argv[]) {
  pDistRTT = 0;
  P_L_D = 0.01;
  P_L_A = 0.01;
  double T_O_min = 0.005;
  double T_O_max = 10;
  typedef enum { PREC = 0, APPROX, P } MODE;
  MODE mode = PREC;
  double step = 0.005;
 
  int arg = 1;
  while (arg < argc) {
    if (!strcmp(argv[arg], "--dist")) {
      arg++;
      assert(arg < argc);
      if (!strcmp(argv[arg], "step")) {
        arg++; assert(arg < argc);
        // get expectation/minimum
        const double m = strtod(argv[arg], 0);
        assert(m > 0);
        pDistRTT = new Step(m);
      } else if (!strcmp(argv[arg], "exp")) {
        arg++; assert(arg < argc);
        // get expectation
        const double e = strtod(argv[arg], 0);
        assert(e > 0);
        arg++; assert(arg < argc);
        // get minimum
        const double m = strtod(argv[arg], 0);
        assert(m >= 0);
        assert(m < e);
        pDistRTT = new Exponential(m, e);
      } else if (!strcmp(argv[arg], "pareto")) {
        arg++; assert(arg < argc);
        // get expectation
        const double e = strtod(argv[arg], 0);
        assert(e > 0);
        arg++; assert(arg < argc);
        // get k
        const double k = strtod(argv[arg], 0);
        assert(k > 1 && k <= 2);
        pDistRTT = new Pareto(e, k);
      } else {
        fprintf(stderr, "unknown distribution function: %s\n", argv[arg]);
        usage();
        exit(1);
      }
    } else if (!strcmp(argv[arg], "--min")) {
      arg++; assert(arg < argc);
      T_O_min = strtod(argv[arg], 0);
      assert(T_O_min > 0);
    } else if (!strcmp(argv[arg], "--max")) {
      arg++; assert(arg < argc);
      T_O_max = strtod(argv[arg], 0);
      assert(T_O_max > 0 && T_O_max > T_O_min);
    } else if (!strcmp(argv[arg], "--stepsize")) {
      arg++; assert(arg < argc);
      step = strtod(argv[arg], 0);
      assert(step > 0);
    } else if (!strcmp(argv[arg], "--mode")) {
      arg++; assert(arg < argc);
      if (!strcmp(argv[arg], "prec")) {
        mode = PREC;
      } else if (!strcmp(argv[arg], "approx")) {
        mode = APPROX;
      } else if (!strcmp(argv[arg], "P_RTT")) {
        mode = P;
      } else {
        fprintf(stderr, "illegal mode\n");
        usage();
        exit(1);
      }
    } else if (!strcmp(argv[arg], "--loss-data")) {
      arg++; assert(arg < argc);
      P_L_D = strtod(argv[arg], 0);
      assert(P_L_D >= 0 && P_L_D <= 1);
    } else if (!strcmp(argv[arg], "--loss-ack")) {
      arg++; assert(arg < argc);
      P_L_A = strtod(argv[arg], 0);
      assert(P_L_A >= 0 && P_L_A <= 1);
    } else {
      fprintf(stderr, "illegal argument\n");
      usage();
      exit(1);
    }
    arg++;
  }
  if (!pDistRTT) {
    usage();
    exit(1);
  }


  // depending on mode, plot E(N), E(N approx), or P_RTT
  switch (mode) {
  case PREC: // plot E(N)
    {
      for (double T_O = T_O_min; T_O <= T_O_max; T_O += step) {
        fprintf(stdout, "%lf %lf\n", T_O, E_N(T_O));
        //fprintf(stdout, "%lf %lf\n", T_O, 1. / P_succ(0, T_O, 1));
      }
    }
    break;
  case APPROX: // plot E(N approx.)
    {
      for (double T_O = T_O_min; T_O <= T_O_max; T_O += step) {
        fprintf(stdout, "%lf %lf\n", T_O, E_N_approx(T_O));
      }
    }
    break;
  case P: // plot P_RTT
    {
      for (double T = T_O_min; T <= T_O_max; T += step) {
        fprintf(stdout, "%lf %lf\n", T, P_RTT(T));
      }      
    }
    break;
  default:
    throw 1;
  }
  
  delete pDistRTT;
  return 0;
}

