// plot P_delay over T for a given distribution function and T_O
// usage: P_delay --dist step|exp|pareto <expectation> [<arg>] --timeout
//                <timeout> --min <min> --max <max> --stepsize <step> --mode
//                prec|approx|inter|hybrid|P_D --loss <P_L_D> 

#include <math.h>
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>

#include "distrib.h"

double P_L_D;
DistribFunction* pDistData;

double P_D(double x) {
  return pDistData->P(x);
}

void usage() {
  fprintf(stderr,
          "P_delay --dist step|exp|pareto <expectation> [<arg>]\n"
          "        --timeout <timeout>\n"
          "        --min <min> --max <max> --stepsize <step>\n"
          "        --mode prec|approx|inter|P_D --loss <P_L_D>\n");
}

double P_T_delay(double T, double T_O) {
  if (!(T > 0)) return 0;
  
  double prod = 1;
  int n = 0;

  while (n * T_O < T) {
    prod *= 1 - (1 - P_L_D) * P_D(T - n * T_O);
    n++;
  }

  return 1 - prod;
}

double P_T_delay_approx(int M, double T_O) {
  if (M <= 0) return 0;
  
  return 1 - pow(1 - (1 - P_L_D) * P_D(T_O), M);
}

double P_T_delay_approx_inter(double T, double T_O) {
  if (!(T > 0)) return 0;

  // find the right interval
  int M = 1;
  while (T >= M * T_O) M++;

  const double T1 = (M - 1) * T_O;
  const double T2 = M * T_O;
  const double P1 = P_T_delay_approx(M - 1, T_O);
  const double P2 = P_T_delay_approx(M, T_O);

  return P1 + (T - T1) * (P2 - P1) / (T2 - T1);
}

double P_T_delay_approx_hybrid(double T, double T_O) {
  if (!(T > 0)) return 0;

  if (T < T_O) return (1 - P_L_D) * P_D(T);

  return P_T_delay_approx_inter(T, T_O);
}

int main(const int argc, const char* argv[]) {
  pDistData = 0;
  double T_O = 1;
  double T_min = 0;
  double T_max = 10;
  P_L_D = 0.01;
  typedef enum { PREC = 0, APPROX, INTER, HYBRID, P } MODE;
  MODE mode = PREC;
  double step = 0.005;
  
  int arg = 1;
  while (arg < argc) {
    if (!strcmp(argv[arg], "--dist")) {
      arg++;
      assert(arg < argc);
      if (!strcmp(argv[arg], "step")) {
        arg++; assert(arg < argc);
        // get expectation/minimum
        const double m = strtod(argv[arg], 0);
        assert(m > 0);
        pDistData = new Step(m);
      } else if (!strcmp(argv[arg], "exp")) {
        arg++; assert(arg < argc);
        // get expectation
        const double e = strtod(argv[arg], 0);
        assert(e > 0);
        arg++; assert(arg < argc);
        // get minimum
        const double m = strtod(argv[arg], 0);
        assert(m >= 0);
        assert(m < e);
        pDistData = new Exponential(m, e);
      } else if (!strcmp(argv[arg], "pareto")) {
        arg++; assert(arg < argc);
        // get expectation
        const double e = strtod(argv[arg], 0);
        assert(e > 0);
        arg++; assert(arg < argc);
        const double k = strtod(argv[arg], 0);
        assert(k > 1 && k <= 2);
        pDistData = new Pareto(e, k);
      } else {
        fprintf(stderr, "unknown distribution function: %s\n", argv[arg]);
        usage();
        exit(1);
      }
    } else if (!strcmp(argv[arg], "--timeout")) {
      arg++; assert(arg < argc);
      T_O = strtod(argv[arg], 0);
      assert(T_O > 0);
    } else if (!strcmp(argv[arg], "--min")) {
      arg++; assert(arg < argc);
      T_min = strtod(argv[arg], 0);
      assert(T_min >= 0);
    } else if (!strcmp(argv[arg], "--max")) {
      arg++; assert(arg < argc);
      T_max = strtod(argv[arg], 0);
      assert(T_max > 0 && T_max > T_min);
    } else if (!strcmp(argv[arg], "--stepsize")) {
      arg++; assert(arg < argc);
      step = strtod(argv[arg], 0);
      assert(step > 0);
    } else if (!strcmp(argv[arg], "--mode")) {
      arg++; assert(arg < argc);
      if (!strcmp(argv[arg], "prec")) {
        mode = PREC;
      } else if (!strcmp(argv[arg], "approx")) {
        mode = APPROX;
      } else if (!strcmp(argv[arg], "inter")) {
        mode = INTER;
      } else if (!strcmp(argv[arg], "hybrid")) {
        mode = HYBRID;
      } else if (!strcmp(argv[arg], "P_D")) {
        mode = P;
      } else {
        fprintf(stderr, "illegal mode\n");
        usage();
        exit(1);
      }
    } else if (!strcmp(argv[arg], "--loss")) {
      arg++; assert(arg < argc);
      P_L_D = strtod(argv[arg], 0);
      assert(P_L_D >= 0 && P_L_D <= 1);
    } else {
      fprintf(stderr, "illegal argument\n");
      usage();
      exit(1);
    }
    arg++;
  }
  if (!pDistData) {
    usage();
    exit(1);
  }

  // depending on mode, plot P_delay, approx. P_delay, interpolated
  // approx. P_delay  or P_D
  switch (mode) {
  case PREC: // plot P_delay
    {
      for (double T = T_min; T < T_max; T += step) {
        fprintf(stdout, "%lf %lf\n", T, P_T_delay(T, T_O));
      }
    }
    break;
  case APPROX: // plot approx. P_delay 
    {
      for (int M = (int) ceil(T_min / T_O); M * T_O < T_max; M++) {
        fprintf(stdout, "%lf %lf\n", M * T_O, P_T_delay_approx(M, T_O));
      }
    }
    break;
  case INTER: // plot interpolated approx. P_delay
    {
      for (double T = T_min; T < T_max; T += step) {
        fprintf(stdout, "%lf %lf\n", T, P_T_delay_approx_inter(T, T_O));
      }
    }
    break;
  case HYBRID: // plot interpolated approx. P_delay
    {
      for (double T = T_min; T < T_max; T += step) {
        fprintf(stdout, "%lf %lf\n", T, P_T_delay_approx_hybrid(T, T_O));
      }
    }
    break;
  case P: // plot P_D
    {
      for (double T = T_min; T < T_max; T += step) {
        fprintf(stdout, "%lf %lf\n", T, P_D(T));
      }
    }
    break;
  default:
    throw 1;
  }
  
  delete pDistData;
  
  return 0;
}
