#include <math.h>
#include <assert.h>

#include "distrib.h"


double Step::P(double x) const {
  if (x < m_m) return 0;

  return 1;
}

double Exponential::P(double x) const {
  if (x < m_m) return 0;

  return 1 - exp(- (x - m_m) / m_a);
}

// comparisons are made within this precision
#define PRECISION 0.000001
// comparisons are made within 50 steps
#define STEPS 50

// E formula is:
// ( (m/c) - (m/c)^k ) * c * k /(k-1)/( 1-(m/c)^k )
// m: x_min
// c: cutoff
// k: shape parameter
double calculate_expect(double m, double k, double c) {
  const double rate = m / c;
  const double pow_of_rate = pow(rate, k);
  return ((rate - pow_of_rate) * c * k / (1 - pow_of_rate) / (k - 1));
}

// calculate the derivative of the expected value pareto_cutoff
// note: the function increases monotonously, so the derivative is positive
// e: expectation of pareto
// k: shape parameter, in (1,2]
// c: cutoff
double calculate_dir_of_expect(double m, double k, double c) {
  // sanity checks
  assert(k > 1 && k <= 2);
  assert(m > 0 && m <= c);

  const double rate = m / c;
  const double pow_of_rate = pow(rate, k);
  const double rest = 1 - pow_of_rate;
  return (k * (rest - k * (pow_of_rate / rate) + k * pow_of_rate)
          / rest / rest / (k - 1) );
}

// simple interpolation with newton method
// note: the function increases monotonously, so the derivative is positiv
// e: expectation from pareto
// k: shape parameter, in (1,2]
// c: cutoff
double calculate_x_min(double e, double k, double c){
  // sanity checks
  assert(k > 1 && k <= 2);
  assert(e > PRECISION && e < 4.5);

  double x_new = 0.05;
  double e_new = calculate_expect(x_new, k, c);
  double dir_e_new;
  int count = 0;

  while( fabs(e_new - e) > PRECISION  && count < STEPS ){
    count++;
    dir_e_new=calculate_dir_of_expect(x_new, k, c);
    x_new=x_new + ( e - e_new )/dir_e_new;
    if( x_new < 0 ) x_new = PRECISION;
    else if( x_new > c ) x_new = c - PRECISION;
    e_new=calculate_expect(x_new,k,c);
  }

  //cout << count ;
  return x_new;
}

// uses a fixed cutoff
Pareto::Pareto(double e, double k, double cutoff)
  : DistribFunction(),
    m_cutoff(cutoff), m_m(calculate_x_min(e, k, m_cutoff)), m_k(k) {
}

double Pareto::P(double x) const {
  if (x < m_m) return 0;
  if (x > m_cutoff) return 0;
  
  return 1 - pow(m_m / x, m_k);
}
