#!/usr/bin/perl
# Copyright (C) 2007, 2008  Condel project, TU-Berlin.
# Written by Jiehua Chen, Sven Grottke, Jan Sablatnig.
# 
# This file is part of Adam, the DVE testbed.
# 
# Adam is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3, or (at your option) any later
# version.
# 
# Adam is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
# 
# You should have received a copy of the GNU General Public License
# along with Adam; see the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.
# 
# Contact the Condel project at <http://www.math.tu-berlin.de/condel/>.

#
# usage: visu_swarm.pl [-c] START_TIME END_TIME views-parm-set vector-file
# if -c is given, splitvec will not be called
# views-parm-set must be the view-ids

use strict;

sub usage{
    print STDERR "usage: visu_swarm.pl [-c] START_TIME END_TIME views-parm-set vector-file\n";
}

sub failure{
    my $failedparm = shift;

    if( defined $failedparm ){
        print STDERR "$failedparm format is wrong\n";
    }else{
        print STDERR "parameters failed\n";
    }

    &usage;
    print STDERR "Please see the usage for more details.\n";
    die "aborting.";
}

my %views;
my %bee_exists;

sub insertValues{
    my $view_id = hex shift;
    my $bee_id = shift;
    my $type = shift;
    my $vec_id = shift;
    if( $bee_id =~ /[a-fA-F\d]+/ ){
        $bee_id = hex $bee_id;
        $bee_exists{$bee_id} = 1;
    }
    $views{$view_id}{$bee_id}{$type} = $vec_id;   
}

#sanity check
if( @ARGV < 4 ) { &failure; }

# parse parameters
my $argc = 0;
# check options
my $to_split = 0;
if( $ARGV[0] eq "-c" ){
    $to_split=1;
    $argc++;
}

# still, we need at least three arguments
if( @ARGV - $argc != 4 ) { &failure; }

# set some constants from parameters
if ( $ARGV[$argc] !~ /^[\.\d]+/ ) { &failure("start time"); }
my $start_time = $ARGV[$argc++];   # start_time
if ( $ARGV[$argc] !~ /^[\.\d]+/ ) { &failure("end time"); }
my $end_time = $ARGV[$argc++];     # end_time

my $views_parm_set = $ARGV[$argc++];

if( $ARGV[$argc] !~ /\.vec$/ ) { &failure("vec file"); }
my $vecfile = $ARGV[$argc];
my $swarm_result_base = `dirname $vecfile`;
#print STDERR "${swarm_result_base}\n";
chomp(${swarm_result_base});


# try to set CONDEL_ROOT
my $CONDEL_ROOT;
chomp($CONDEL_ROOT = `pwd`);
$CONDEL_ROOT =~ s:/[^/]+/[^/]+$::;

# set path to splitvec utility
my $SPLITVEC = "${CONDEL_ROOT}/omnetpp/bin/splitvec";
# set path to visualisation tool
my $VISU_TOOL = "${CONDEL_ROOT}/util/visu_swarm/visu_swarm";

print STDERR "START_TIME: ${start_time}\n";
print STDERR "END_TIME: ${end_time}\n";

# split the vector file unless $CACHE is set to yes
if( !$to_split ){
    print STDERR "split vecfile\n";
    system("$SPLITVEC $vecfile");
}
my @view_ids = split(/\s+/, $views_parm_set);

open(VECFILE, $vecfile);
# parse views-parm-set
# find the queen and bees ids in different views
my $vecfiles;
my $nr_views = @view_ids;

while (<VECFILE>){
    # parse line
    # parse queen attributes
    if( /^vector\s+([\d]+)\s+\"swarm_[a-z]+\.player@([a-fA-F\d]+)\[[\d]+\]\"\s+\"queenPosX\"/ ){
        insertValues($2, "q", "x", $1);
    }
    elsif( /^vector\s+([\d]+)\s+\"swarm_[a-z]+\.player@([a-fA-F\d]+)\[[\d]+\]\"\s+\"queenPosY\"/ ){
        insertValues($2, "q", "y", $1);
    }

    # parse bee attributes
    elsif( /^vector\s+([\d]+)\s+\"swarm_[a-z]+\.player@([a-fA-F\d]+)\[[\d]+\]\"\s+\"cBee@([a-fA-F\d]+).PosX\"/ ){
        insertValues($2, $3, "x", $1);
    }
    elsif( /^vector\s+([\d]+)\s+\"swarm_[a-z]+\.player@([a-fA-F\d]+)\[[\d]+\]\"\s+\"cBee@([a-fA-F\d]+).PosY\"/ ){
        insertValues($2, $3, "y", $1);
    }
    elsif( /^vector\s+([\d]+)\s+\"swarm_[a-z]+\.player@([a-fA-F\d]+)\[[\d]+\]\"\s+\"cBee@([a-fA-F\d]+).Dead\"/ ){
        insertValues($2, $3, "d", $1);
    }
    elsif( /^vector\s+([\d]+)\s+\"swarm_[a-z]+\.player@([a-fA-F\d]+)\[[\d]+\]\"\s+\"cScore@([a-fA-F\d]+)\"/ ){
        insertValues($2, $3, "s", $1);
  }   
}

# map the bee ids
my $nummer = 0;
my %num_to_id;

my @id_order;
for (@view_ids){
    if (! exists $bee_exists{ hex $_}){
        failure("unknown view id!");
    }
    push(@id_order, hex $_);
    delete $bee_exists{hex $_};
}
push(@id_order, sort { $a <=> $b } keys %bee_exists);


# TODO:  find correct order
# my %num_to_id = { 0 => 4, 1 => 2, 2 => 3, 3 => 5, 4 => 6, 5 => 7};
#0 1 2 3 4 5
#4 2 3 5 6 7
#print STDERR @id_order;

for my $vid ( @view_ids ){
    next if( !exists $views{$vid} );      
    $vecfiles .= "${swarm_result_base}/omnetpp-$views{$vid}{q}{x}.vec ${swarm_result_base}/omnetpp-$views{$vid}{q}{y}.vec ";
    for my $bid ( @id_order ){
        next if( !exists $views{$vid}{$bid} );
        $vecfiles .= "${swarm_result_base}/omnetpp-$views{$vid}{$bid}{x}.vec ${swarm_result_base}/omnetpp-$views{$vid}{$bid}{y}.vec ${swarm_result_base}/omnetpp-$views{$vid}{$bid}{s}.vec ${swarm_result_base}/omnetpp-$views{$vid}{$bid}{d}.vec ";
    }
}

# call it
print STDERR "${VISU_TOOL} ${start_time} ${end_time} ${nr_views} ${vecfiles}\n";
system("${VISU_TOOL} ${start_time} ${end_time} ${nr_views} ${vecfiles}");
